; -----------------------------------------------------------------------------
; 680x0 Second Processor CiscOS v1.99 (c) Eelco Huininga 2007-2015
; DOS Filename   : bios.x68
; Load address   : FFFF0000 for 68020+ / 003F0000 for 68008
; Exec address   : FFFF0800 for 68020+ / 003F0800 for 68008
; Length         : 00008000
; -----------------------------------------------------------------------------
;
; Thanks to J.G.Harston for his code reviews and additions
;
; v1.99e (20150302)
; * Renamed the OS to CiscOS (in reference to RiscOS)
; * Updated: SWI &06: OS_Byte &8E handler (Enter language ROM) - program start time is now used
; * Updated: reset_entry: os start time is now used
; * Updated: reset_entry: interrupts are now disabled during boot time
; * Updated: Completely rewrote the OS_Word handler
; * Updated: *XFER handler is now working correctly
; * Updated: Startup message
; * Updated: *MON handler is now working correctly
; * Updated: *HELP handler is now working correctly
; * Updated: unknown_exception handler now shows extended information
; * Updated: SWI &01: OS_WriteC - now calls wrchv instead of wrchv calling OS_Writec. This ensures that all output can be redirected by intercepting wrchv
; * Updated: SWI &19: OS_UnusedSWI - Now preserves registers on exit
; * Updated: SWI &29: OS_FSControl - Control the filing system manager and the filing system itself
; * Updated: SWI &2B: OS_GenerateError - speed
; * Updated: SWI &40: OS_ChangeEnvitonment - speed
; * Updated: SWI &5D: OS_PrintChar - bugfixes
; * Updated: SWI &7D: bugfixes
; * Updated: Optimized the SWI handler and SWI table 
; * Updated: Address error exception handler: now correctly displays the registers pushed to the stack
; * Updated: SWI_OS_ConvertHex to correctly handle nibbles in the A-F range
; * Updated: SWI table OS_Word description (was: OS_Byte)
; * Updated: not_supported used A5 as SWI register instead of A0
; * Updated: reset_entry: Set the stack pointer temporary to $00000500 so we can safely jump to subroutines like checkramsize
; * Updated: checkramsize: changed the address from $00000000 to $00000008 to prevent hangup on reset when in the checkramsize function
; * Added: Integer divide by zero handler
; * Added: Illegal instruction handler
; * Added: osstarttime variable
; * Added: programstarttime variable
; * Added: error_pc variable
; * Added: SWI &10: OS_GetEnv: Read environment parameters
; * Added: SWI &12: OS_SetEnv: Set environment parameters
; * Added: SWI &32: OS_ReadPoint: Returns the colour and tint of a point
; * Added: SWI &34: OS_CallAVector: Call a vector directly
; * Added: SWI &36: OS_RemoveCursors: Remove the cursors from the screen (VDU 23,1,0;0;0;0;)
; * Added: SWI &37: OS_RestoreCursors: Restore the cursors from the screen (VDU 23,1,1;0;0;0;)
; * Added: SWI &3F: OS_CheckModeValid: Check if it is possible to change to a specified screen mode
; * Added: SWI &56: OS_SetECFOrigin: Set the origin of the ECF pattern (VDU23,17,6,x;y;0,0,0)
;
; v1.99d (20140806):
; * Changed: INT5 to INT2 (TUBE IRQ) for easy designing MC68008P circuits
; * Changed: INT7 to INT5 (TUBE NMI) for easy designing MC68008P circuits
;
; v1.99c (20140319):
; * Added: *HELP SWI handler (displays all supported SWI calls)
; * Added: *XFER
; * Added: NMI handler (JGH/EH)
; * Added: Privilege exception handler
; * Updated: *HELP message: *HELP now shows a short help message, and *HELP TUBE now shows the extended help message
; * Updated: Updated the low-level TUBE calls (JGH) and renamed them to make them more consistent (EH)
; * Updated: SWI &40: SWI_OS_ChangeEnvironment: now re-routed through changeenvironmentv vector
; * Updated: Replaced enter_language_rom with SWI_OS_Byte_8E
; * Updated: SWI &05: OS_CLI handler; it now uses a jump table
; * Updated: SWI &0C: OS_GBPB handler (JGH/EH)
; * Added: Escape flag
; * Added: SWI &13: OS_IntOn - Enable interrupts
; * Added: SWI &14: OS_IntOff - Disable interrupts
; * Added: SWI &2C: OS_ReadEscapeState: Check whether an escape contition has occurred
; * Added: SWI &5B: OS_CRC: Calculate 16-bit CRC value
;
; v1.99b (20140303):
; * Updated: Replaced dump_d0 routine with SWI &D4:OS_ConvertHex8
; * Updated: Replaced dump_sr routine with SWI &E0:OS_ConvertBinary2
; * Updated: findram procedure is now called checkramsize
; * Added: SWI &0F: OS_Control: Read or write address handlers
; * Added: SWI &16: OS_EnterOS - Switch to supervisor mode
; * Added: SWI &19: OS_UnusedSWI: Read or write handler for unused SWI's
; * Added: SWI &29: OS_FSControl: Control the filing system manager and the filing system itself
; * Added: SWI &29: OS_FSControl $10: Set file system options
; * Added: SWI &29: OS_FSControl $16: Close all open files
; * Added: SWI &3A: OS_ValidateAddress: Check that a range of addresses are in logical RAM
; * Added: SWI &40: OS_ChangeEnvironment: Read or write address handlers (handler 6, 9, 10)
; * Added: SWI &45: OS_Plot: Perform a direct PLOT call using VDU
; * Added: SWI &46: OS_WriteN: Write the first D1 characters of the string pointed to by D0
; * Added: SWI &5D: OS_PrintChar: Send character to the printer
; * Added: SWI &7C: OS_LeaveOS - Switch to user mode
; * Added: SWI &D0: OS_ConvertHex1
; * Added: SWI &D1: OS_ConvertHex2
; * Added: SWI &D2: OS_ConvertHex4
; * Added: SWI &D3: OS_ConvertHex6
; * Added: SWI &D4: OS_ConvertHex8
; * Added: SWI &DD: OS_ConvertBinary1
; * Added: SWI &DE: OS_ConvertBinary2
; * Added: SWI &DF: OS_ConvertBinary3
; * Added: SWI &E0: OS_ConvertBinary4
; * Added: Added beeb (&07) to startup message :-)
;
; v1.99a (20140301):
; Major rewrite of the v0.99 code, abandoning the old 6502-style API calls and implementing ARM SWI-style API calls.
; * Replaced old 6502-style vectors with new SWI-style vectors
;   D0-7=R0-7, A0=SWI, A1-2=R8-9, A4=R12 (Workspace pointer)
; * Replaced old 6502-style API calls and replaced them with new SWI-style API calls
; * RISC OS style SWI handler written for TRAP 12 (JGH/EH)
; * Removed jump table at $FFFFFF7C (you can use SWI calls from now on)
; * HIMEM, OSHWM and Stack Pointer are now set accordingly to size of RAM
; * Startup message now correctly displays CPU type and RAM size (JGH)
; * Error handler is now SWI &2B: OS_GenerateError
; * Error handler now routes errors through the ErrorV vector
; * SWI handler now uses jump table (JGH/EH)
; * SWI handler now routes known SWI's with a Software Vector through it's vector addresses
; * SWI Handler now routes unknown SWI's through the UKSWIV vector
; * Exception handler now prints debug information
; * Subroutines grouped according to functionality (JGH)
; * Optimized SWI &06: OS_Byte (JGH)
; * Added: SWI &0E: OS_ReadLine: Read a line from the current input stream
; * Added: SWI &7D: OS_ReadLine32: Read a line from the current input stream
; * checkramsize now returns the RAM size in bytes in D0
; * Renamed some messages and errors
; * Error numbers are now 32-bit
; * Added: Built-in application jump table (JGH)
; * Replaced OSWORD &00 calls with OS_ReadLine32 calls (also removed the OSWORD &00 parameterblock)
; * OSWORD &00 now uses hoaddr as high order address for buffer
; * Added IRQ handler (JGH)
; * Different error messages for unknown exceptions and interrupts
; * Various code optimizations (JGH/EH)
; * Added: changelog
;
; Todo / known issues:
; Write NMI handler function 4 to 7
; Finish *FLASH code so it runs in RAM and programs 28F256
; Check for register consistency (also: compatibility with ARM SWI calls)
; Error handling (see also PRM 1-28):
;   All SWI`s should setup an error block on exit if an error occurred. On exit, D0 should point to error block and V should be set.
;   UKSWIV should be replaced by setting up a SWI error block: error handling should be done by the SWI routine itself
;   Change error handler: should set up an error block, +0-+3: PC where error occurred, +4-+7: Error nr., +8...: Error string terminated by &00
;   Known issue: XOS_GenerateError is handled exactly the same as OS_GenerateError (not compliant with RiscOS)
; Escape flag: Isn`t vectored yet; find out how escv should be implemented in 32-bit RiscOS style OS-es
; Write function hex2dec (OS_ConvertInteger) (used for displaying RAM size at startup) (Use OS_ConvertFixedFileSize/OS_ConvertFileSize?)
; Write SWI &40: OS_ChangeEnvironment handlers:
;    0: Memory limit ()
;    1: Undefined instruction (reroute through 68k 4: Illegal instruction)
;    2: Prefetch abort ()
;    3: Data abort ()
;    4: Address exception (reroute through 68k 3: Address error)
;    5: Other exceptions ()
;    7: CallBack ()
;    8: BreakPoint ()
;    9: Escape ()
;   11: Exit ()
;   13: Exception registers ()
;   14: Application space ()
;   15: Currently active object ()
;   16: UpCall ()
;   If an attempt is made to access protected memory from an insufficiently privileged mode, MEMC traps the exception and sends an abort signal to the ARM.
; Write &15:OS_CallBack routine (reroute through OS_ChangeEnvironment)
; Write &18:OS_BreakCtrl routine (reroute through OS_ChangeEnvironment)
; Write &1C:OS_Mouse routine using either OSWORD &40 or to read the mouse position using ADVAL(7) and ADVAL(8) and the buttons with INKEY(-10) INKEY(-11) INKEY(-12) : check if OSWORD &40 is properly handled by the OS_Word handler
; Write &1F:OS_Claim and &20:OS_Release (Claim and release vector) related to &40:OS_ChangeEnvironment
; Write &21:OS_ReadUnsigned as a replacement for the get_hex_value
; Write &29:OS_FSControl other handlers
; Write &38:OS_SWINumberToString
; Write &39:OS_SWINumberFromString
; Write &48:OS_WriteEnv
; Write &55:OS_ReadDefaultHandler
; Write &59:OS_Confirm (use the code in the *FLASH handler)
; Write &EB:OS_ConvertFixedFileSize
; Write &EC:OS_ConvertFileSize
; Write &280:Cache_Control (and *CACHE handler)
; Write &40380:Debugger_Disassemble
; Write &43F40:Joystick_Read
; SWI &34: OS_CallAVector: pass Carry and oVerflow bits to vector
; SWI &40: handler 6 (error) returns previous error buffer address in D3: is this compatible with RiscOS? (used by &0F:OS_Control)
; SWI &45: OS_Plot : Unknown plot codes (216-231,240-255) are sent to the unknown plot vector?, which by default does nothing.
; SWI &72 not finished yet
; SWI &7D write handler for flags in D4: rewrite code so it runs locally on the copro, not using the OSWORD &00 TUBE call but OS_ReadC instead
; SWI &D5: OS_ConvertCardinal1: Debug and check routine
; SWI &D6: OS_ConvertCardinal2: Debug and check routine
; SWI &D7: OS_ConvertCardinal3: Debug and check routine
; SWI &D8: OS_ConvertCardinal4: Debug and check routine
; Finish SWI &C0: SWI_OS_ConvertStandardDateAndTime and SWI &C1: SWI_OS_ConvertDateAndTime
; When starting a program, get the current time (5 bytes) and store it somewhere in memory together with pointer to *command (used by SWI &10:OS_GetEnv and others)
; Describe memory layout correctly for various CPU`s
; Switch from/to supervisor/user-mode at appropiate locations in the code
; Efficiency: make generic set/clear overflow/carry flag and RTS exits for all SWI`s (BRA is 4 bytes vs. AND+RTS is 6 bytes)
; Known issue: SWI &07:OS_Word $00 handler doesn`t preserve the Carry flag returned by the OS_ReadLine32 routine
; Known issue: SWI &0C:OS_OSGBPB - reason &09-&0C are not supported by standard BBC B with DFS: find out if this is gonna be a problem
; Known issue: SWI &7C:OS_LeaveOS - supervisor and user stack are different so the RTS returns to a long stored on the user stack, but the addresses is actually stored on the supervisor stack
; Known issue: SWI &06:OS_Byte $82, $83, $84 and $9D - don't know if these give the correct value
; Known issue: The error_pc variable is not updated every time an error is generated
; Known issue: Address exception handler: now probably only handles 68000/68008 stacks.
; Enable/disable interrupts at appropiate locations in the code
; Differentiate between SSP and USP (System Stack Pointer & User Stack Pointer)
; Follow all jumps to check all RTS' and RTE's if they are used in the correct order
; Internal * programs:
;   *ERROR
;   *FLASH: Flash ROM routines
;   *QUIT
;   *TZAP: Memory editor for TUBE
;   *XFER: * use generic string API calls for getting arguments???
; Decide wether or not the 'current' variable is nessecary
; Change exception_handler so it generates RiscOS compatible errors (see PRM 1-43)
; When jumping into code (enter language/OSBYTE etc) the check-for-copyright string should use the copyright offset byte at +&0007



; -----------------------------------------------------------------------------
; ALIGN macro
; -----------------------------------------------------------------------------

ALIGN   MACRO
	IFARG   1
	DS.B    (*-1)/\1*\1+\1-*
	MEXIT
	ENDC
	DS.B    (*+3)&-4-*
	ENDM

; -----------------------------------------------------------------------------
; Compiler constants
; -----------------------------------------------------------------------------


DEBUG	SET 1

; Memory map for 68008 issue 3 board:
; RAM		000000-2FFFFF
; (nothing)	300000-3DFFFF
; TUBECS	3E0000-3EFFFF
; ROMCS		3F0000-3FFFFF


SWI                                     EQU 12			; Use TRAP #12 for SWI calls
stacksize                               EQU $00000200		; Size of stack is 512 bytes
prompt                                  EQU $2A			; Prompt
flash_rom_sectors			EQU $08			; Am29F040 has 8 sectors, 64k each
max_ram_size                            EQU $00300000           ; Address at which checkramsize will stop
bios_address				EQU $003F0000		; BIOS base address FFFF0000-FFFFFFFF (3F0000-3FFFFF for 68008)
tube_address                            EQU $003E0000           ; TUBE base address FFFE0000-FFFEFFFF
bios_version                            SET 'v1.99c (20140314)'	; BIOS version


r1status                                EQU tube_address        ; TUBE ULA addresses
r1data                                  EQU tube_address+1
r2status                                EQU tube_address+2
r2data                                  EQU tube_address+3
r3status                                EQU tube_address+4
r3data                                  EQU tube_address+5
r4status                                EQU tube_address+6
r4data                                  EQU tube_address+7


OS_WriteC                               EQU $00000000           ; SWI calls
OS_WriteS                               EQU $00000001
OS_Write0                               EQU $00000002
OS_NewLine                              EQU $00000003
OS_ReadC                                EQU $00000004
OS_CLI                                  EQU $00000005
OS_Byte                                 EQU $00000006
OS_Word                                 EQU $00000007
OS_File                                 EQU $00000008
OS_Args                                 EQU $00000009
OS_BGet                                 EQU $0000000A
OS_BPut                                 EQU $0000000B
OS_GBPB                                 EQU $0000000C
OS_Find                                 EQU $0000000D
OS_ReadLine                             EQU $0000000E
OS_Control                              EQU $0000000F
OS_GetEnv                               EQU $00000010
OS_Exit                                 EQU $00000011
OS_SetEnv                               EQU $00000012
OS_IntOn                                EQU $00000013
OS_IntOff                               EQU $00000014
OS_CallBack                             EQU $00000015
OS_EnterOS                              EQU $00000016
OS_BreakPt                              EQU $00000017
OS_BreakCtrl                            EQU $00000018
OS_UnusedSWI                            EQU $00000019
OS_UpdateMEMC                           EQU $0000001A
OS_SetCallBack                          EQU $0000001B
OS_Mouse                                EQU $0000001C
OS_Heap                                 EQU $0000001D
OS_Module                               EQU $0000001E
OS_Claim                                EQU $0000001F
OS_Release                              EQU $00000020
OS_ReadUnsigned                         EQU $00000021
OS_GenerateEvent                        EQU $00000022
OS_ReadVarVal                           EQU $00000023
OS_SetVarVal                            EQU $00000024
OS_GSInit                               EQU $00000025
OS_GSRead                               EQU $00000026
OS_GSTrans                              EQU $00000027
OS_BinaryToDecimal                      EQU $00000028
OS_FSControl                            EQU $00000029
OS_ChangeDynamicArea                    EQU $0000002A
OS_GenerateError                        EQU $0000002B
OS_ReadEscapeState                      EQU $0000002C
OS_EvaluateExpression                   EQU $0000002D
OS_SpriteOp                             EQU $0000002E
OS_ReadPalette                          EQU $0000002F
OS_ServiceCall                          EQU $00000030
OS_ReadVduVariables                     EQU $00000031
OS_ReadPoint                            EQU $00000032
OS_UpCall                               EQU $00000033
OS_CallAVector                          EQU $00000034
OS_ReadModeVariable                     EQU $00000035
OS_RemoveCursors                        EQU $00000036
OS_RestoreCursors                       EQU $00000037
OS_SWINumberToString                    EQU $00000038
OS_SWINumberFromString                  EQU $00000039
OS_ValidateAddress                      EQU $0000003A
OS_CallAfter                            EQU $0000003B
OS_CallEvery                            EQU $0000003C
OS_RemoveTickerEvent                    EQU $0000003D
OS_InstallKeyHandler                    EQU $0000003E
OS_CheckModeValid                       EQU $0000003F
OS_ChangeEnvironment                    EQU $00000040
OS_ClaimScreenMemory                    EQU $00000041
OS_ReadMonotonicTime                    EQU $00000042
OS_SubstituteArgs                       EQU $00000043
OS_PrettyPrint                          EQU $00000044
OS_Plot                                 EQU $00000045
OS_WriteN                               EQU $00000046
OS_AddToVector                          EQU $00000047
OS_WriteEnv                             EQU $00000048
OS_ReadArgs                             EQU $00000049
OS_ReadRAMFsLimits                      EQU $0000004A
OS_ClaimDeviceVector                    EQU $0000004B
OS_ReleaseDeviceVector                  EQU $0000004C
OS_DelinkApplication                    EQU $0000004D
OS_RelinkApplication                    EQU $0000004E
OS_HeapSort                             EQU $0000004F
OS_ExitAndDie                           EQU $00000050
OS_ReadMemMapInfo                       EQU $00000051
OS_ReadMemMapEntries                    EQU $00000052
OS_SetMemMapEntries                     EQU $00000053
OS_AddCallBack                          EQU $00000054
OS_ReadDefaultHandler                   EQU $00000055
OS_SetECFOrigin                         EQU $00000056
OS_SerialOp                             EQU $00000057
OS_ReadSysInfo                          EQU $00000058
OS_Confirm                              EQU $00000059
OS_ChangedBox                           EQU $0000005A
OS_CRC                                  EQU $0000005B
OS_ReadDynamicArea                      EQU $0000005C
OS_PrintChar                            EQU $0000005D
OS_ChangeRedirection                    EQU $0000005E
OS_RemoveCallBack                       EQU $0000005F
OS_FindMemMapEntries                    EQU $00000060
OS_SetColour                            EQU $00000061
OS_Pointer                              EQU $00000064
OS_ScreenMode                           EQU $00000065
OS_DynamicArea                          EQU $00000066
OS_Memory                               EQU $00000068
OS_ClaimProcessorVector                 EQU $00000069
OS_Reset                                EQU $0000006A
OS_MMUControl                           EQU $0000006B
OS_ResyncTime                           EQU $0000006C
OS_PlatformFeatures                     EQU $0000006D
OS_SynchroniseCodeAreas                 EQU $0000006E
OS_CallASWI                             EQU $0000006F
OS_AMBControl                           EQU $00000070
OS_CallASWIR12                          EQU $00000071
OS_SpecialControl                       EQU $00000072
OS_EnterUSR32                           EQU $00000073
OS_EnterUSR26                           EQU $00000074
OS_VIDCDivider                          EQU $00000075
OS_NVMemory                             EQU $00000076
OS_ClaimOSSWI                           EQU $00000077
OS_TaskControl                          EQU $00000078
OS_DeviceDriver                         EQU $00000079
OS_Hardware                             EQU $0000007A
OS_IICOp                                EQU $0000007B
OS_LeaveOS                              EQU $0000007C
OS_ReadLine32                           EQU $0000007D
OS_Substitute                           EQU $0000007E
OS_HeapSort32                           EQU $0000007F
OS_ConvertStandardDateAndTime           EQU $000000C0
OS_ConvertDateAndTime                   EQU $000000C1
OS_ConvertHex1                          EQU $000000D0
OS_ConvertHex2                          EQU $000000D1
OS_ConvertHex4                          EQU $000000D2
OS_ConvertHex6                          EQU $000000D3
OS_ConvertHex8                          EQU $000000D4
OS_ConvertCardinal1                     EQU $000000D5
OS_ConvertCardinal2                     EQU $000000D6
OS_ConvertCardinal3                     EQU $000000D7
OS_ConvertCardinal4                     EQU $000000D8
OS_ConvertInteger1                      EQU $000000D9
OS_ConvertInteger2                      EQU $000000DA
OS_ConvertInteger3                      EQU $000000DB
OS_ConvertInteger4                      EQU $000000DC
OS_ConvertBinary1                       EQU $000000DD
OS_ConvertBinary2                       EQU $000000DE
OS_ConvertBinary3                       EQU $000000DF
OS_ConvertBinary4                       EQU $000000E0
OS_ConvertSpacedCardinal1               EQU $000000E1
OS_ConvertSpacedCardinal2               EQU $000000E2
OS_ConvertSpacedCardinal3               EQU $000000E3
OS_ConvertSpacedCardinal4               EQU $000000E4
OS_ConvertSpacedInteger1                EQU $000000E5
OS_ConvertSpacedInteger2                EQU $000000E6
OS_ConvertSpacedInteger3                EQU $000000E7
OS_ConvertSpacedInteger4                EQU $000000E8
OS_ConvertFixedNetStation               EQU $000000E9
OS_ConvertNetStation                    EQU $000000EA
OS_ConvertFixedFileSize                 EQU $000000EB
OS_ConvertFileSize                      EQU $000000EC
OS_WriteI                               EQU $00000100       ; $100-$1FF is VDU$00-VDU$FF
ADFS_DiscOp                             EQU $00040240
ADFS_HDC                                EQU $00040241
ADFS_Drives                             EQU $00040242
ADFS_FreeSpace                          EQU $00040243
ADFS_Retries                            EQU $00040244
ADFS_DescribeDisc                       EQU $00040245
ADFS_VetFormat                          EQU $00040246
ADFS_FlpProcessDCB                      EQU $00040247
ADFS_ControllerType                     EQU $00040248
ADFS_PowerControl                       EQU $00040249
ADFS_SetIDEController                   EQU $0004024A
ADFS_IDEUserOp                          EQU $0004024B
ADFS_MiscOp                             EQU $0004024C
ADFS_ECCSAndRetries                     EQU $00040250


buffer					EQU $00000600       ; OSWORD &00, OSCLI and Error buffer
errorbuffer				EQU $00000600       ; OSWORD &00, OSCLI and Error buffer
os_workspace				EQU $00000500       ; OS workspace: SWI vectors / jump table


; -----------------------------------------------------------------------------
; Start of code
; -----------------------------------------------------------------------------

START ORG bios_address
vector_table                                                ; CPU Reset vectors
	                DC.L	$00008000                   ;   0: Initial stack pointer (Top of RAM; in this example RAM would go from 0000 to 7FFF)
	                DC.L	reset_entry                 ;   1: Initial program counter (start address)
	                                                    ; CPU Exception vectors
	                DC.L	bus_error_handler           ;   2: Bus error
	                DC.L	address_error_handler       ;   3: Address error
	                DC.L	illegal_instruction_handler ;   4: Illegal instruction
	                DC.L	int_divide_by_zero_handler  ;   5: Integer divide-by-zero
	                DC.L	unknown_exception           ;   6: CHK, CHK2 instruction
	                DC.L	unknown_exception           ;   7: FTRAPcc, TRAPcc, TRAPV instruction
	                DC.L	privilege_violation_handler ;   8: Privilege violation
	                DC.L	unknown_exception           ;   9: Trace
	                DC.L	unknown_exception           ;  10: Line 1010 emulator (unimplemented A-line opcode)
	                DC.L	unknown_exception           ;  11: Line 1111 emulator (unimplemented F-line opcode) / Floating-point unimplemented instruction / Floating-point disabled
	                DC.L	unknown_exception           ;  12: Emulator interrupt (MC68060 only)
	                DC.L	unknown_exception           ;  13: Coprocessor Protocol Violation (MC68020 & MC68030 only)
	                DC.L	unknown_exception           ;  14: Format Error ( >= MC68010)
	                DC.L	unknown_exception           ;  15: Uninitialized interrupt vector
	                DC.L	unknown_exception           ;  16: (reserved)
	                DC.L	unknown_exception           ;  17: (reserved)
	                DC.L	unknown_exception           ;  18: (reserved)
	                DC.L	unknown_exception           ;  19: (reserved)
	                DC.L	unknown_exception           ;  20: (reserved)
	                DC.L	unknown_exception           ;  21: (reserved)
	                DC.L	unknown_exception           ;  22: (reserved)
	                DC.L	unknown_exception           ;  23: (reserved)
	                DC.L	unknown_exception           ;  24: Spurious Interrupt (Bus error during interrupt)
	                                                    ; Auto-vector interrupts
	                DC.L	unknown_exception           ;  25: Level-1 autovector (INT1)
	                DC.L	irq_handler                 ;  26: Level-2 autovector (INT2) TUBE ULA IRQ
	                DC.L	unknown_exception           ;  27: Level-3 autovector (INT3)
	                DC.L	unknown_exception           ;  28: Level-4 autovector (INT4)
	                DC.L	nmi_handler                 ;  29: Level-5 autovector (INT5) TUBE ULA NMI
	                DC.L	unknown_exception           ;  30: Level-6 autovector (INT6)
	                DC.L	unknown_exception           ;  31: Level-7 autovector (INT7)
	                                                    ; TRAP instruction vectors
	                                                    ; Atari uses TRAP #1 (Atari GEMDOS BDOS interface), #2 (Atari AES and VDI), #13 (BIOS) and #14 (XBIOS)
	                                                    ; Sinclair QDOS uses TRAP #0 (Switch to supervisor mode), #1 (QDOS manager), #2 (QDOS basic I/O), #3 (QDOS complex I/O) and #4 (BASIC command interpreter)
	                DC.L	unknown_exception           ;  32: TRAP 0 vector (Unix system calls & OS/9 system call)
	                DC.L	unknown_exception           ;  33: TRAP 1 vector (Unix indirect system calls & Atari GEMDOS BDOS interface)
	                DC.L	unknown_exception           ;  34: TRAP 2 vector (CP/M BDOS calls & Atari AES and VDI)
	                DC.L	unknown_exception           ;  35: TRAP 3 vector (CP/M BIOS calls)
	                DC.L	unknown_exception           ;  36: TRAP 4 vector
	                DC.L	unknown_exception           ;  37: TRAP 5 vector
	                DC.L	unknown_exception           ;  38: TRAP 6 vector
	                DC.L	unknown_exception           ;  39: TRAP 7 vector
	                DC.L	unknown_exception           ;  40: TRAP 8 vector
	                DC.L	unknown_exception           ;  41: TRAP 9 vector
	                DC.L	unknown_exception           ;  42: TRAP 10 vector
	                DC.L	unknown_exception           ;  43: TRAP 11 vector
	                DC.L	SWI_Handler                 ;  44: TRAP 12 vector 68k BIOS SWI handler
	                DC.L	unknown_exception           ;  45: TRAP 13 vector (Atari BIOS / OS/9 C I/O calls)
	                DC.L	unknown_exception           ;  46: TRAP 14 vector (Atari XBIOS)
	                DC.L	unknown_exception           ;  47: TRAP 15 vector (OS/9 Extended math calls)
	                DC.L	unknown_exception           ;  48: FPU: Branch or Set on Unordered Condition
	                DC.L	unknown_exception           ;  49: FPU: Inexact Result
	                DC.L	unknown_exception           ;  50: FPU: Divide by Zero
	                DC.L	unknown_exception           ;  51: FPU: Underflow
	                DC.L	unknown_exception           ;  52: FPU: Operand Error
	                DC.L	unknown_exception           ;  53: FPU: Overflow
	                DC.L	unknown_exception           ;  54: FPU: Signaling NAN (Not-A-Number)
	                DC.L	unknown_exception           ;  55: FPU: Unimplemented data type (MC68040 & MC68060 only)
	                DC.L	unknown_exception           ;  56: MMU: Configuration Error (MC68030 & MC68851 only)
	                DC.L	unknown_exception           ;  57: MMU: Illegal Operation Error (MC68851 only)
	                DC.L	unknown_exception           ;  58: MMU: Access Level Violation Error (MC68851 only)
	                                                    ; Miscellaneous vectors
	                DC.L	unknown_exception           ;  59: (reserved)
	                DC.L	unknown_exception           ;  60: FPU: Unimplemented effective address (MC68060 only)
	                DC.L	unknown_exception           ;  61: Unimplemented integer instruction (MC68060 only)
	                DC.L	unknown_exception           ;  62: (reserved)
	                DC.L	unknown_exception           ;  63: (reserved)
	                                                    ;  64-255: user (acknowledged) interrupt vectors

; -----------------------------------------------------------------------------
; OS workspace: Software Vectors
; -----------------------------------------------------------------------------
	ORG     bios_address+$00000400
	                                                    ; RiscOS vectors
userv	                DC.L	not_supported               ; $00000400 $00 UserV              (Not implemented)
errorv                  DC.L	_errorv                     ; $00000404 $01 ErrorV
irqv	                DC.L	_irqv                       ; $00000408 $02 IrqV
wrchv	                DC.L	_wrchv                      ; $0000040C $03 WrchV
readcv	                DC.L	SWI_OS_ReadC                ; $00000410 $04 ReadCV
cliv	                DC.L	SWI_OS_CLI                  ; $00000414 $05 CliV
bytev	                DC.L	SWI_OS_Byte                 ; $00000418 $06 ByteV
wordv	                DC.L	SWI_OS_Word                 ; $0000041C $07 WordV
filev	                DC.L	SWI_OS_File                 ; $00000420 $08 FileV
argsv	                DC.L	SWI_OS_Args                 ; $00000424 $09 ArgsV
bgetv	                DC.L	SWI_OS_Bget                 ; $00000428 $0A BgetV
bputv	                DC.L	SWI_OS_Bput                 ; $0000042C $0B BputV
gbpbv	                DC.L	SWI_OS_GBPB                 ; $00000430 $0C GBPBV
findv	                DC.L	SWI_OS_Find                 ; $00000434 $0D FindV
readlinev               DC.L	SWI_OS_ReadLine             ; $00000438 $0E ReadLineV
fscontrolv              DC.L	SWI_OS_FSControl            ; $0000043C $0F FSControlV
eventv	                DC.L	_eventv                     ; $00000440 $10 EventV
	                DC.L	$0                          ; $00000444 $11 Reserved (UPTV ?)
	                DC.L	$0                          ; $00000448 $12 Reserved (NETV ?)
keyv	                DC.L	not_supported               ; $0000044C $13 KeyV               (Not implemented)
insv	                DC.L	not_supported               ; $00000450 $14 InsV               (Not implemented)
remv	                DC.L	not_supported               ; $00000454 $15 RemV               (Not implemented)
cnpv                    DC.L	not_supported               ; $00000458 $16 CnpV               (Not implemented)
ukvdu23v                DC.L	not_supported               ; $0000045C $17 UKVDU23V           (Not implemented)
ukswiv                  DC.L	_ukswiv                     ; $00000460 $18 UKSWIV
ukplotv                 DC.L	not_supported               ; $00000464 $19 UKPLOTV            (Not implemented)
mousev                  DC.L	not_supported               ; $00000468 $1A MouseV             (Not implemented)
vduxv                   DC.L	not_supported               ; $0000046C $1B VDUXV              (Not implemented)
tickerv                 DC.L	not_supported               ; $00000470 $1C TickerV            (Not implemented)
upcallv                 DC.L	not_supported               ; $00000474 $1D UpCallV            (Not implemented)
changeenvironmentv      DC.L	_ChangeEnvironmentV         ; $00000478 $1E ChangeEnvironmentV
spritev                 DC.L	not_supported               ; $0000047C $1F SpriteV            (Not implemented)
drawv                   DC.L	not_supported               ; $00000480 $20 DrawV              (Not implemented)
econetv                 DC.L	not_supported               ; $00000484 $21 EconetV            (Not implemented)
colourv                 DC.L	not_supported               ; $00000488 $22 ColourV            (Not implemented)
palettev                DC.L	not_supported               ; $0000048C $23 PaletteV           (Not implemented)
serialv                 DC.L	not_supported               ; $00000490 $24 SerialV            (Not implemented)
fontv                   DC.L	not_supported               ; $00000494 $25 FontV              (Not implemented)
pointerv                DC.L	not_supported               ; $00000498 $26 PointerV           (Not implemented)
timesharev              DC.L	not_supported               ; $0000049C $27 TimeShareV         (Not implemented)
lowpriorityeventv       DC.L	not_supported               ; $000004A0 $28 LowPriorityV       (Not implemented)
fasttickerv             DC.L	not_supported               ; $000004A4 $29 FastTickerV        (Not implemented)
graphicsv               DC.L	not_supported               ; $000004A8 $2A GraphicsV          (Not implemented)
	                DC.L	$0                          ; $000004AC $2B Reserved
	                DC.L	$0                          ; $000004B0 $2C Reserved
	                DC.L	$0                          ; $000004B4 $2D Reserved
	                DC.L	$0                          ; $000004B8 $2E Reserved
	                DC.L	$0                          ; $000004BC $2F Reserved


; -----------------------------------------------------------------------------
; OS workspace: OS variables
; -----------------------------------------------------------------------------
	ORG     $000004C0

himem	                DC.L	$00000000                   ; $000004C0 HIMEM; top of RAM excluding the stack
oshwm	                DC.L	$00000000                   ; $000004C4 OSHWM; bottom of RAM; reserve page 4-7 for the OS
ramsize	                DC.L	$00000000                   ; $000004C8 RAMSIZE; Total size of RAM available
error	                DC.L	$00000000                   ; $000004CC Pointer to last generated error block
error_pc                DC.L	$00000000                   ; $000004D0 Pointer to location where last error occurred
current
langrom	                DC.L	$00000000                   ; $000004D4 Pointer to start of language ROM
tubetransfer            DC.L	$00000000                   ; $000004D8 Start address used for TUBE data transfers (NMI/INT2)
hoaddr	                DC.W	$0000                       ; $000004DC High order pointer
osstarttime             DC.B	$00, $00, $00, $00, $00     ; $000004DE OS start time (5 bytes)
programstarttime        DC.B	$00, $00, $00, $00, $00     ; $000004E3 Program start time (5 bytes)
cputype	                DC.B	$00                         ; $000004E8 CPU type ($01=68008, $02=68000, $03=68010, $04=68020, $05=68030, $06=68040, $07=68060, $FF=unknown CPU)
fputype	                DC.B	$00                         ; $000004E9 FPU type ($00=no FPU, $01=68881, $02=68882, $03=CPU integrated(68040/68060), $FF=unknown FPU)
mmutype	                DC.B	$00                         ; $000004EA MMU type ($00=no MMU, $01=68841, $02=68851, $03=CPU integrated(68030/68040/68060), $FF=unknown MMU)
escapeflag              DC.B	$00                         ; $000004EB Escape flag (bit 6=1 if ESCAPE condition)

; -----------------------------------------------------------------------------
; OS workspace: OS variables
; -----------------------------------------------------------------------------
	ORG     $00000500
standarddatetimestring	DC.B	'%24:%mi:%se %dy-%m3-%ce%yr', $00	; SYS$DateFormat


; -----------------------------------------------------------------------------
; Reset entry
; -----------------------------------------------------------------------------
	ORG	bios_address+$00000800            ; RAM at $00000000 is enabled when A19 is asserted (The 68030 issue 1 board uses A31 instead of A19)

reset_entry
	ORI.W	#$0700, SR                        ; Disable all interrupts
	MOVE.B	r2data, D0                        ; Enable RAM
	MOVE.L	#$00000600, SP                    ; Temporary set the stackpointer so we can safely jump to subroutines

	MOVE.L	#$00000000, A0                    ; Destination address
	MOVE.L	#vector_table, A1                 ; Source address
	MOVE.W	#$3F, D0                          ; Number of system vectors minus one
init_sysvectors
	MOVE.L	(A1)+, (A0)+                      ; Move data
	DBF	D0, init_sysvectors
	                                          ; Set all user vectors to 'Unknown Exception' handler
	MOVE.W	#$BF, D0                          ; Number of user vectors minus one
init_uservectors
	MOVE.L	#unknown_exception, (A0)+         ; Move data
	DBF	D0, init_uservectors

	MOVE.W	#$3F, D0                          ; Number of words minus one
	MOVE.L	#$00000400,  A0                   ; Destination address
	MOVE.L	#userv,  A1                       ; Source address
init_osworkspace
	MOVE.L	(A1)+, (A0)+                      ; Move data
	DBF	D0, init_osworkspace

	MOVE.W	$0000, (hoaddr)                   ; Set high order pointer to $0000 (used by OS_Byte &82 and OS_Word &00)
	BSR	checkramsize                      ; Check how much RAM this system has
	MOVE.L	D0, ramsize                       ; ...and store the hex value
	MOVE.L	D0, SP                            ; Set supervisor stack pointer to top of RAM

	SUBI.L	#stacksize, D0                    ; Reserve supervisor stack
	MOVE.L	D0, himem                         ; Set himem to top of RAM minus supervisor stack

	BSR	check_cpu                         ; Find out which CPU is present
	MOVE.B	D0, cputype                       ; ...and store it
	BSR	check_fpu                         ; Find out which FPU is present
	MOVE.B	D0, fputype                       ; ...and store it
	BSR	check_mmu                         ; Find out which MMU is present
	MOVE.B	D0, mmutype                       ; ...and store it

	BSR	SWI_OS_IntOn                      ; Enable interrupts
	ANDI.W	#$DFFF, SR
;	BSR	SWI_OS_LeaveOS                    ; Switch to user mode

	MOVE.L	himem, D0                         ; Get top of RAM
	MOVE.L	D0, SP                            ; Set user stack pointer to top of RAM minus supervisor stack
	SUBI.L	#stacksize, D0                    ; Reserve user stack
	MOVE.L	D0, himem                         ; Set himem to top of RAM minus supervisor and user stack
	MOVE.L	#$00000800, oshwm
	
	MOVE.L	#startup_message1, D0             ; Location of startup message in D0
	BSR	SWI_OS_Write0                     ; Print the message

	MOVE.B	cputype, D0
	CMP	#$09, D0
	BLO	init_startup_msg
	MOVEQ	#$00, D0                          ; Unknown CPU, use CPU type 0

init_startup_msg
	CMP.B	#$FF, (cputype)
	BEQ	init_startup_msg_unknown_cpu

	MOVE.L	#startup_message_cpus, D0         ; A0 = A0 + (3*D0)
	ADD.B	(cputype), D0
	ADD.B	(cputype), D0
	ADD.B	(cputype), D0
;	ADD.L	D0, A0                            ; A0 = A0 + (3*D0)
;	ADD.L	D0, A0
;	ADD.L	D0, A0                            ; Point to CPU name
;	MOVE.L	A0, D0
	MOVE.L	#$3, D1                           ; Write 3 characters
	BSR	SWI_OS_WriteN
	BRA	init_startup_msg2

init_startup_msg_unknown_cpu
	MOVE.B	#'K', D0
	BSR	SWI_OS_WriteC

init_startup_msg2
	MOVE.L	#startup_message2, D0             ; Location of message in D0
	BSR	SWI_OS_Write0                     ; Print the message

	MOVE.L	ramsize, D0                       ; Get RAM size (use OS_ConvertFileSize???)
	LSR.L	#5, D0                            ; Divide by 1024
	LSR.L	#5, D0 
	MOVE.L	#buffer, D1                       ; Pointer to buffer
	MOVE.L	#$FF, D2                          ; Size of buffer
	BSR	SWI_OS_ConvertCardinal4           ; Convert value to string
	MOVE.L	#buffer, D0                       ; Pointer to buffer
	BSR	SWI_OS_Write0                     ; And print the RAM size in K

	MOVE.L	#startup_message3, D0             ; Print 'K' <cr> <cr>
	BSR	SWI_OS_Write0                     ; Print the message

	MOVE.L	#$00000001, D0
	MOVE.L	#osstarttime, D1
	BSR	SWI_OS_Word                       ; Get OS start time
	MOVE.B	#$00, programstarttime            ; Set program start time to 0
	MOVE.B	#$00, programstarttime+1
	MOVE.B	#$00, programstarttime+2
	MOVE.B	#$00, programstarttime+3
	MOVE.B	#$00, programstarttime+4

	BSR	readbyte_from_tuber2              ; 
	CMP.B	#$80, D0                          ; Check if we need to enter a program at startup
	BEQ	SWI_OS_Byte_8E                    ; Enter language ROM

main_loop
	MOVE.B	#prompt, D0                       ; * Prompt
	BSR	SWI_OS_WriteC

	MOVE.L	#buffer, D0                       ; Buffer address
	MOVE.L	#$FF, D1                          ; Maximum line length is 255
	MOVE.B	#$20, D2                          ; Minimum acceptable ASCII value
	MOVE.B	#$FF, D3                          ; Maximum acceptable ASCII value
	MOVE.L	#OS_ReadLine32, A0                ; SWI OS_ReadLine32: Read line from input
	TRAP	#SWI
	BCS	escape                            ; Jump if user pressed ESCAPE

	MOVE.L	#buffer, D0                       ; D0 points to the line read by the OS_ReadLine32 call
	MOVE.L	#OS_CLI, A0                       ; SWI OS_CLI: Process command
	TRAP	#SWI
	BRA	main_loop                         ; Loop infinitely

escape
	MOVEQ	#$7E, D0                          ; OSBYTE $7E: Acknowledge detection of an ESCAPE condition
	BSR	SWI_OS_Byte
	MOVE.L	#err_escape, D0                   ; Point D0 to the ESCAPE error message
	MOVE.L	#OS_GenerateError, A0             ; SWI OS_GenerateError
	TRAP	#SWI                              ; SWI call
	BRA	main_loop                         ; This command should never be reached



; -----------------------------------------------------------------------------
; Low level hardware stuff
;
; Wait for data in Tube Data 1 and return it in D0, while allowing requests via Tube R4
; -----------------------------------------------------------------------------
readbyte_from_tuber1
	BTST	#7, r1status                      ; Check Tube Status 1 for data present in Tube Data Register 1
	BNE	readbyte_from_tuber1_fetch        ; Data present, fetch it
	BTST	#7, r4status                      ; Check Tube Status 1 for data present in Tube Data Register 1
	BEQ	readbyte_from_tuber1              ; Nothing in Tube R4, loop back
	BSR	irq_transfer                      ; Deal with Tube R4 request
	BRA	readbyte_from_tuber1              ; Loop back to wait for Tube R1 data

readbyte_from_tuber1_fetch
	MOVE.B	r1data, D0                        ; Fetch byte from Tube Register 1

	IFNE DEBUG
		MOVE.B  #5, D0
		TRAP    #15
		MOVE.B  D1, D0
	ENDC

	RTS

; -----------------------------------------------------------------------------
; Wait for data in TUBE ULA Data register 2 and return it in D0
; -----------------------------------------------------------------------------
readbyte_from_tuber2
	BTST	#7, r2status                      ; Check the DA2 flag of TUBE ULA Status register 2 (Data present in 2st register)
	BEQ	readbyte_from_tuber2              ; No? Then wait
	MOVE.B	r2data, D0                        ; D0.B = ?r2data
	RTS

; -----------------------------------------------------------------------------
; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2
; -----------------------------------------------------------------------------
writebyte_to_tuber2
	BTST	#6, r2status                      ; Check the NF2 flag of TUBE ULA Status register 2 (2nd data register not filled)
	BEQ	writebyte_to_tuber2               ; Loop until TUBE ULA Data register 2 is empty
	MOVE.B	D0, r2data                        ; Write byte to r2data
	RTS

; -----------------------------------------------------------------------------
; Wait for data in TUBE ULA Data register 4 and return it in D0
; -----------------------------------------------------------------------------
readbyte_from_tuber4
	BTST	#7, r4status                      ; Check Tube Status 4 for data present in Tube Data Register 4
	BEQ	readbyte_from_tuber4              ; Loop until data present
	MOVE.B	r4data, D0                        ; Fetch byte from Tube Register 4
	RTS

; -----------------------------------------------------------------------------
; Read 32-bit word from TUBE via Data register 2 to D0
; -----------------------------------------------------------------------------
readlong_from_tuber2
	CLR.L	D0
	BSR	readbyte_from_tuber2              ; D0=&000000aa
	ROL	#8, D0                            ; D0=&0000aa00
	BSR	readbyte_from_tuber2              ; D0=&0000aabb
	ROL	#8, D0                            ; D0=&00aabb00
	BSR	readbyte_from_tuber2              ; D0=&00aabbcc
	ROL	#8, D0                            ; D0=&aabbcc00
	BRA	readbyte_from_tuber2              ; Fall through to get D0=&aabbccdd

; -----------------------------------------------------------------------------
; Send 32-bit word in D0 to TUBE via Data register 2
; -----------------------------------------------------------------------------
writelong_to_tuber2
	ROL	#8, D0
	BSR	writebyte_to_tuber2               ; Send b24-b31
	ROL	#8, D0
	BSR	writebyte_to_tuber2               ; Send b16-b23
	ROL	#8, D0
	BSR	writebyte_to_tuber2               ; Send b8-b15
	ROL	#8, D0
	BRA	writebyte_to_tuber2               ; Drop through to send b0-b7

; -----------------------------------------------------------------------------
; Read 32-bit word from TUBE via Data register 2 to D4
; -----------------------------------------------------------------------------
readlong_from_tuber4
	CLR.L	D0
	BSR	readbyte_from_tuber4              ; D0=&000000aa
	ROL	#8, D0                            ; D0=&0000aa00
	BSR	readbyte_from_tuber4              ; D0=&0000aabb
	ROL	#8, D0                            ; D0=&00aabb00
	BSR	readbyte_from_tuber4              ; D0=&00aabbcc
	ROL	#8, D0                            ; D0=&aabbcc00
	BRA	readbyte_from_tuber4              ; Fall through to get D0=&aabbccdd

; -----------------------------------------------------------------------------
; Send string at (A6) through TUBE ULA Data register 2
; -----------------------------------------------------------------------------
writestring_to_tuber2
	BTST	#6, r2status                      ; Test Tube Register 2 Status
	BEQ	writestring_to_tuber2             ; Loop until Tube Register 2 is empty
	MOVE.B	(A6)+, D0                         ; Get byte in D0
	BSR	writebyte_to_tuber2
	CMP.B	#$0D, D0                          ; Is it the last character in the string?
	BNE	writestring_to_tuber2             ; No? Then process next character
	RTS

; -----------------------------------------------------------------------------
; IRQ handler
; -----------------------------------------------------------------------------
irq_handler
	BTST	#7, r4status                      ; Check Tube Status 4 to see if any data in Tube Register 4
	BNE	irq_transfer                      ; Data in Tube R4, jump to process errors and transfers
	BTST	#7, r1status                      ; Check the DA2 flag of TUBE ULA Status register 2 (Data present in 2st register)
	BNE	irq_events                        ; Data in Tube R1, jump to process Escape and Events

	MOVE.L	A6, -(SP)                         ; Save A6
	MOVE.L	(irqv-bios_address).L, A6                      ; IRQ not recognised, pass on to IRQV
	JSR	(A6)                              ; Jump to IRQV
	MOVE.L	(SP)+, A6                         ; Restore A6

	RTE

; -----------------------------------------------------------------------------
; Interupt generated by data in Tube Register 1 - Escape change or Event
; -----------------------------------------------------------------------------
irq_events
	MOVE.L	D0, -(SP)                         ; Save D0
	MOVE.B	r1data, D0                        ; Get data byte from Tube Register 1
	BMI	irq_escape                        ; b7=1, change Escape state

	MOVE.L	D1, -(SP)                         ; Save D1
	
	CLR.L	D0
	BSR	readbyte_from_tuber1              ; Get Event parameter Y
	ROL	#8, D0                            ; D0=&0000aa00
	BSR	readbyte_from_tuber1              ; Get Event parameter X
	MOVE.W	D0, D1                            ; Get 16-bit event parameter in D1

	BSR	readbyte_from_tuber1              ; Get Event number

	MOVE.L	A6, -(SP)                         ; Save A6
	MOVE.L	(eventv-bios_address).L, A6       ; IRQ not recognised, pass on to EventV
	JSR	(A6)                              ; Jump to EventV
	MOVE.L	(SP)+, A6                         ; Restore A6

	MOVE.L	(SP)+, D1                         ; Restore registers
	MOVE.L	(SP)+, D0
	RTE                                       ; Return from interupt

; -----------------------------------------------------------------------------
; Escape state changed
; -----------------------------------------------------------------------------
irq_escape
	ADD	D0, D0                            ; Move Escape state to bit 7 (????? should already be in bit 7)
	JSR	_escapev                          ; Call the Escape handler (not vectored since there''s no EscV in RiscOS; find out if I need to change this)
	MOVE.L	(SP)+, D0
	RTE

; -----------------------------------------------------------------------------
; Interrupt generated by data in Tube Register 4 - Initiate data transfer
; -----------------------------------------------------------------------------
irq_transfer
	MOVE.B	r4data, D0                        ; Get data byte from Tube Register 4
	BPL	irq_data_transfer                 ; b7=0, jump for data transfer

; -----------------------------------------------------------------------------
; Data transfer type b7=1 - Host error
; -----------------------------------------------------------------------------
	MOVE.L	D0, -(SP)                         ; Save D0
	MOVE.L	A0, -(SP)                         ; Save A0
	MOVE.L	#buffer, A0                       ; Point A0 to error buffer
	BSR	readbyte_from_tuber2              ; Wait for BRK byte

	MOVEQ	#0, D0                            ; Only use lower 8 bits for error number
	BSR	readbyte_from_tuber2              ; Wait for error number
	MOVE.L	D0, (A0)+                         ; Store error number

irq_error_lp
	BSR	readbyte_from_tuber2              ; Wait for error string character
	MOVE.B	D0, (A0)+                         ; Store in error buffer
;	CMP.B	#0, D0
	BNE	irq_error_lp                      ; Loop until final &00 byte

	MOVE.L	(SP)+, A0                         ; Restore A0
	MOVE.L	#buffer, D0                       ; Point D0 to error buffer
	MOVE.L	D0, error
	MOVE.L	#$FFFFC000, error_pc              ; Location of error is somewhere in the host
	BSR	SWI_OS_GenerateError              ; Generate error
	MOVE.L	(SP)+, D0                         ; Restore D0
	RTE                                       ; 

; -----------------------------------------------------------------------------
; Data transfer type b7=0 - Data transfer
; D0 = action
; Tube data via R4: ID address[4] sync, then via R3: data
; -----------------------------------------------------------------------------
irq_data_transfer
	MOVE.L	A0, -(SP)                         ; Save A0
	MOVE.L	D0, -(SP)                         ; Save D0
	ANDI.L	#%11111111, D0                    ; Ensure 8-bit value of reason code in D0
	LSL.L	#2, D0                            ; D0 = D0 * 4
	LEA	nmi_handler_table, A0             ; Pointer to NMI routine in A0
	ADDA.L	D0, A0                            ; Add offset
	MOVE.L	A0, $00000074                     ; Store in NMI vector (INT5)

	BSR	readbyte_from_tuber4              ; Wait for ID byte
	MOVE.L	(SP), D0                          ; Restore D0 but leave it on the stack for now

	CMP.B	#5, D0                            ; Was transfer type 'Tube Release'?
	BEQ	irq_data_transfer_exit            ; TubeRelease - exit    

	BSR	readlong_from_tuber4              ; Read word from host
	MOVE.L	D0, tubetransfer                  ; Store start address of data transfer

	MOVE.B	r3data, D0                        ; Read from Tube R3 twice to clear FIFO
	MOVE.B	r3data, D0
	BSR	readbyte_from_tuber4              ; Wait for sync byte and discard it

	MOVE.L	(SP), D0                          ; Restore D0, but leave it on the stack for now
	CMP.B	#$06, D0
	BLO	irq_data_transfer_exit            ; Exit if it`s reason code &00-&05
	BNE	irq_data_transfer_07              ; Jump if it`s not reason code &06

	MOVEQ	#$FF, D0                          ; Reason code &06 handler, transfer 256 bytes from co-pro to host

irq_data_transfer_06_loop
	BTST	#7, r3status                      ; Check for data in TUBE register 3
	BEQ	irq_data_transfer_06_loop         ; If not, then loop
	MOVE.B	(A2)+, r3data                     ; Send byte to TUBE register 3
	DBF	D0, irq_data_transfer_06_loop

	BTST	#7, r3status                      ; Check for data in TUBE register 3
	BEQ	irq_data_transfer_06_loop         ; If not, then loop
	MOVE.B	-(A2), r3data                     ; Send byte to TUBE register 3

irq_data_transfer_exit                            ; Reason code &00-&05 handler
	MOVE.L	(SP)+, D0                         ; Restore D0
	MOVE.L	(SP)+, A0                         ; Restore A0
	RTE

irq_data_transfer_07                              ; Reason code &07 handler, transter 256 bytes from host to co-pro
	MOVEQ	#$FF, D0

irq_data_transfer_07_loop
	BTST	#7, r3status                      ; Check for data in TUBE register 3
	BEQ	irq_data_transfer_07_loop         ; If not, then loop
	MOVE.B	r3data, (A2)+                     ; Get byte from TUBE register 3
	DBF	D0, irq_data_transfer_07_loop     ; Loop 256 times
	BRA	irq_data_transfer_exit

; -----------------------------------------------------------------------------
; NMI handler reason code &00 (Multiple single byte transfer from parasite to host)
; -----------------------------------------------------------------------------
nmi_data_transfer_00
	MOVE.L	A2, -(SP)                         ; Save A2
	MOVE.L	tubetransfer, A2
	MOVE.B	(A2)+, r2data
	MOVE.L	A2, tubetransfer
	MOVE.L	(SP)+, A2
	RTE

; -----------------------------------------------------------------------------
; NMI handler reason code &01 (Multiple single byte transfer from host to parasite)
; -----------------------------------------------------------------------------
nmi_data_transfer_01
	MOVE.L	A2, -(SP)                         ; Save A2
	MOVE.L	tubetransfer, A2
	MOVE.B	r2data, (A2)+
	MOVE.L	A2, tubetransfer
	MOVE.L	(SP)+, A2
	RTE

; -----------------------------------------------------------------------------
; NMI handler reason code &02 (Multiple double byte transfer from parasite to host)
; -----------------------------------------------------------------------------
nmi_data_transfer_02
	MOVE.L	A2, -(SP)                         ; Save A2
	MOVE.L	tubetransfer, A2
	MOVE.B	(A2)+, r2data                     ; Timing???
	MOVE.B	(A2)+, r2data
	MOVE.L	A2, tubetransfer
	MOVE.L	(SP)+, A2
	RTE

; -----------------------------------------------------------------------------
; NMI handler reason code &03 (Multiple double byte transfer from host to parasite)
; -----------------------------------------------------------------------------
nmi_data_transfer_03
	MOVE.L	A2, -(SP)                         ; Save A2
	MOVE.L	tubetransfer, A2
	MOVE.B	r2data, (A2)+                     ; Timing???
	MOVE.B	r2data, (A2)+
	MOVE.L	A2, tubetransfer
	MOVE.L	(SP)+, A2
	RTE

; -----------------------------------------------------------------------------
; NMI handler reason codes &04-&07
; -----------------------------------------------------------------------------
nmi_handler
	MOVE.B	D0, r3data
	RTE

nmi_handler_table
	DC.L	nmi_data_transfer_00              ; Multiple single byte transfer from parasite to host
	DC.L	nmi_data_transfer_01              ; Multiple single byte transfer from host to parasite
	DC.L	nmi_data_transfer_02              ; Multiple double byte transfer from parasite to host
	DC.L	nmi_data_transfer_03              ; Multiple double byte transfer from host to parasite
	DC.L	nmi_handler                       ; Execute code in parasite
	DC.L	nmi_handler                       ; Release TUBE
	DC.L	nmi_handler                       ; Transfer 256 bytes from co-pro to host
	DC.L	nmi_handler                       ; Transfer 256 bytes from host to co-pro



; -----------------------------------------------------------------------------
; 68k CPU exception handlers
;
; Bus error exception handler
; -----------------------------------------------------------------------------
bus_error_handler
	MOVE.L	#msg_bus_error, D0                 ; Exception message
	BSR	SWI_OS_Write0                      ; Not using OS_GenerateError since there must be something seriously wrong
	BRA	address_error_handler2

; -----------------------------------------------------------------------------
; Address error exception handler
; -----------------------------------------------------------------------------
address_error_handler
	BSR	SWI_OS_NewLine
	MOVE.L	#msg_address_error, D0            ; Bus error message
	BSR	SWI_OS_Write0                     ; Not using OS_GenerateError since there must be something seriously wrong

address_error_handler2
	MOVE.L	#msg_address_error_mf, D0         ; Bus error message
	BSR	SWI_OS_Write0
	MOVE.L	#buffer, D1                       ; D1 is pointer to buffer
	MOVE.B	#$FF, D2                          ; Size of buffer
	MOVEQ	#$00, D0
	MOVE.W	(SP)+, D0                         ; Pop "Memory access type and function code" word
	BSR	SWI_OS_ConvertHex4                ; Prepare string
	BSR	SWI_OS_Write0                     ; Write string
	BSR	SWI_OS_NewLine

	MOVE.L	#msg_address_error_aa, D0         ; Bus error message
	BSR	SWI_OS_Write0
	MOVE.L	#buffer, D1                       ; D1 is pointer to buffer
	MOVE.B	#$FF, D2                          ; Size of buffer
	MOVEQ	#$00, D0
	MOVE.L	(SP)+, D0                         ; Pop "Access address" long
	BSR	SWI_OS_ConvertHex8
	BSR	SWI_OS_Write0                     ; Write string
	BSR	SWI_OS_NewLine

	MOVE.L	#msg_address_error_ir, D0         ; Bus error message
	BSR	SWI_OS_Write0
	MOVE.L	#buffer, D1                       ; D1 is pointer to buffer
	MOVE.B	#$FF, D2                          ; Size of buffer
	MOVEQ	#$00, D0
	MOVE.W	(SP)+, D0                         ; Pop "Instruction register" word
	BSR	SWI_OS_ConvertHex4
	BSR	SWI_OS_Write0                     ; Write string
	BSR	SWI_OS_NewLine

	MOVE.L	#msg_address_error_sr, D0         ; Bus error message
	BSR	SWI_OS_Write0
	MOVE.L	#buffer, D1                       ; D1 is pointer to buffer
	MOVE.B	#$FF, D2                          ; Size of buffer
	MOVEQ	#$00, D0
	MOVE.W	(SP)+, D0                         ; Pop "Status register" word
	BSR	SWI_OS_ConvertBinary2
	BSR	SWI_OS_Write0                     ; Write string
	BSR	SWI_OS_NewLine

	MOVE.L	#msg_address_error_pc, D0         ; Bus error message
	BSR	SWI_OS_Write0
	MOVE.L	#buffer, D1                       ; D1 is pointer to buffer
	MOVE.B	#$FF, D2                          ; Size of buffer
	MOVEQ	#$00, D0
	MOVE.L	(SP), D0                          ; Pop "Program counter" long
	BSR	SWI_OS_ConvertHex8
	BSR	SWI_OS_Write0                     ; Write string
	BSR	SWI_OS_NewLine

exception0handler2
	BRA	exception0handler2
	BSR	SWI_OS_ReadC
	BRA	main_loop                         ; We cannot return from an access error exception, so jump straight back to the main loop

; -----------------------------------------------------------------------------
; Illegal Instruction handler
; -----------------------------------------------------------------------------
illegal_instruction_handler
	MOVE.L	D0, -(SP)                         ; Push D0 to the stack
	MOVE.L	#err_illegal_instruction, D0      ; Location of message in A6
	BRA	exception_handler

; -----------------------------------------------------------------------------
; Integer divide by zero handler
; -----------------------------------------------------------------------------
int_divide_by_zero_handler
	MOVE.L	D0, -(SP)                         ; Push D0 to the stack
	MOVE.L	#err_int_divide_by_zero, D0       ; Location of message in A6
	BRA	exception_handler

; -----------------------------------------------------------------------------
; Privilege violation handler
; -----------------------------------------------------------------------------
privilege_violation_handler
	MOVE.L	D0, -(SP)                         ; Push D0 to the stack
	MOVE.L	#err_privilege_violation, D0      ; Location of message in A6
	BRA	exception_handler

; -----------------------------------------------------------------------------
; Unknown exception handler
; -----------------------------------------------------------------------------
unknown_exception
	MOVE.L	D0, -(SP)                         ; Push D0 to the stack
	MOVE.L	#err_unknown_exception, D0        ; Location of message in A6

exception_handler
	MOVE.L	6(SP), error_pc                   ; Pull program counter from the stack
	BSR	SWI_OS_GenerateError              ; Print the message
	MOVE.L	(SP)+, D0                         ; Pull D0 from the stack
	RTE

; -----------------------------------------------------------------------------
; Handler for not supported OS calls
; -----------------------------------------------------------------------------
not_supported
	MOVE.L	#err_not_supported, D0
	MOVE.L	#OS_GenerateError, A0
	TRAP	#12
	RTS

; -----------------------------------------------------------------------------
; Check CPU type ($00=68008, $01=68000, $02=68010, $03=68020, $04=68030, $05=68040, $06=68060, $07=68070, $08=68300, $FF=unknown CPU)
; 68010 got new registers: RFC, DFC and VBR
; 68020 got new registers: CACR and CAAR
; -----------------------------------------------------------------------------

check_cpu                                         ; this needs to be run at reset or else a way to get the vector base address needs to be known
	MOVEQ	#$00,D0                           ; exception vector base, assumes 68000 here
	MOVEA.L	D0, A5                            ; point to vector base
	MOVEA.L	$0C(A5), A0                       ; save the address exception vector
	LEA	check_cpu2(PC), A1                ; get the continue address
	MOVE.L	A1, $0C(A5)                       ; save as the address exception vector
	MOVE.L	SP, D7                            ; copy the stack pointer
	MOVE.L	D7, D6                            ; copy the stack pointer again
	JMP	*+5(PC)                           ; cause an address error - all processors

	                                          ; now there is an address error exception frame on the stack the processor type can be
	                                          ; deduced from the size of this frame

check_cpu2
	MOVE.L	A0, $0C(A5)                       ; restore the address exception vector
	SUB.L	SP, D7                            ; calculate the size of the stack frame

	MOVEQ	#$01, D0                          ; default to 68000 processor
	CMPI.B	#$12, D7                          ; compare frame size with that of 68000
	BEQ.S	proc_found                        ; if 68000 size then exit

	MOVEQ	#$02, D0                          ; set 68010 processor
	CMPI.B	#$3E, D7                          ; compare frame size with that of 68010
	BEQ.S	proc_found                        ; if 68010 size then exit

	MOVEQ	#$07, D0                          ; set 68070 processor
	CMPI.B	#$26, D7                          ; compare frame size with that of 68070
	BEQ.S	proc_found                        ; if 68070 size then exit

	MOVE.L	#$08, D0                          ; set 68300 processor
	CMPI.B	#$1C, D7                          ; compare frame size with that of 68300
	BEQ.S	proc_found                        ; if 68300 size then exit

	MOVEQ	#$05, D0                          ; set 68040 processor
	CMPI.B	#$10, D7                          ; compare frame size with that of 68040
	BEQ.S	proc_found                        ; if 68040 size then exit

	MOVEQ	#$FF, D0                          ; set unknown processor
	CMPI.B	#$24, D7                          ; compare frame size with that of 68020/030
	BNE.S	proc_found                        ; if not 68020/030 size then exit

	                                          ; test for 68020/68030

	MOVE.L	$1A(A5), A0                       ; get illegal instruction exception vector
	MOVE.L	$7E(a5), A1                       ; get format error exception vector

	LEA	is_68020(pc), A2                  ; set our format error exception vector
	MOVE.L	A2, $7E(A5)                       ; save format error exception vector

	LEA	is_68030(pc), A2                  ; set our illegal instruction exception vector
	MOVE.L	A2, $1A(A5)                       ; save illegal instruction exception vector
	MOVEQ	#$03, D0                          ; default to 68020 processor

	DC.W	$06FA, $0000, $0002               ; CALLM #0, next_word(PC)
	                                          ; call module, 68020 only
next_word
	DC.W	-1                                ; dummy data, cause format error

is_68030                                          ; illegal instruction exception arrives here
	MOVEQ	#$04, D0                          ; set 68030 processor

is_68020                                          ; format error exception arrives here
	MOVE.L	A1, $7E(A5)                       ; restore format error exception vector
	MOVE.L	A0, $1A(A5)                       ; restore illegal instruction exception vector

proc_found
	MOVE.L	D6, SP                            ; restore the stack pointer
	RTS

; -----------------------------------------------------------------------------
; Check FPU type ($00=no FPU, $01=68881, $02=68882, $03=CPU integrated(68040/68060), $FF=unknown FPU)
; -----------------------------------------------------------------------------

check_fpu
	MOVEQ	#0, D0                            ; exception vector base, assumes 68000 here
	MOVEA.L	D0, A5                            ; point to vector base
	MOVEA.L	$2C(A5), A0                       ; save the address exception vector
	LEA	no_fpu(pc), A1                    ; get the continue address
	MOVE.L	A1, $2C(A5)                       ; save as the address exception vector
	MOVE.L	SP, D7                            ; copy the stack pointer

;	FMOVE.S	#0, FP0                           ; FPU instruction; if no FPU is available, an exception $10 will be generated

	MOVEQ	#$FF, D0                          ; Set FPU flag
	BRA	check_fpu2

no_fpu
	MOVEQ	#$00, D0                          ; Set FPU flag

check_fpu2
	MOVE.L	D7, SP                            ; restore the stack pointer
	MOVE.L	A0, $2C(A5)                       ; save as the address exception vector
	RTS

; -----------------------------------------------------------------------------
; MMU type ($00=no MMU, $01=68841, $02=68851, $03=CPU integrated(68030/68040/68060), $FF=unknown MMU)
; -----------------------------------------------------------------------------
check_mmu
	MOVEQ	#$00, D0
	RTS

; -----------------------------------------------------------------------------
; Resets the flash ROM to data mode
; -----------------------------------------------------------------------------
flash_reset
	MOVE.B	#$F0, bios_address
	MOVE.B	#$FF, bios_address
	RTS	

; -----------------------------------------------------------------------------
; Returns the flash ROMs vendor and device ID in D0
; -----------------------------------------------------------------------------
flash_identify
	MOVE.B	#$AA, bios_address+$5555          ; Switch flash ROM to command mode
	MOVE.B	#$55, bios_address+$AAAA
	MOVE.B	#$90, bios_address+$5555          ; Read identification command
	MOVE.B	D0, bios_address                  ; Get vendor ID
	ASL.W	#8, D0                            ; Shift D0 to make room for device ID
	BSR	flash_reset

	MOVE.B	#$AA, bios_address+$5555          ; Switch flash ROM to command mode
	MOVE.B	#$55, bios_address+$AAAA
	MOVE.B	#$90, bios_address+$5555          ; Read identification command
	MOVE.B	D0, bios_address+1                ; Get device ID
	BSR	flash_reset

	RTS

; -----------------------------------------------------------------------------
; Writes one sector (65536 bytes) in the flash ROM (D7=sector, A6=memory address)
; -----------------------------------------------------------------------------
flash_write_sector
	CMP.B	#flash_rom_sectors, D7            ; Check if sector is < 8
	BHS	flash_write_sector99              ; If not, exit
	ASL	#8, D7
	ASL	#8, D7

	MOVE.L	#$FFFF, D1                        ; 65536-1 bytes

flash_write_sector1
	MOVE.B	(A5), D0                          ; Get byte from memory
	BSR	flash_write_byte
;;;;;;
;	DBF	D2, shift_wo                      ; Do next nibble

flash_write_sector99
	RTS

; -----------------------------------------------------------------------------
; Writes one byte in the flash ROM
; -----------------------------------------------------------------------------
flash_write_byte
	CMP.B	(A6), D0                          ; Check if the address already contains this byte
	BEQ	flash_write_byte2                 ; Yes? Then exit

	MOVE.B	#$AA, bios_address+$5555          ; Switch flash ROM to command mode
	MOVE.B	#$55, bios_address+$AAAA
	MOVE.B	#$A0, bios_address+$5555          ; Read identification command
	MOVE.B	(A6), D0                          ; Write byte

flash_write_byte1
	CMP.B	#2, (A6)
	BNE	flash_write_byte1

	BSR	flash_reset

flash_write_byte2
	RTS

; -----------------------------------------------------------------------------
; Find out how much RAM is available
; On exit: D0 = total number of bytes in hex
; -----------------------------------------------------------------------------
checkramsize
	MOVE.L	A0, -(SP)                         ; Push A0 to the stack
	MOVE.L	$00000008, -(SP)                  ; Save original contents of $00000008 on the stack

	MOVE.L	#$00000008, A0                    ; Reset hex counter
	MOVE.L	#$DEADBEEF, $00000008             ; Set $00000008 to the magic number

checkramsize_loop
	ADDA.L	#$00000400, A0                    ; Increase hex counter with 1Kbyte
	CMPI.L	#$DEADBEEF, (A0)                  ; Is it the same as location $00000008?
	BEQ	checkramsize_exit                 ; Yes, then exit
	CMP.L	#max_ram_size, A0                 ; Prevent indefenitely looping
	BLO	checkramsize_loop

checkramsize_exit
	MOVE.L	A0, D0
	SUBQ	#8, D0
	MOVE.L	(SP)+, $00000000                  ; Pull original contents of $00000000 from the stack
	MOVE.L	(SP)+, A0                         ; Pull A0 from the stack
	RTS



; -----------------------------------------------------------------------------
; SWI call support
; On entry: A0=SWI number, D0-D7 = R0-R7, A1-A2=R8-R9
;
; SWI handler (TRAP #12)
; -----------------------------------------------------------------------------
SWI_Handler
	MOVE.L	A2, -(SP)                         ; Push A2 to the stack
	MOVE.L	A1, -(SP)                         ; Push A1 to the stack
	MOVE.L	A0, -(SP)                         ; Push A0 to the stack
	MOVE.L	D1, -(SP)                         ; Push D1 to the stack
	EXG	A0, D0                            ; Swap A0 (SWI number) and D0
	ANDI.L	#$00FDFFFF, D0                    ; Remove 'X' bit and superflous bits

	LEA	SWITable, A1                      ; Point A1 to SWI lookup table

SWI_Handler_loop
	CMP.L	(A1)+, D0                         ; Check SWI number
	BEQ	SWI_Handler_dispatch              ; Yes, this is our SWI handler

SWI_Handler_loop2
	MOVE.L	(A1)+, D1                         ; Get next word
	CMP.B	#$FF, D1                          ; End of SWI-table?
	BEQ	SWI_Handler_WriteI                ; Yes, so pass this SWI call to the _UKSWIV
	CMP.B	#$00, D1                          ; Check for end of string marker
	BNE	SWI_Handler_loop2                 ; This loop can probably be optimized, but it works for now
	BRA	SWI_Handler_loop

SWI_Handler_dispatch
	MOVE.L	(A1)+, A2                         ; Get address of SWI function (this needs to be word-aligned; if not it will generate an address exception)
	EXG	A0, D0                            ; Swap A0 (SWI number) and D0
	CMP.L	#oshwm, A2                        ; Check if address in SWI jump table is a vector
	BHS	SWI_Handler_dispatch2             ; If not, jump
	MOVE.L	(A2), A2

SWI_Handler_dispatch2
	MOVE.L	(SP)+, D1                         ; Pull D1 from the stack
	MOVE.L	(SP)+, A0                         ; Pull A0 from the stack
	MOVE.L	(SP)+, A1                         ; Pull A1 from the stack
	JSR	(A2)                              ; Jump to SWI function

SWI_Handler_exit
	BVS	SWI_ErrorReturn                   ; V set, return or generate error
	MOVE.L	(SP)+, A2                         ; Pull A2 from the stack
	BCS	SWI_ReturnCarrySet                ; ????? What is this for?
	ANDI	#$FC, CCR                         ; Clear Carry and Overflow in stacked PSW                     ; Don't know if this will work since we have two SP's: SSP and USP (Supervisor and User Stack Pointer)
	RTE                                       ; Return to User world

SWI_ReturnCarrySet
	ANDI	#$FD, CCR                         ; Clear Overflow in stacked PSW
	ORI.W	#1, (SP)                          ; Set Carry in stacked PSW
	RTE                                       ; Return to User world

SWI_ErrorReturn
	MOVE.L	(SP)+, A2                         ; Restore A2
	EXG	A0, D0                            ; Put the raw SWI number in D0 so we can test for the 'X' bit
	BTST	#17, D0                           ; Was 'X' bit set in raw SWI number?
	BEQ	SWI_ErrorGenerate                 ; 'X' was clear, generate an error
	EXG	A0, D0
	ORI.W	#$02, (SP)                        ; Set V flag in stacked PSW
	RTE                                       ; Return to User world (D0 points to error block and V is set)

SWI_ErrorGenerate
	EXG	A0, D0
	MOVE.L	D1, -(SP)                         ; Push D1 to the stack
	MOVE.L	#$00000006, D1                    ; Service call: Service_Error
	BSR	SWI_OS_ServiceCall                ; Issue an OS_ServiceCall (D0 still points to the error block)

	MOVE.L	D0, (error)                       ; Store current error address
	MOVE.L	2(SP), (error_pc)
	MOVE.L	(errorv-bios_address).L, 2(SP)                 ; Overwrite return address on stack with ErrorV
	RTE                                       ; Enter ErrorV handler in User world

SWI_Handler_WriteI
	MOVE.L	(SP), D0                          ; Pull D0 from the stack (SWI number)
	AND.L	#$00FDFF00, D0
	CMP.L	#OS_WriteI, D0
	BNE	SWI_Handler_unknown
	EXG	A0, D0                            ; Restore original value of D0
	MOVE.L	(SP)+, D0                         ; Pull D0 from the stack (SWI number)
	JSR	SWI_OS_WriteI
	BRA	SWI_Handler_exit

SWI_Handler_unknown
	EXG	A0, D0                            ; Restore original value of A0
	MOVE.L	(SP)+, D0                         ; Pull D0 from the stack
	LEA	(ukswiv-bios_address), A2                      ; Get the vector for UKSWIV
	JSR	(A2)                              ; And jump to it
	BRA	SWI_Handler_exit                  ; Jump to SWI handler exit for error handling

; -----------------------------------------------------------------------------
; SWI &100-1FF: OS_WriteI: Write immediate byte to all active output strams
; -----------------------------------------------------------------------------
SWI_OS_WriteI
	MOVE.L	(SP)+, D0                         ; Pull D0 from the stack
	BRA	SWI_OS_WriteC


; -----------------------------------------------------------------------------
; SWI &00: OS_WriteC: Write D0.B character to output stream
; -----------------------------------------------------------------------------
SWI_OS_WriteC
	MOVE.L	(wrchv-bios_address), -(SP)
	RTS

; -----------------------------------------------------------------------------
; SWI &01: OS_WriteS - Print string to output stream
; -----------------------------------------------------------------------------
SWI_OS_WriteS
	MOVE.L	D0, -(SP)                         ; Save D0
	MOVE.L	4(SP), D0                         ; Pull address from stack
	BSR	SWI_OS_Write0
	MOVE.L	D0, 4(SP)
	MOVE.L	(SP)+, D0
	RTS

; -----------------------------------------------------------------------------
; SWI &02: OS_Write0 - Print string to output stream
; -----------------------------------------------------------------------------
SWI_OS_Write0
	MOVE.L	A0, -(SP)                         ; Save A0
	MOVE.L	D0, A0

SWI_OS_Write0_loop
	MOVE.B	(A0)+, D0                         ; Get byte in D0
	BEQ	SWI_OS_Write0_exit                ; Exit if it was the last byte
	BSR	SWI_OS_WriteC                     ; Print byte
	BRA	SWI_OS_Write0_loop                ; And get the next one
	
SWI_OS_Write0_exit
	MOVE.L	A0, D0
	MOVE.L	(SP)+, A0                         ; Restore A0
	RTS

; -----------------------------------------------------------------------------
; SWI &03: OS_NewLine: Write a newline to selected output stream
; -----------------------------------------------------------------------------
SWI_OS_NewLine
	MOVE.L	D0, -(SP)                         ; Push D0 to the stack
	MOVEQ	#$0A, D0
	BSR	SWI_OS_WriteC
	MOVEQ	#$0D, D0
	BSR	SWI_OS_WriteC
	MOVE.L	(SP)+, D0                         ; Restore D0
	RTS

; -----------------------------------------------------------------------------
; SWI &04: OS_ReadC: Read character from input stream
; -----------------------------------------------------------------------------
SWI_OS_ReadC
	MOVEQ	#$00, D0                          ; Code $00 for OSRDCH call
	BSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2

SWI_OS_ReadC1
	BSR	readbyte_from_tuber2              ; Wait for data in TUBE ULA Data register 2 and return it in A
	ASR.B	#8, D0                            ; Move bit 7 to carry flag
	BRA	readbyte_from_tuber2              ; And get the last byte

; -----------------------------------------------------------------------------
; SWI &05: OS_CLI: Process a command as if typed in the CLI
; -----------------------------------------------------------------------------
SWI_OS_CLI
	MOVE.L	A6, -(SP)                         ; Save A6
	MOVE.L	D0, -(SP)                         ; Save D0
	MOVE.L	D0, A6                            ; A6 => command string

SWI_OS_CLI2
	BSR	skip_spaces                       ; Find first non-space character
	MOVE.B	(A6)+, D0
	CMP.B	#prompt, D0
	BEQ	SWI_OS_CLI2                       ; Skip leading '*'

	MOVE.B	-(A6), D0
	LEA	OS_CLI_command_table, A1          ; A1 => command table

start_compare
	MOVE.L	A6, -(SP)                         ; Save command pointer for now
	MOVE.L	(A1)+, A2                         ; Pre-fetch jump address
	CMPA	#$FFFFFFFF, A2                    ; End of table marker
	BEQ	SWI_OS_CLI_to_host                ; Jump if end of table has been reached

SWI_OS_CLI_compare_loop
	MOVE.B	(A6)+, D0                         ; Get first character in D0
	ANDI.B	#$DF, D0                          ; Make character upper-case
	CMP.B	(A1)+, D0                         ; Compare command agains table entries
	BNE	SWI_OS_CLI_skip_entry             ; No match, so skip to next entry
	CMPI.B	#'.', (A6)                        ; Is the next character a '.'?
	BEQ	SWI_OS_CLI_matched                ; Yes, so we've got a match
	CMPI.B	#' ', (A1)                        ; Has the full table entry been processed?
	BEQ	SWI_OS_CLI_matched                ; Yes, so we`ve got a match
	CMPI.B	#$00, (A1)                        ; Has the end of the table entry been reached?
	BNE	SWI_OS_CLI_compare_loop           ; If not, check next character

SWI_OS_CLI_matched
	CMPI.B	#$0D, (A6)                        ; Check if the end of command has been reached
	BEQ	SWI_OS_CLI_matched2               ; If not, skip this entry and compare against next entry
	CMPI.B	#' ', (A6)                        ; Check if the end of command has been reached
	BNE	SWI_OS_CLI_skip_entry             ; If not, skip this entry and compare against next entry

SWI_OS_CLI_matched2
	BSR	skip_spaces                       ; Set A6 to first command line parameter
	MOVE.L	(SP)+, A1                         ; Clean up stack
	JSR	(A2)                              ; We`ve found the correct entry, so jump to it
	MOVE.L	(SP)+, D0                         ; Restore D0
	MOVE.L	(SP)+, A6                         ; Restore A6
	BCS	SWI_OS_CLI_to_host2
	RTS
	
SWI_OS_CLI_skip_entry
	CMPI.B	#$00, (A1)+                       ; Loop through
	BNE	SWI_OS_CLI_skip_entry

	MOVE.L	(SP)+, A6                         ; Restore original command pointer
	BRA	start_compare

SWI_OS_CLI_to_host
	MOVE.L	(SP)+, A6                         ; Clean up stack
	MOVE.L	(SP)+, D0                         ; Restore D0

SWI_OS_CLI_to_host2
	MOVE.L	D0, A6                            ; Put pointer to string in A6
	MOVEQ	#$02, D0                          ; Code $02 for OSCLI
	BSR	writebyte_to_tuber2
	BSR	writestring_to_tuber2             ; Send string to the host

	BSR	readbyte_from_tuber2              ; Get response from the host
	MOVE.L	(SP)+, A6                         ; Restore original A0 value
	CMP.B	#$80, D0
	BEQ	SWI_OS_Byte_8E                    ; Enter language rom
	ANDI	#$FD, CCR                         ; Clear overflow flag
	RTS

; -----------------------------------------------------------------------------
; SWI &06: OS_Byte handler
; -----------------------------------------------------------------------------
SWI_OS_Byte
	CMP.B	#$80, D0
	BHS	SWI_OS_Byte80ff

	CMP.B	#$7E, D0                          ; Is it OSBYTE $7E? (Acknowledge detection of an ESCAPE condition)
	BNE	SWI_OS_Byte_007f
	BCLR.B	#6, escapeflag                    ; Clear the Escape flag, and pass the OSBYTE call to the host

SWI_OS_Byte_007f
	MOVE.L	D0, -(SP)                         ; Push D0 to the stack

	MOVE.B	#$04, D0                          ; Code $04 for OSBYTE $00-$7F call
	BSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2

	MOVE.B	D1, D0                            ; Send X value to the host
	BSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2

	MOVE.L	(SP), D0                          ; Pull D0 from the stack
	BSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2

	BSR	readbyte_from_tuber2              ; Wait for data in TUBE ULA Data register 2 and return it in A
	MOVE.B	D0, D1                            ; Set D1 to the new X value received from the host
	MOVE.L	(SP)+, D0                         ; Restore D0
	RTS

SWI_OS_Byte80ff
	CMP.B	#$82, D0                          ; Is it OSBYTE $82? (Read machine high order address)
	BEQ	SWI_OS_Byte_82
	CMP.B	#$83, D0                          ; Is it OSBYTE $83? (Read top of operating system RAM address (OSHWM)
	BEQ	SWI_OS_Byte_83
	CMP.B	#$84, D0                          ; Is it OSBYTE $84? (Read bottom of display RAM address (HIMEM)
	BEQ	SWI_OS_Byte_84

	MOVE.L	D0, -(SP)                         ; Push D0 to the stack

	MOVE.B	#$06, D0                          ; Code $06 for OSBYTE $80-$81, $85-$FF call
	BSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2

	MOVE.B	D1, D0                            ; Send X value to the host
	BSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2

	MOVE.B	D2, D0                            ; Send Y value to the host
	BSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2

	MOVE.L	(SP)+, D0                         ; Pull D0 from the stack
	BSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2

	CMP.B	#$8E, D0                          ; Is it OSBYTE $8E? (Enter language ROM)
	BEQ	SWI_OS_Byte_9D

	CMP.B	#$9D, D0                          ; Is it OSBYTE $9D? (Fast TUBE BPUT)
	BEQ	SWI_OS_Byte_9D

	MOVE.L	D0, -(SP)                         ; Push D0 to the stack

	BSR	readbyte_from_tuber2              ; Wait for data in TUBE ULA Data register 2 and return it in A
	MOVE.W	D0, -(SP)                         ; Push returned Carry flag to stack

	BSR	readbyte_from_tuber2              ; Wait for data in TUBE ULA Data register 2 and return it in A
	MOVE.B	D0, D2                            ; Set D2 to the new Y value received from the host

	BSR	readbyte_from_tuber2              ; Wait for data in TUBE ULA Data register 2 and return it in A
	MOVE.B	D0, D1                            ; Set D1 to the new X value received from the host

	MOVE.L	(SP)+, D0                         ; Pull carry flag from the stack
	ASR.B	#8, D0                            ; Move bit 7 to carry flag
	MOVE.L	(SP)+, D0                         ; Pull D0 from the stack
	RTS

; -----------------------------------------------------------------------------
; SWI &06: OS_Byte $82 handler (Read machine high order address)
; -----------------------------------------------------------------------------
SWI_OS_Byte_82                                    ; SWI &06: OS_Byte $82 handler (Read machine high order address)
	MOVE.W	(hoaddr), D1
	MOVE.L	D1, D2
	ASR.L	#8, D2                            ; D2=D1 DIV 256 (don`t know if this is useful; hoaddr is a 32-bit value)
	RTS

; -----------------------------------------------------------------------------
; SWI &06: OS_Byte $83 handler (Read bottom of user memory)
; -----------------------------------------------------------------------------
SWI_OS_Byte_83                                    ; SWI &06: OS_Byte $83 handler (Read top of operating system RAM address: OSHWM)
	MOVE.L	(oshwm), D1
	MOVE.L	D1, D2
	ASR.L	#8, D2                            ; D2=D1 DIV 256 (don`t know if this is useful; oshwm is a 32-bit value)
	RTS

; -----------------------------------------------------------------------------
; SWI &06: OS_Byte $84 handler (Read top of user memory)
; -----------------------------------------------------------------------------
SWI_OS_Byte_84                                    ; SWI &06: OS_Byte $84 handler (Read bottom of display RAM address: HIMEM)
	MOVE.L	(himem), D1
	MOVE.L	D1, D2
	ASR.L	#8, D2                            ; D2=D1 DIV 256 (don`t know if this is useful; himem is a 32-bit value)
	RTS

; -----------------------------------------------------------------------------
; SWI &06: OS_Byte &8E handler (Enter language ROM)
; -----------------------------------------------------------------------------
SWI_OS_Byte_8E
	MOVE.L	current, -(SP)                    ; Save current program on the stack
	MOVE.L	A0, current                       ; Set 'Current program'

	CMPI.L	#$00284329, 7(A0)                 ; Check for a copyright string
	BNE	SWI_OS_Byte_8E_entercode          ; It`s not a ROM, so no way to check the processor type
	MOVE.B	6(A0), D0                         ; Get ROM type byte
	ANDI.B	#%00001111, D0                    ; Mask out the processor type bits
	CMPI.B	#%00000011, D0                    ; Is it 68k code?
	BEQ	SWI_OS_Byte_8E_entercode          ; It`s the correct type of code, so start it

	MOVE.L	#err_cannot_run_code, D0          ; Point D0 to the ESCAPE error message
	MOVE.L	#0, error_pc
	BRA	SWI_OS_GenerateError              ; SWI OS_GenerateError

SWI_OS_Byte_8E_entercode
	MOVE.L	#$00000001, D0
	MOVE.L	#osstarttime, D1
	BSR	SWI_OS_Word                       ; Get program start time

	MOVEQ	#$01, D0                          ; Enter with A=1
	JMP	(A2)
	MOVE.L	(SP)+, current                    ; Restore current program
	RTS
	
; -----------------------------------------------------------------------------
; SWI &06: OS_Byte $9D handler (Fast TUBE BPUT)
; -----------------------------------------------------------------------------
SWI_OS_Byte_9D                                    ; SWI &06: OS_Byte $9D handler (Fast TUBE BPUT)
	RTS

; -----------------------------------------------------------------------------
; SWI &07: OS_Word handler
; -----------------------------------------------------------------------------
SWI_OS_Word
	CMP.L	#$000000FF, D0
	BHI	SWI_OS_Word_error

	MOVE.L	A6, -(SP)
	MOVE.L	D0, -(SP)
	MOVE.L	D1, -(SP)
	MOVE.L	D2, -(SP)
	MOVE.L	D1, A6

	TST	D0                                ; Is the reason code $00?
	BNE	SWI_OS_Word_01ff                  ; No, then jump to OSWORD $01-$FF handler
	
	MOVE.L	D3, -(SP)                         ; OSWORD &00 handler
	MOVE.L	D4, -(SP)

	MOVE.W	(hoaddr), D0                      ; Get High Order pointer
	SWAP	D0                                ; Shift to bit 16-31

	MOVE.W	(A6), D0                          ; OS_Word &00 uses only a 16-bit buffer pointer
	MOVEQ	#$00, D1
	MOVE.B	2(A6), D1                         ; Maximum length
	MOVEQ	#$00, D2
	MOVE.B	3(A6), D2                         ; Minimum allowable ASCII value
	MOVEQ	#$00, D3
	MOVE.B	4(A6), D3                         ; Maximum allowable ASCII value
	MOVEQ	#$00, D4                          ; No special flags needed for this call

	BSR	SWI_OS_ReadLine32                 ; Jump to the OS_ReadLine32 handler

	MOVE.L	(SP)+, D4                         ; Clean up stack / restore registers
	MOVE.L	(SP)+, D3
	MOVE.L	(SP)+, D2                         ; D2 isn`t used since on OSWORD $00 exit it will hold the number of bytes in the string
	MOVE.L	D1, D2
	ADDQ	#1, D2                            ; Be sure to include the <CR>
	MOVE.L	(SP)+, D1
	MOVE.L	(SP)+, D0
	MOVE.L	(SP)+, A6
	ANDI	#$FD, CCR                         ; Clear overflow flag
	RTS

SWI_OS_Word_01ff
	MOVEQ	#$00, D1
	MOVEQ	#$00, D2
	CMP.B	#$14, D0                          ; 
	BHI	SWI_OS_Word_15ff

	MOVE.L	#osword_send_table, A0
	ADD.L	D0, A0                            ; Add displacement to A0
	SUBQ	#1, A0
	MOVE.B	(A0), D1                          ; Get the send length of the parameterblock from the table for OSWORD &01-&14
	MOVE.L	#osword_recv_table, A0
	ADD.L	D0, A0                            ; Add displacement to A0
	SUBQ	#1, A0
	MOVE.B	(A0), D2                          ; Get the receive length of the parameterblock from the table for OSWORD &01-&14
	BRA	SWI_OS_Word_send

SWI_OS_Word_15ff
	CMP.B	#$7F, D0
	BHI	SWI_OS_Word_80ff
	MOVEQ	#$10, D1                          ; Number of parameters to send
	MOVEQ	#$10, D2                          ; Number of parameters to be received
	BRA	SWI_OS_Word_send

SWI_OS_Word_80ff
	MOVE.B	(A6), D1                          ; Number of parameters to send
	MOVE.B	1(A6), D2                         ; Number of parameters to be received

SWI_OS_Word_send
	MOVE.L	D0, -(SP)                         ; Push D0 to the stack
	MOVE.B	#$08, D0                          ; Code &08 for OSWORD &01-&FF calls
	BSR	writebyte_to_tuber2               ; Send OSWORD &01-&FF command to the host
	MOVE.L	(SP)+, D0                         ; Pull D0 from the stack
	BSR	writebyte_to_tuber2               ; Send OSWORD reason code to the host
	MOVE.B	D1, D0                            ; 
	BSR	writebyte_to_tuber2               ; Send number of send-parameters to the host
	SUBQ	#1, D1                            ; Replace with a DBxx instruction?
	BMI	SWI_OS_Word_receive

SWI_OS_Word_send_loop
	MOVE.B	(A6, D1), D0                      ; Send the parameter block to the host processor
	BSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2
	DBF	D1, SWI_OS_Word_send_loop

SWI_OS_Word_receive
	MOVE.B	D2, D0                            ; 
	BSR	writebyte_to_tuber2               ; Send number of receive-parameters to the host
	SUBQ	#1, D2                            ; Replace with a DBxx instruction?
	BMI	SWI_OS_Word_exit

SWI_OS_Word_receive_loop
	BSR	readbyte_from_tuber2              ; Wait for error number
	MOVE.B	D0, (A6, D2)
	DBF	D2, SWI_OS_Word_receive_loop

SWI_OS_Word_exit
	MOVE.L	(SP)+, D2
	MOVE.L	(SP)+, D1
	MOVE.L	(SP)+, D0
	MOVE.L	(SP)+, A6
	ANDI	#$FD, CCR                         ; Clear overflow flag
	RTS

SWI_OS_Word_error
	ORI	#$02, CCR                         ; Set overflow fag
	RTS

; -----------------------------------------------------------------------------
; SWI &08: OS_File: Read or write a whole file or its attributes
; -----------------------------------------------------------------------------
SWI_OS_File
	MOVE.L	D0, -(SP)                         ; Push D0 to the stack
	MOVE.B	#$14, D0                          ; Code $14 for OSFILE call
	BSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2

	MOVE.L	D5, D0                            ; Send attrs/end address
	BSR	writelong_to_tuber2
	MOVE.L	D4, D0                            ; Send length/start address
	BSR	writelong_to_tuber2
	MOVE.L	D3, D0                            ; Send exec address
	BSR	writelong_to_tuber2
	MOVE.L	D2, D0                            ; Send load address
	BSR	writelong_to_tuber2
	MOVE.L	D1, A6
	BSR	writestring_to_tuber2             ; Send string at (A6) through TUBE ULA Data register 2

	MOVE.L	(SP)+, D0                         ; Pull D0 from the stack
	BSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2

	BSR	readbyte_from_tuber2              ; Wait for data in TUBE ULA Data register 2 and return it in A
	MOVE.L	D0, -(SP)                         ; Push D0 to the stack (result code)

	BSR	readlong_from_tuber2              ; Read returned attrs
	MOVE.L	D0, D5
	BSR	readlong_from_tuber2              ; Read returned length
	MOVE.L	D0, D4
	BSR	readlong_from_tuber2              ; Read returned exec address
	MOVE.L	D0, D3
	BSR	readlong_from_tuber2              ; Read returned load address
	MOVE.L	D0, D2

	MOVE.L	(SP)+, D0                         ; Pull D0 from the stack
	RTS

; -----------------------------------------------------------------------------
; SWI &09: OS_Args: Read or write a file`s attributes
; -----------------------------------------------------------------------------
SWI_OS_Args
	MOVE.L	D0, -(SP)                         ; Push D0 to the stack
	MOVE.B	#$0C, D0                          ; Code $0C for OSARGS call
	BSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2

	MOVE.B	D1, D0                            ; File handle
	BSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2

	MOVE.L	D2, D0                            ; Pass data word in D2 to D0
	BSR	writelong_to_tuber2               ; Send 32-bit word from D0

	MOVE.L	(SP)+, D0                         ; Pull D0 from the stack (function)
	BSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2

	BSR	readbyte_from_tuber2              ; Wait for data in TUBE ULA Data register 2 and return it in A
	MOVE.L	D0, -(SP)                         ; Push D0 to the stack (result)

	BSR	readlong_from_tuber2              ; Wait for 32-bit word from Tube
	MOVE.L	D0, D2                            ; Pass data word to D2

	MOVE.L	(SP)+, D0                         ; Pull D0 from the stack
	RTS

; -----------------------------------------------------------------------------
; SWI &0A: OS_BGet: Read a single byte from an open file
; -----------------------------------------------------------------------------
SWI_OS_BGet
	MOVE.B	#$0E, D0                          ; Code $0E for BGET call
	JSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2
	MOVE.B	D1, D0                            ; File handle
	JSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2
	BRA	SWI_OS_ReadC1                     ; Get carry bit and the byte read

; -----------------------------------------------------------------------------
; SWI &0B: OS_BPut: Write a single byte to an open file
; -----------------------------------------------------------------------------
SWI_OS_BPut
	MOVE.L	D0, -(SP)                         ; Push D0 to the stack
	MOVE.B	#$10, D0                          ; Code $10 for OSBPUT call
	BSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2
	MOVE.B	D1, D0                            ; File handle
	BSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2
	MOVE.L	(SP), D0                          ; Restore D0 from the stack
	BSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2
	BSR	readbyte_from_tuber2              ; Wait for data in TUBE ULA Data register 2 and return it in A
	MOVE.L	(SP)+, D0                         ; Pull D0 from the stack
	RTS

; -----------------------------------------------------------------------------
; SWI &0C: OS_GBPB handler
; -----------------------------------------------------------------------------
SWI_OS_GBPB
	MOVE.L	D0, -(SP)                         ; Push D0 to the stack
	MOVE.B	#$16, D0                          ; Code $16 for OSGBPB call
	BSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2

	MOVE.L	D4, D0
	BSR	writelong_to_tuber2               ; Send pointer
	MOVE.L	D3, D0
	BSR	writelong_to_tuber2               ; Send count
	MOVE.L	D2, D0
	BSR	writelong_to_tuber2               ; Send address
	MOVE.L	D1, D0
	BSR	writebyte_to_tuber2               ; Send handle
	MOVE.L	(SP)+, D0
	BSR	writelong_to_tuber2               ; Send function

	BSR	readbyte_from_tuber2              ; Read returned pointer
	MOVE.L	D0, D4
	BSR	readlong_from_tuber2              ; Read returned count
	MOVE.L	D0, D3
	BSR	readlong_from_tuber2              ; Read returned address
	MOVE.L	D0, D2
	BSR	readbyte_from_tuber2              ; Read returned cycle/handle
	MOVE.L	D0, D1

	BRA	SWI_OS_ReadC1                     ; Get carry bit and result byte

; -----------------------------------------------------------------------------
; SWI &0D: OS_Find: Open or close a file for random or sequential access
; -----------------------------------------------------------------------------
SWI_OS_Find
	MOVE.L	D0, -(SP)                         ; Push D0 to the stack
	MOVE.B	#$12, D0                          ; Code $12 for OSFIND call
	BSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2
	MOVE.L	(SP)+, D0                         ; Pull D0 from the stack
	BSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2
	TST	D0                                ; Should the file be closed? (result = $00)
	BNE	SWI_OS_Find2

	MOVE.L	D0, -(SP)                         ; Push D0 to the stack
	MOVE.B	D1, D0                            ; File handle
	BSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2
	BSR	readbyte_from_tuber2              ; Wait for data in TUBE ULA Data register 2 and return it in A
	MOVE.L	(SP)+, D0                         ; Pull D0 from the stack
	RTS

SWI_OS_Find2
	BSR	writestring_to_tuber2             ; Send string at (A6) through TUBE ULA Data register 2
	BRA	readbyte_from_tuber2              ; Wait for data in TUBE ULA Data register 2 and return it in A

; -----------------------------------------------------------------------------
; SWI &0E: OS_ReadLine: Read a line from the current input stream
; -----------------------------------------------------------------------------
SWI_OS_ReadLine
	MOVE.L	D0, -(SP)                         ; Push D0 to the stack
	AND.L	#$D0000000, D0                    ; Mask out the OS_ReadLine flags (bit 31-30)
	MOVE.L	D0, D4                            ; And put them in D4
	MOVE.L	(SP)+, D0                         ; Pull D0 from the stack
	BRA	SWI_OS_ReadLine32                 ; Jump to _OS_ReadLine32 handler

; -----------------------------------------------------------------------------
; SWI &0F: OS_Control: Read or write address handlers
; -----------------------------------------------------------------------------
SWI_OS_Control
	MOVE.L	D5, -(SP)                         ; Push D0 to the stack
	MOVE.L	D4, -(SP)                         ; Push D0 to the stack
	MOVE.L	D1, -(SP)                         ; Push D0 to the stack
	MOVE.L	D0, -(SP)                         ; Push D0 to the stack

	MOVE.L	D3, D1                            ; First we`re gonna deal with the event handler
	MOVE.L	#10, D0
	BSR	SWI_OS_ChangeEnvironment
	MOVE.L	D1, D4                            ; Temporary register for address of previous event handler

	MOVE.L	D2, D1                            ; Second is the escape state handler
	MOVE.L	#9, D0
	BSR	SWI_OS_ChangeEnvironment
	MOVE.L	D1, D5                            ; Temporary register for previous escape state handler

		                                  ; And finally, the error handler and buffer
	MOVE.L	(SP)+, D1                         ; Pull D1 from the stack (address of error handler)
	MOVE.L	(SP)+, D3                         ; Pull D3 from the stack (address of error handler buffer)

	MOVE.L	#6, D0
	BSR	SWI_OS_ChangeEnvironment
	MOVE.L	D1, D0                            ; On exit, D0 contains address of previous error handler
	MOVE.L	D3, D1                            ; On exit, D1 contains address of previous error handler buffer

	MOVE.L	D4, D3                            ; On Exit, D3 contains address of previous event handler
	MOVE.L	D5, D2                            ; On Exit, D2 contains address of previous escape state handler

	MOVE.L	(SP)+, D4                         ; Pull D4 from the stack (temporary register)
	MOVE.L	(SP)+, D5                         ; Pull D5 from the stack (temporary register)

	ANDI	#$FD, CCR                         ; Clear overflow flag
	RTS

; -----------------------------------------------------------------------------
; SWI &10: OS_GetEnv: Read environment parameters
; -----------------------------------------------------------------------------
SWI_OS_GetEnv
	MOVE.L	#buffer, D0
	MOVE.L	#himem, D1
	MOVE.L	#programstarttime, D2
	RTS

; -----------------------------------------------------------------------------
; SWI &12: OS_SetEnv: Set environment parameters
; -----------------------------------------------------------------------------
SWI_OS_SetEnv
	MOVE.L	D3, -(SP)                         ; Push D3 to the stack
	MOVE.L	D2, -(SP)                         ; Push D2 to the stack
	MOVE.L	D1, -(SP)                         ; Push D1 to the stack
	MOVE.L	D0, -(SP)                         ; Push D0 to the stack

	MOVE.L	D7, D1                            ; First we`re gonna deal with the address exception handler
	MOVE.L	#4, D0
	BSR	SWI_OS_ChangeEnvironment
	MOVE.L	D1, D7

	MOVE.L	D6, D1                            ; Data abort handler
	MOVE.L	#3, D0
	BSR	SWI_OS_ChangeEnvironment
	MOVE.L	D1, D6

	MOVE.L	D5, D1                            ; Prefetch abort handler
	MOVE.L	#2, D0
	BSR	SWI_OS_ChangeEnvironment
	MOVE.L	D1, D5

	MOVE.L	D4, D1                            ; Undefined instruction handler
	MOVE.L	#1, D0
	BSR	SWI_OS_ChangeEnvironment
	MOVE.L	D1, D4

	MOVE.L	(SP)+, D1                         ; Exit handler
	MOVE.L	#11, D0
	BSR	SWI_OS_ChangeEnvironment
	MOVE.L	D1, D0                            ; We don`t have to worry about D2 becaouse OS_SetEnv doesn`t use it

	MOVE.L	(SP)+, D1                         ; Memory limit
	MOVE.L	D0, -(SP)
	MOVE.L	#0, D0
	BSR	SWI_OS_ChangeEnvironment
	
	MOVE.L	(SP)+, D0
	MOVE.L	(SP)+, D2
	MOVE.L	(SP)+, D3

	ANDI	#$FD, CCR                         ; Clear overflow flag
	RTS

; -----------------------------------------------------------------------------
; SWI &13: OS_IntOn - Enable interrupts
; -----------------------------------------------------------------------------
SWI_OS_IntOn
	ANDI.W	#$F8FF, SR
	RTS

; -----------------------------------------------------------------------------
; SWI &14: OS_IntOff - Disable interrupts
; -----------------------------------------------------------------------------
SWI_OS_IntOff
	ORI.W	#$0700, SR
	RTS

; -----------------------------------------------------------------------------
; SWI &16: OS_EnterOS - Switch to supervisor mode
; -----------------------------------------------------------------------------
SWI_OS_EnterOS
	ORI.W	#$2000, SR
	RTS

; -----------------------------------------------------------------------------
; SWI &19: OS_UnusedSWI: Read or write handler for unused SWI`s
; -----------------------------------------------------------------------------
SWI_OS_UnusedSWI
	MOVE.L	D3, -(SP)                         ; Push D3 to the stack
	MOVE.L	D2, -(SP)                         ; Push D3 to the stack
	MOVE.L	D1, -(SP)                         ; Push D3 to the stack
	MOVE.L	D0, D1
	MOVE.B	#12, D0
	BSR	SWI_OS_ChangeEnvironment
	MOVE.L	D1, D0
	MOVE.L	(SP)+, D1
	MOVE.L	(SP)+, D2
	MOVE.L	(SP)+, D3
	ANDI	#$FD, CCR                         ; Clear overflow flag
	RTS	

; -----------------------------------------------------------------------------
; SWI &29: OS_FSControl: Control the filing system manager and the filing system itself
; -----------------------------------------------------------------------------
SWI_OS_FSControl
	CMP.B	#$16, D0                          ; Is D0 within the allowed range?
	BHI	SWI_OS_FSControl_error            ; If not, exit with Not Supported error message

	MOVE.L	A2, -(SP)                         ; Push A2 to the stack

	ADD	D0, D0                            ; Multiply A2 by 4 (ADD` are quicker than MUL`s)
	ADD	D0, D0

	MOVE.L	OS_FSControl_jumptable, A2        ; Optimize this???
	ADD.L	D0, A2
	LEA	(A2), A2                          ; Get vector
	JSR	(A2)                              ; Jump to reason handler

	MOVE.L	(SP)+, A2                         ; Restore A2
	ANDI	#$FD, CCR                         ; Clear overflow flag
	RTS

SWI_OS_FSControl_error
	MOVE.L	#err_bad_fscontrol_call, D0
	ORI	#$02, CCR                         ; Set overflow fag
	RTS

; -----------------------------------------------------------------------------
; SWI &29: OS_FSControl &0A: Set filing system options
; -----------------------------------------------------------------------------
SWI_OS_FSControl0A
	MOVE.B	#$8B, D0                          ; OSBYTE &8B: Select file options
	BSR	SWI_OS_Byte                       ; Do OSBYTE call
	MOVE.B	#$10, D0                          ; Preserve old value
	RTS

; -----------------------------------------------------------------------------
; SWI &29: OS_FSControl &16: Close all open files
; -----------------------------------------------------------------------------
SWI_OS_FSControl16
	MOVEQ	#$00, D0                          ; OSFILE &00: Close file
	MOVEQ	#$00, D1                          ; File handle = 0 so close all files
	BSR	SWI_OS_File                       ; Do OSFILE call
	MOVE.B	#$16, D0                          ; Preserve old value
	RTS

; -----------------------------------------------------------------------------
; SWI &2B: OS_GenerateError: Generate an error and invoke the error handler
; -----------------------------------------------------------------------------
SWI_OS_GenerateError
	MOVE.L	(errorv-bios_address), -(SP)
	RTS

; -----------------------------------------------------------------------------
; SWI &2C: OS_ReadEscapeState: Check whether an escape contition has occurred
; -----------------------------------------------------------------------------
SWI_OS_ReadEscapeState
	BTST.B	#6, escapeflag                    ; Test if the escape flag bit is set
	BNE	SWI_OS_ReadEscapeState_set        ; Jump if the escape flag is set
	ANDI	#$FE, CCR                         ; Clear carry flag
	RTS

SWI_OS_ReadEscapeState_set
	ORI	#$01, CCR                         ; Set Carry flag
	RTS                                       ; Return

; -----------------------------------------------------------------------------
; SWI &30: OS_ServiceCall: Issue service call to module
; -----------------------------------------------------------------------------
SWI_OS_ServiceCall
	RTS

; -----------------------------------------------------------------------------
; SWI &32: OS_ReadPoint: Returns the colour and tint of a point
; -----------------------------------------------------------------------------
SWI_OS_ReadPoint
	CMP.L	#$000004FF, D0                    ; Is D0 (R0) within the allowed range?
	BHI	SWI_OS_ReadPoint_error            ; If not, exit with Out Of Range error message
	CMP.L	#$000003FF, D1                    ; Is D1 (R1) within the allowed range?
	BHI	SWI_OS_ReadPoint_error            ; If not, exit with Out Of Range error message

	MOVE.L	D0, -(SP)                         ; Push D0 to the stack
	MOVE.L	D1, -(SP)                         ; Push D1 to the stack
	MOVE.W	D0, buffer
	MOVE.W	D1, buffer+2
	MOVE.L	#buffer, D1
	MOVEQ	#$09, D0                          ; OSWORD call 9: Read pixel value
	BSR	SWI_OS_Word
	MOVEQ	#$00, D2
	MOVEQ	#$00, D3
	MOVE.B	buffer+4, D2
	MOVE.L	(SP)+, D1                         ; Pull D1 from the stack
	MOVE.L	(SP)+, D0                         ; Pull D0 from the stack
	CMP.B	#$FF, D2
	BEQ	SWI_OS_ReadPoint_error            ; Jump if point was off screen or non-graphical
	MOVEQ	#$00, D4
	ANDI	#$FD, CCR                         ; Clear overflow flag
	RTS

SWI_OS_ReadPoint_error
	MOVE.L	#err_out_of_range, D0             ; Pointer to error block in D0
	MOVEQ	#$FF, D4                          ; Point is off screen so D4(R4) = -1
	ORI	#$02, CCR                         ; Set overflow fag
	RTS

; -----------------------------------------------------------------------------
; SWI &34: OS_CallAVector: Call a vector directly
; -----------------------------------------------------------------------------
SWI_OS_CallAVector
	CMP.L	#$0000002F, A2                    ; Is A2 (R9) within the allowed range?
	BHI	SWI_OS_CallAVector_error          ; If not, exit with Not Supported error message

	MOVE.L	A2, -(SP)                         ; Push A3 to the stack

	ADD	A2, A2                            ; Multiply A2 by 4 (ADD's are quicker than MUL's)
	ADD	A2, A2

	ADD.L	vector_table, A2                  ; Optimize this???
	LEA	(A2), A2                          ; Get vector
	JSR	(A2)                              ; Jump to reason handler

	MOVE.L	(SP)+, A2                         ; Pull A2 from the stack
	ANDI	#$FD, CCR                         ; Clear overflow flag
	RTS

SWI_OS_CallAVector_error
	MOVE.L	#err_bad_vector_number, D0
	ORI	#$02, CCR                         ; Set overflow fag
	RTS

; -----------------------------------------------------------------------------
; SWI &36: OS_RemoveCursors: Remove the cursors from the screen (VDU 23,1,0;0;0;0;)
; -----------------------------------------------------------------------------
SWI_OS_RemoveCursors
	MOVE.L	D0, -(SP)                         ; Push D0 to the stack
	MOVE.L	D1, -(SP)                         ; Push D1 to the stack
	MOVE.L	#msg_swi_removecursors, D0        ; Location of message in D0
	MOVE.L	#$0B, D1                          ; Length of string
	BSR	SWI_OS_WriteN
	MOVE.L	(SP)+, D1                         ; Pull D0 from the stack
	MOVE.L	(SP)+, D0                         ; Pull D1 from the stack
	ANDI	#$FD, CCR                         ; Clear overflow flag
	RTS

; -----------------------------------------------------------------------------
; SWI &37: OS_RestoreCursors: Restore the cursors from the screen (VDU 23,1,1;0;0;0;)
; -----------------------------------------------------------------------------
SWI_OS_RestoreCursors
	MOVE.L	D0, -(SP)                         ; Push D0 to the stack
	MOVE.L	D1, -(SP)                         ; Push D1 to the stack
	MOVE.L	#msg_swi_restorecursors, D0       ; Location of message in D0
	MOVE.L	#$0B, D1                          ; Length of string
	BSR	SWI_OS_WriteN
	MOVE.L	(SP)+, D1                         ; Pull D0 from the stack
	MOVE.L	(SP)+, D0                         ; Pull D1 from the stack
	ANDI	#$FD, CCR                         ; Clear overflow flag
	RTS

; -----------------------------------------------------------------------------
; SWI &38: OS_SWINumberToString: Convert a SWI number to a SWI name string
; -----------------------------------------------------------------------------
SWI_OS_SWINumberToString
	RTS

; -----------------------------------------------------------------------------
; SWI &39: OS_SWINumberFromString: Convert a SWI name string to a SWI number
; -----------------------------------------------------------------------------
SWI_OS_SWINumberFromString
	RTS

; -----------------------------------------------------------------------------
; SWI &3A: OS_ValidateAddress: Check that a range of addresses are in logical RAM
; -----------------------------------------------------------------------------
SWI_OS_ValidateAddress
	CMPI.L	#$00000000, D0                    ; Check minimum address against lower limit of RAM
	BLT	SWI_OS_ValidateAddress_error      ; Jump if it's less than the lower limit
	CMP.L	(ramsize), D1                     ; Check maximum address against upper limit of RAM
	BHI	SWI_OS_ValidateAddress_error      ; Jump if it's more than the lower limit			; Optimize the BHI and ANDI? Carry will be set by CMP
	
	ANDI	#$FE, CCR                         ; Clear carry flag
	RTS

SWI_OS_ValidateAddress_error
	ORI	#$01, CCR                         ; Set Carry flag
	RTS

; -----------------------------------------------------------------------------
; SWI &3F: SWI_OS_CheckModeValid: Check if it is possible to change to a specified screen mode
; -----------------------------------------------------------------------------
SWI_OS_CheckModeValid
	CMPI.L	#$00000007, D0
	BHI	SWI_OS_CheckModeValid_error
	ANDI	#$FE, CCR                         ; Clear carry flag
	RTS

SWI_OS_CheckModeValid_error
	MOVE.L	#-1, D0                           ; Screen mode is non-existent
	MOVE.L	#-2, D1                           ; Unable to select alternative screen mode
	ORI	#$01, CCR                         ; Set Carry flag
	RTS

; -----------------------------------------------------------------------------
; SWI &40: SWI_OS_ChangeEnvironment: Read or write address handlers
; -----------------------------------------------------------------------------
SWI_OS_ChangeEnvironment
	MOVE.L	(changeenvironmentv-bios_address), -(SP)
	RTS

; -----------------------------------------------------------------------------
; SWI &45: OS_Plot: Perform a direct PLOT call using VDU
; -----------------------------------------------------------------------------
SWI_OS_Plot
	CMP.L	#$000004FF, D0                    ; Is D0 (R0) within the allowed range?
	BHI	SWI_OS_Plot_error                 ; If not, exit with Out of range error message
	CMP.L	#$000003FF, D1                    ; Is D1 (R1) within the allowed range?
	BHI	SWI_OS_Plot_error                 ; If not, exit with Out of range error message

	MOVE.L	D0, -(SP)                         ; Push D0 to the stack

	MOVEQ	#$19, D0                          ; VDU code for PLOT (VDU 25)
	BSR	SWI_OS_WriteC                     ; Output to host

	MOVE.L	(SP)+, D0                         ; Pull D0 from the stack
	BSR	SWI_OS_WriteC                     ; And output the PLOT command code to host

	MOVE.W	D1, D0
	BSR	SWI_OS_WriteC                     ; Output LSB of X-co-oordinate first
	LSR.L	#8, D0
	BSR	SWI_OS_WriteC                     ; And then MSB of X-co-oordinate

	MOVE.W	D2, D0
	BSR	SWI_OS_WriteC                     ; Output LSB of X-co-oordinate first
	LSR.L	#8, D0
	BSR	SWI_OS_WriteC                     ; And then MSB of X-co-oordinate

	ANDI	#$FD, CCR                         ; Clear overflow flag
	RTS                                       ; Return

SWI_OS_Plot_error
	MOVE.L	#err_out_of_range, D0             ; Pointer to error block in D0
	MOVEQ	#$FF, D4                          ; Point is off screen so D4(R4) = -1
	ORI	#$02, CCR                         ; Set overflow fag
	RTS


; -----------------------------------------------------------------------------
; SWI &46: OS_WriteN - Write the first D1 characters of the string pointed to by D0
; -----------------------------------------------------------------------------
SWI_OS_WriteN
	CMPI.L	#0, D1                            ; Check if size is 0
	BEQ	SWI_OS_WriteN_exit
	SUBQ	#1, D1

	MOVE.L	A0, -(SP)                         ; Save A0
	MOVE.L	D0, -(SP)
	MOVE.L	D1, -(SP)

	MOVE.L	D0, A0

SWI_OS_WriteN_loop
	MOVE.B	(A0)+, D0                         ; Get byte in D0
	BSR	SWI_OS_WriteC                        ; Print byte
	DBF	D1, SWI_OS_WriteN_loop

	MOVE.L	(SP)+, D1                         ; Preserve D1 on exit
	MOVE.L	(SP)+, D0                         ; Preserve D0 on exit
	MOVE.L	(SP)+, A0                         ; Restore A0

SWI_OS_WriteN_exit
	RTS

; -----------------------------------------------------------------------------
; SWI &56: OS_SetECFOrigin: Set the origin of the ECF pattern (VDU23,17,6,x;y;0,0,0)
; -----------------------------------------------------------------------------
SWI_OS_SetECFOrigin
	CMP.L	#$000004FF, D0                    ; Is D0 (R0) within the allowed range?
	BHI	SWI_OS_SetECFOrigin_error         ; If not, exit with Not Supported error message
	CMP.L	#$000003FF, D1                    ; Is D1 (R1) within the allowed range?
	BHI	SWI_OS_SetECFOrigin_error         ; If not, exit with Not Supported error message

	MOVE.L	D0, -(SP)                         ; Push D0 to the stack
	MOVE.L	D1, -(SP)
	MOVE.B	#23, D0
	BSR	SWI_OS_WriteC
	MOVE.B	#17, D0
	BSR	SWI_OS_WriteC
	MOVE.B	#6, D0
	BSR	SWI_OS_WriteC
	MOVE.L	(SP)+, D0
	ROR.L	#8, D0
	BSR	SWI_OS_WriteC                     ; X-coord MSB
	ROL.L	#8, D0
	BSR	SWI_OS_writeC                     ; X-coord LSB
	MOVE.L	D1, D0
	ROR.L	#8, D0
	BSR	SWI_OS_WriteC                     ; Y-coord MSB
	ROL.L	#8, D0
	BSR	SWI_OS_writeC                     ; Y-coord LSB
	MOVEQ	#$00, D0
	BSR	SWI_OS_WriteC
	BSR	SWI_OS_WriteC
	BSR	SWI_OS_WriteC
	MOVE.L	(SP)+, D1                         ; Pull D1 from the stack
	MOVE.L	(SP)+, D0                         ; Pull D0 from the stack
	ANDI	#$FE, CCR                         ; Clear carry flag
	RTS

SWI_OS_SetECFOrigin_error
	MOVE.L	#err_out_of_range, D0             ; Pointer to error block in D0
	ORI	#$02, CCR                         ; Set overflow fag
	RTS

; -----------------------------------------------------------------------------
; SWI &5B: OS_CRC: Calculate 16-bit CRC value
; -----------------------------------------------------------------------------
SWI_OS_CRC
	MOVE.L	A2, -(SP)                         ; Push A2 to the stack
	MOVE.L	D4, -(SP)                         ; Push D4 to the stack

	MOVE.L	D1, A2

	OR.W	D3, D3                            ; Test for zero size
	BEQ	SWI_OS_CRC_exit

SWI_OS_CRC_Byteloop
	MOVE.B	(A2), D4                          ; Get next byte
	ASL.W	#8, D4                            ; move to high order byte within word
	EOR.W	D4, D0                            ; xor data into current CRC
	MOVEQ	#7, D4                            ; 8 bits in byte - 1 for DBRA

SWI_OS_CRC_Bitloop
	ASL.W	#1, D0                            ; Shift CRC register
	BCC	SWI_CrcBlock_30                   ; Check for high order bit rolling into carry
	EORI.W	#$1021, D0                        ; Apply feedback polynomial

SWI_CrcBlock_30:
	DBRA	D4, SWI_OS_CRC_Bitloop            ; Bit loop
	ADD.W	D3, A2                            ; Increase pointer with D1
	CMP.L	A2, D2                            ; Have we reached the end?
	BLO	SWI_OS_CRC_Byteloop               ; No, process next byte (note: according to the RiscOS documentation this should be a BNE)

SWI_OS_CRC_exit
	MOVE.L	(SP)+, D4                         ; Pull D4 from the stack
	MOVE.L	(SP)+, A2                         ; Pull A2 from the stack
	ANDI	#$FD, CCR                         ; Clear overflow flag
	RTS	                                  ; Return new CRC in D0

; -----------------------------------------------------------------------------
; SWI &5D: OS_PrintChar: Send character to the printer
; -----------------------------------------------------------------------------
SWI_OS_PrintChar
	MOVE.B	D0, D2                            ; Byte to print
	MOVEQ	#$8A, D0                          ; OSWORD &8A (138): Insert value into buffer
	MOVEQ	#$03, D1                          ; Buffer 3: printer
	BSR	SWI_OS_Byte                       ; Call OS_Byte

	                                          ; Carry is set if buffer was full

;	ANDI	#$FD, CCR                         ; Clear overflow flag

	RTS                                       ; Return

; -----------------------------------------------------------------------------
; SWI &7C: OS_LeaveOS - Switch to user mode
; -----------------------------------------------------------------------------
SWI_OS_LeaveOS
	ANDI.W	#$DFFF, SR
	RTS

; -----------------------------------------------------------------------------
; SWI &7D: OS_ReadLine32: Read a line from the current input stream
; -----------------------------------------------------------------------------
SWI_OS_ReadLine32
	IFNE DEBUG
		MOVE.L	D0, A1
		MOVE.B  #2, D0
		TRAP    #15
		MOVE.B	#$0D, (A1, D1.W)
	ENDC
	IFEQ DEBUG

	CMP.L	#$FF, D1                          ; Is buffer size larger than 255 bytes?
	BHI	SWI_OS_ReadLine32_error

	MOVE.L	A0, -(SP)                         ; Push A0 to the stack
	MOVE.L	D0, A0                            ; Put buffer address in A0

	MOVE.B	#$0A, D0                          ; Code &0A for OSWORD &00 call
	BSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2

	MOVE.B	D3, D0                            ; Send parameter 6 (Maximum allowable ASCII value)
	BSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2
	MOVE.B	D2, D0                            ; Send parameter 5 (Minimum allowable ASCII value)
	BSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2
	MOVE.B	D1, D0                            ; Send parameter 4 (Maximum length)
	BSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2

	MOVE.B	#$07, D0                          ; MSB of buffer in host (&0700)
	BSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2
	MOVE.B	#$00, D0                          ; MSB of buffer in host (&0700)
	BSR	writebyte_to_tuber2               ; Wait for TUBE ULA Data register 2 to be clear and send A to Data register 2

	BSR	readbyte_from_tuber2              ; Get the parameter and store it in the parameterblock
	CMP.B	#$80, D0
	BEQ	SWI_OS_ReadLine32_escape

	MOVEQ	#$00, D1                           ; Set 'characters read' to 0

SWI_OS_ReadLine32_loop
	BSR	readbyte_from_tuber2              ; Get the parameter and store it in the parameterblock
	MOVE.B	D0, (A0)+                         ; Store byte in buffer
	ADDQ	#1, D1                            ; Increment line length
	CMP.B	#$0D, D0
	BNE	SWI_OS_ReadLine32_loop

	SUBQ	#1, D1                            ; D1=length of line without <cr>
	MOVE.L	(SP)+, A0                         ; ...and restore A0 and D0
	ENDC

	ANDI	#$FE, CCR                         ; Clear carry flag
	RTS

SWI_OS_ReadLine32_escape                          ; ESCAPE was pressed, so exit with carry set and X = &FF
	MOVE.L	(SP)+, A0                         ; ...and restore A0 and D0
	SUBQ	#1, D1                            ; Length in D1=-1 (ESCAPE was pressed)
	ORI	#$01, CCR                         ; Set Carry flag
	RTS

SWI_OS_ReadLine32_error                           ; Set overflow because buffer is larger than 255 bytes (not compatible with TUBE protocol)
	ORI	#$02, CCR                         ; Set overflow fag
	RTS


;		MOV	#LINE, A0
;.WAITKEY	TST	#0, STATUS
;		BNE	WAITKEY
;		MOVE	DATAIN, D0
;.WAITDISPLAY	TST	#1, STATUS
;		BNE	WAITDISPLAY
;		MOVE	D0, DATAOUT
;		MOVE	D0, (A0)+
;		CMP	#$0D, D0
;		BNE	WAITKEY
;		MOVE	#$0A, DATAOUT

; -----------------------------------------------------------------------------
; SWI &C0: SWI_OS_ConvertStandardDateAndTime
; -----------------------------------------------------------------------------
SWI_OS_ConvertStdDateAndTime
	MOVE.L	D3, -(SP)
	MOVE.L	standarddatetimestring, D3
	BSR	SWI_OS_ConvertDateAndTime
	MOVE.L	(SP)+, D3
	RTS

; -----------------------------------------------------------------------------
; SWI &C1: SWI_OS_ConvertDateAndTime
; -----------------------------------------------------------------------------
SWI_OS_ConvertDateAndTime
	MOVE.L	A6, -(SP)
	MOVE.L	A5, -(SP)
	MOVE.L	D1, A6
	MOVE.L	D3, A5

OS_ConvertDateAndTime_loop
	MOVE.B	(A6)+, D0
	CMP.B	#'%', D0
	BEQ	OS_ConvertDateAndTime_conv
	MOVE.B	D0, (A5)+
	CMP.B	#$00, D0
	BNE	OS_ConvertDateAndTime_loop
	MOVE.L	(SP)+, A5
	MOVE.L	(SP)+, A6
	RTS

OS_ConvertDateAndTime_conv
	MOVE.B	(A6)+, D0
	CMP.B	#'0', D0
	BNE	OS_ConvertDateAndTime_conv2
	MOVE.B	#$00, (A5)+                            ; Insert $00 byte to split string
	BRA	OS_ConvertDateAndTime_loop

OS_ConvertDateAndTime_conv2
	AND.B	#$DF, D0
	CMP.B	#'%', D0
	BNE	OS_ConvertDateAndTime_conv3
	MOVE.B	#'%', (A5)+
	BRA	OS_ConvertDateAndTime_loop

OS_ConvertDateAndTime_conv3
	CMP.B	#'Z', D0
	BNE	OS_ConvertDateAndTime_conv4
	MOVE.B	#'Z', D1                               ; Remove leading zero's marker

OS_ConvertDateAndTime_conv4
	ROL	#8, D0
	MOVE.B	(A6)+, D0
	AND.W	#$DFDF, D0
	CMP.W	#'CS', D0
	BEQ	OS_ConvertDateAndTime_cs          ; CS: Centiseconds
	CMP.W	#'SE', D0
	BEQ	OS_ConvertDateAndTime_se          ; SE: Seconds
	CMP.W	#'MI', D0
	BEQ	OS_ConvertDateAndTime_mi          ; MI: Minutes
	CMP.W	#'12', D0
	BEQ	OS_ConvertDateAndTime_12          ; 12: Hours
	CMP.W	#'24', D0
	BEQ	OS_ConvertDateAndTime_24          ; 24: Hours
	CMP.W	#'AM', D0
	BEQ	OS_ConvertDateAndTime_ampm        ; 
	CMP.W	#'PM', D0
	BEQ	OS_ConvertDateAndTime_ampm
	CMP.W	#'WE', D0
	BEQ	OS_ConvertDateAndTime_we
	CMP.W	#'W3', D0
	BEQ	OS_ConvertDateAndTime_w3
	CMP.W	#'WN', D0
	BEQ	OS_ConvertDateAndTime_wn
	CMP.W	#'DY', D0
	BEQ	OS_ConvertDateAndTime_dy
	CMP.W	#'ST', D0
	BEQ	OS_ConvertDateAndTime_st          ; Pre-/Suffix nd st rd th th th
	CMP.W	#'MO', D0
	BEQ	OS_ConvertDateAndTime_mo
	CMP.W	#'M3', D0
	BEQ	OS_ConvertDateAndTime_m3
	CMP.W	#'MN', D0
	BEQ	OS_ConvertDateAndTime_mn
	CMP.W	#'CE', D0
	BEQ	OS_ConvertDateAndTime_ce
	CMP.W	#'YR', D0
	BEQ	OS_ConvertDateAndTime_yr
	CMP.W	#'WK', D0
	BEQ	OS_ConvertDateAndTime_wk                     ; Week in the year
	CMP.W	#'DN', D0
	BEQ	OS_ConvertDateAndTime_dn                     ; Day of the year
	CMP.W	#'TZ', D0                                   ; Timezone 3 chars
	BEQ	OS_ConvertDateAndTime_tz

	MOVE.L	(SP)+, A5
	MOVE.L	(SP)+, A6
	MOVE.L	#err_unknown_percent_field, D0
	ORI	#$02, CCR                         ; Set overflow fag
	RTS

OS_ConvertDateAndTime_cs
OS_ConvertDateAndTime_se
OS_ConvertDateAndTime_mi
OS_ConvertDateAndTime_12
OS_ConvertDateAndTime_24
OS_ConvertDateAndTime_ampm
OS_ConvertDateAndTime_we
OS_ConvertDateAndTime_w3
OS_ConvertDateAndTime_wn
OS_ConvertDateAndTime_dy
OS_ConvertDateAndTime_st
OS_ConvertDateAndTime_mo
OS_ConvertDateAndTime_m3
OS_ConvertDateAndTime_mn
OS_ConvertDateAndTime_ce
OS_ConvertDateAndTime_yr
OS_ConvertDateAndTime_wk
OS_ConvertDateAndTime_dn
OS_ConvertDateAndTime_tz
	BRA	OS_ConvertDateAndTime_loop

; -----------------------------------------------------------------------------
; SWI &D0: OS_ConvertHex1 - convert a value into a string of 1 hex character
; -----------------------------------------------------------------------------
SWI_OS_ConvertHex1
	CMP.L	#2, D2                            ; Check if buffer is big enough for 8 chars + terminator byte
	BLE	SWI_OS_ConvertHex_error           ; If not, exit with error flag (V) set

	SUB.L	#2, D2                            ; We don't need D2 (buffer size) anymore, so prepare D2 for exit
	MOVE.L	A2, -(SP)                         ; Push A2 to the stack (used as pointer to string)
	MOVEM.L	D1-D3, -(SP)                      ; Push D1-3 to the stack (used as pointer to string)
	MOVE.W	#$00, D2                          ; Nibbles to do
	ROR.L	#4, D0                            ; Prepare first nibble to process
	BRA	SWI_OS_ConvertHex		

; -----------------------------------------------------------------------------
; SWI &D1: OS_ConvertHex2 - convert a value into a string of 2 hex characters
; -----------------------------------------------------------------------------
SWI_OS_ConvertHex2
	CMP.L	#3, D2                            ; Check if buffer is big enough for 8 chars + terminator byte
	BLE	SWI_OS_ConvertHex_error           ; If not, exit with error flag (V) set

	SUB.L	#3, D2                            ; We don't need D2 (buffer size) anymore, so prepare D2 for exit
	MOVE.L	A2, -(SP)                         ; Push A2 to the stack (used as pointer to string)
	MOVEM.L	D1-D3, -(SP)                      ; Push D1-3 to the stack (used as pointer to string)
	MOVE.W	#$01, D2                          ; Nibbles to do
	ROR.L	#8, D0                            ; Prepare first nibble to process
	BRA	SWI_OS_ConvertHex		

; -----------------------------------------------------------------------------
; SWI &D2: OS_ConvertHex4 - convert a value into a string of 4 hex characters
; -----------------------------------------------------------------------------
SWI_OS_ConvertHex4
	CMP.L	#5, D2                            ; Check if buffer is big enough for 8 chars + terminator byte
	BLE	SWI_OS_ConvertHex_error           ; If not, exit with error flag (V) set

	SUB.L	#5, D2                            ; We don`t need D2 (buffer size) anymore, so prepare D2 for exit
	MOVE.L	A2, -(SP)                         ; Push A2 to the stack (used as pointer to string)
	MOVEM.L	D1-D3, -(SP)                      ; Push D1-3 to the stack (used as pointer to string)
	MOVE.W	#$03, D2                          ; Nibbles to do
	ROL.L	#8, D0                           ; Prepare first nibble to process
	ROL.L	#8, D0
	BRA	SWI_OS_ConvertHex		

; -----------------------------------------------------------------------------
; SWI &D3: OS_ConvertHex6 - convert a value into a string of 6 hex characters
; -----------------------------------------------------------------------------
SWI_OS_ConvertHex6
	CMP.L	#7, D2                            ; Check if buffer is big enough for 8 chars + terminator byte
	BLE	SWI_OS_ConvertHex_error           ; If not, exit with error flag (V) set

	SUB.L	#7, D2                            ; We don`t need D2 (buffer size) anymore, so prepare D2 for exit
	MOVE.L	A2, -(SP)                         ; Push A2 to the stack (used as pointer to string)
	MOVEM.L	D1-D3, -(SP)                      ; Push D1-3 to the stack (used as pointer to string)
	MOVE.W	#$05, D2                          ; Nibbles to do
	ROR.L	#8, D0                            ; Prepare first nibble to process
	BRA	SWI_OS_ConvertHex		

; -----------------------------------------------------------------------------
; SWI &D4: OS_ConvertHex8 - Convert a value into a string of 8 hex characters
; -----------------------------------------------------------------------------
SWI_OS_ConvertHex8
	CMP.L	#9, D2                            ; Check if buffer is big enough for 8 chars + terminator byte
	BLE	SWI_OS_ConvertHex_error           ; If not, exit with error flag (V) set

	SUB.L	#9, D2                            ; We don`t need D2 (buffer size) anymore, so prepare D2 for exit
	MOVE.L	A2, -(SP)                         ; Push A2 to the stack (used as pointer to string)
	MOVEM.L	D1-D3, -(SP)                      ; Push D1-3 to the stack (used as pointer to string)
	MOVE.W	#$07, D2                          ; Nibbles to do

SWI_OS_ConvertHex
	MOVE.L	D1, A2                            ; Pointer to string in A2
	MOVE.L	D0, D1                            ; copy source

SWI_OS_ConvertHex_loop
	ROL.L	#4, D1                            ; Top nibble to bottom
 	MOVE.B	D1, D0                            ; Copy it
	AND.B	#$0F, D0                          ; Mask the nibble
	MOVE.B	#$09, D3                          ; set compare value
 	SUB.B	D0, D3                            ; Set extend for +1 if >9
	ADDI.B	#'0', D0                          ; Add '0' + extend
	CMPI.B	#$3A, D0                          ; Check wether or not it is in the range of A-F
	BLO	SWI_OS_ConvertHex_loop2           ; Jump if it`s not
	ADDI.B	#$07, D0                          ; Make ASCII A-F

SWI_OS_ConvertHex_loop2
	MOVE.B	D0, (A2)+                         ; Store in string
	DBF	D2, SWI_OS_ConvertHex_loop        ; Do next nibble

	MOVE.B	#$00, (A2)                        ; Store termination byte in string

	MOVEM.L	(SP)+, D1-D3                      ; Pull D1-D3 from the stack
	MOVE.L	D1, D0                            ; D0 is pointer to buffer on exit
	MOVE.L	A2, D1                            ; D1 is pointer to terminator on exit
	MOVE.L	(SP)+, A2                         ; Pull A2 from the stack
	ANDI	#$FD, CCR                         ; Clear overflow flag
	RTS

SWI_OS_ConvertHex_error
	ORI	#$02, CCR                         ; Set overflow fag
	RTS

; -----------------------------------------------------------------------------
; SWI &D8: OS_ConvertCardinal1
; -----------------------------------------------------------------------------
SWI_OS_ConvertCardinal1
	CMP.L	#4, D2                            ; Max string size is 3 characters plus $00 terminator
	BLO	SWI_OS_ConvertCardinal_Error
	ANDI.L	#$000000FF, D0
	BRA	SWI_OS_ConvertCardinal4

; -----------------------------------------------------------------------------
; SWI &D8: OS_ConvertCardinal2
; -----------------------------------------------------------------------------
SWI_OS_ConvertCardinal2
	CMP.L	#4, D2                            ; Max string size is 5 characters plus $00 terminator
	BLO	SWI_OS_ConvertCardinal_Error
	ANDI.L	#$0000FFFF, D0
	BRA	SWI_OS_ConvertCardinal4

; -----------------------------------------------------------------------------
; SWI &D8: OS_ConvertCardinal3
; -----------------------------------------------------------------------------
SWI_OS_ConvertCardinal3
	CMP.L	#9, D2                            ; Max string size is 9 characters plus $00 terminator
	BLO	SWI_OS_ConvertCardinal_Error
	ANDI.L	#$00FFFFFF, D0
	BRA	SWI_OS_ConvertCardinal4

; -----------------------------------------------------------------------------
; SWI &D8: OS_ConvertCardinal4
; -----------------------------------------------------------------------------
SWI_OS_ConvertCardinal4
	CMP.L	#10, D2                           ; Max string size is 9 characters plus $00 terminator
	BLO	SWI_OS_ConvertCardinal_Error

	EXG	D1, A5                            ; Put destination in A5
	MOVEM.L	D1-D7, -(SP)                      ; SAVE REGISTERS
 	MOVE.L	D0, D7                            ; SAVE IT HERE
	BPL.S	SWI_OS_ConvertCardinal4_01
	NEG.L	D7	                          ; CHANGE TO POSITIVE
	BMI.S	SWI_OS_ConvertCardinal4_57        ; SPECIAL CASE (-0)
	MOVE.B	#'-', (A5)+                       ; PUT IN NEG SIGN

SWI_OS_ConvertCardinal4_01
	CLR.W	D4	                          ; FOR ZERO SURPRESS
	MOVEQ.L	#10, D6	                          ; COUNTER

SWI_OS_ConvertCardinal4_0
	MOVEQ.L	#1, D2	                          ; VALUE TO SUB
 	MOVE.L	D6, D1	                          ; COUNTER
	SUBQ.L	#1, D1	                          ; ADJUST - FORM POWER OF TEN
	BEQ.S	SWI_OS_ConvertCardinal4_2         ; IF POWER IS ZERO

SWI_OS_ConvertCardinal4_1
	MOVE.W	D2, D3	                          ; D3=LOWER WORD
	MULU.W	#10, D3
	SWAP.W	D2	                          ; D2=UPPER WORD
	MULU.W	#10, D2
	SWAP.W	D3	                          ; ADD UPPER TO UPPER
	ADD.W	D3, D2
	SWAP.W	D2	                          ; PUT UPPER IN UPPER
	SWAP.W	D3	                          ; PUT LOWER IN LOWER
 	MOVE.W	D3 ,D2	                          ; D2=UPPER & LOWER
	SUBQ.L	#1, D1
	BNE	SWI_OS_ConvertCardinal4_1

SWI_OS_ConvertCardinal4_2
	CLR.L	D0                                ; HOLDS SUB AMT

SWI_OS_ConvertCardinal4_22
	CMP.L	D2, D7
	BLT.S	SWI_OS_ConvertCardinal4_3         ; IF NO MORE SUB POSSIBLE
	ADDQ.L	#1, D0                            ; BUMP SUBS
	SUB.L	D2, D7                            ; COUNT DOWN BY POWERS OF TEN
	BRA.S	SWI_OS_ConvertCardinal4_22        ; DO MORE

SWI_OS_ConvertCardinal4_3
	TST.B	D0                                ; ANY VALUE?
	BNE.S	SWI_OS_ConvertCardinal4_4
	TST.W	D4                                ; ZERO SURPRESS
	BEQ.S	SWI_OS_ConvertCardinal4_5

SWI_OS_ConvertCardinal4_4
	ADDI.B	#$30, D0                          ; BINARY TO ASCII
	MOVE.B	D0, (A5)+                         ; PUT IN BUFFER
	MOVE.B	D0, D4                            ; MARK AS NON ZERO SURPRESS

SWI_OS_ConvertCardinal4_5
	SUBQ.L	#1, D6                            ; NEXT POWER
	BNE	SWI_OS_ConvertCardinal4_0
	TST.W	D4                                ; SEE IF ANYTHING PRINTED
	BNE	SWI_OS_ConvertCardinal4_6

SWI_OS_ConvertCardinal4_57
	MOVE.B	#'0', (A5)+                       ; PRINT AT LEAST A ZERO

SWI_OS_ConvertCardinal4_6
	MOVE.B	#0, (A5)                          ; PUT TERMINATOR
	MOVEM.L	(SP)+, D1-D7                      ; RESTORE REGISTERS
	EXG	A5, D1                            ; Restore A5
	ANDI	#$FD, CCR                         ; Clear overflow flag
 	RTS                                       ; END OF ROUTINE

SWI_OS_ConvertCardinal_Error
	ORI	#$02, CCR                         ; Set overflow fag
	RTS

;		lea	string,A0
;		adda.l	#40,A0
;		move.l	#47110815,D0
;hex2dec_loop:
;		divul.l	#10,D1:D0
;		add.b	#$30,D1
;		move.b	D1,-(A0)
;		cmpi.w	#10,D0
;		bhs.s	hex2dec_loop
;		add.b	#$30,D0
;		move.b	D0,-(A0)
;		move.l	A0,-(SP)
;		move.w	#$09,-(SP)
;		trap	#1
;		addq.l	#6,SP
;		move.w	#8,-(SP)	; warten 
;		trap	#1
;		addq.l	#2,SP
;		clr.w	-(SP)		; ..und tsch
;		trap	#1
;		BSS
;		string:	DS.L50
;		END

; -----------------------------------------------------------------------------
; SWI &DD: OS_ConvertBinary1
; -----------------------------------------------------------------------------
SWI_OS_ConvertBinary1
	CMP.L	#9, D2                            ; Check if buffer is big enough for 8 chars + terminator byte
	BLE	SWI_OS_ConvertHex_error           ; If not, exit with error flag (V) set

	SUB.L	#9, D2                            ; We don`t need D2 (buffer size) anymore, so prepare D2 for exit
	MOVE.L	A2, -(SP)                         ; Push A2 to the stack (used as pointer to string)
	MOVEM.L	D1-D3, -(SP)                      ; Push D1-3 to the stack (used as pointer to string)
	MOVEQ	#7, D2                            ; Result is 8-bit wide
	ROR.L	#8, D0                            ; Prepare first bit to process
	BRA	SWI_OS_ConvertBinary

; -----------------------------------------------------------------------------
; SWI &DE: OS_ConvertBinary2
; -----------------------------------------------------------------------------
SWI_OS_ConvertBinary2
	CMP.L	#17, D2                           ; Check if buffer is big enough for 8 chars + terminator byte
	BLE	SWI_OS_ConvertHex_error           ; If not, exit with error flag (V) set

	SUB.L	#17, D2                           ; We don`t need D2 (buffer size) anymore, so prepare D2 for exit
	MOVE.L	A2, -(SP)                         ; Push A2 to the stack (used as pointer to string)
	MOVEM.L	D1-D3, -(SP)                      ; Push D1-3 to the stack (used as pointer to string)
	MOVEQ	#15, D2                           ; Result is 16-bit wide
	ROL.L	#8, D0                            ; Prepare first bit to process
	ROL.L	#8, D0
	BRA	SWI_OS_ConvertBinary

; -----------------------------------------------------------------------------
; SWI &DF: OS_ConvertBinary3
; -----------------------------------------------------------------------------
SWI_OS_ConvertBinary3
	CMP.L	#25, D2                           ; Check if buffer is big enough for 8 chars + terminator byte
	BLE	SWI_OS_ConvertHex_error           ; If not, exit with error flag (V) set

	SUB.L	#25, D2                           ; We don`t need D2 (buffer size) anymore, so prepare D2 for exit
	MOVE.L	A2, -(SP)                         ; Push A2 to the stack (used as pointer to string)
	MOVEM.L	D1-D3, -(SP)                      ; Push D1-3 to the stack (used as pointer to string)
	MOVEQ	#23, D2                           ; Result is 24-bit wide
	ROR.L	#8, D0                            ; Prepare first bit to process
	BRA	SWI_OS_ConvertBinary

; -----------------------------------------------------------------------------
; SWI &E0: OS_ConvertBinary4
; -----------------------------------------------------------------------------
SWI_OS_ConvertBinary4
	CMP.L	#33, D2                           ; Check if buffer is big enough for 32 chars + terminator byte
	BLE	SWI_OS_ConvertBinary_error        ; If not, exit with error flag (V) set

	SUB.L	#33, D2                           ; We don`t need D2 (buffer size) anymore, so prepare D2 for exit
	MOVE.L	A2, -(SP)                         ; Push A2 to the stack (used as pointer to string)
	MOVEM.L	D1-D3, -(SP)                      ; Push D1-3 to the stack (used as pointer to string)
	MOVEQ	#31, D2                           ; Result is 16-bit wide

SWI_OS_ConvertBinary
	MOVE.L	D1, A2                            ; Pointer to buffer in A2
	MOVEQ	#0, D3                            ; Temp register needed for extend bit

SWI_OS_ConvertBinary_loop
	MOVEQ	#$30, D1                          ; ASCII value of '0'
	ASL.L	#1, D0                            ; Shift leftmost bit into carry
	ABCD	D3, D1                            ; Add carry to ASCII value
	MOVE.B	D1, (A2)+                         ; Put character in buffer
	DBF	D2, SWI_OS_ConvertBinary_loop

	MOVE.B	#$00, (A2)                        ; Store termination byte in string

	MOVEM.L	(SP)+, D1-D3                      ; Pull D1-D3 from the stack
	MOVE.L	D1, D0                            ; D0 is pointer to buffer on exit
	MOVE.L	A2, D1                            ; D1 is pointer to terminator on exit
	MOVE.L	(SP)+, A2                         ; Pull A2 from the stack
	ANDI	#$FD, CCR                         ; Clear overflow flag
	RTS

SWI_OS_ConvertBinary_error
	ORI	#$02, CCR                         ; Set overflow fag
	RTS

; -----------------------------------------------------------------------------
; SWI &EC: OS_ConvertFileSize
; -----------------------------------------------------------------------------
SWI_OS_ConvertFileSize
	MOVE.L	D7, -(SP)                         ; Push D7 to the stack (temp variable)
	MOVE.L	D1, -(SP)                         ; Push D1 to the stack

	MOVE.L	D0, D1
	ANDI.L	#$C0000000, D1                    ; Mask out the GB bits
	BEQ	SWI_OS_ConvertFileSizeM           ; Not in GB range
	LSR.L	#8, D1                            ; Divide by 1024*1024*1024 (=256*256*256*64)
	LSR.L	#8, D1
	LSR.L	#8, D1
	LSR.L	#6, D1
	MOVE.L	D1, D0                            ; Put size in D0
	MOVE.B	'M', D7
	BRA	SWI_OS_ConvertFileSizePrint

SWI_OS_ConvertFileSizeM
	MOVE.L	D0, D1
	ANDI.L	#$FFF00000, D1                    ; Mask out the MB bits
	BEQ	SWI_OS_ConvertFileSizeK           ; Not in MB range
	LSR.L	#8, D1                            ; Divide by 1024*1024 (=256*256*16)
	LSR.L	#8, D1
	LSR.L	#4, D1
	MOVE.L	D1, D0                            ; Put size in D0
	MOVE.B	'K', D7
	BRA	SWI_OS_ConvertFileSizePrint

SWI_OS_ConvertFileSizeK
	MOVE.L	D0, D1
	ANDI.L	#$FFFFFC00, D1                    ; Mask out the KB bits
	BEQ	SWI_OS_ConvertFileSizePrint       ; Not in KB range
	LSR.L	#8, D1                            ; Divide by 1024 (=256*4)
	LSR.L	#2, D1
	MOVE.L	D1, D0                            ; Put size in D0
	MOVE.B	#$00, D7

SWI_OS_ConvertFileSizePrint
	MOVE.L	(SP)+, D1                         ; Restore D1
	BSR	SWI_OS_ConvertCardinal4           ; Convert value in D0 to string in D1
	MOVE.L	D1, A6
	MOVE.B	#$20, (A6)+
	MOVE.B	D7, (A6)+
	MOVE.B	#'b', (A6)+
	MOVE.B	#'y', (A6)+
	MOVE.B	#'t', (A6)+
	MOVE.B	#'e', (A6)+
	MOVE.B	#'s', (A6)+
	MOVE.B	#$00, (A6)+
	RTS

; -----------------------------------------------------------------------------
; SWI &40380: Debugger_Disassemble
; -----------------------------------------------------------------------------
SWI_Debugger_Disassemble
	MOVE.L	D0, -(SP)                         ; Push D0 to the stack
	ANDI.L	#$00000001, D0                    ; Check if it`s word aligned
	BNE	SWI_Debugger_Disassemble_Error    ; If not, then exit with error
	MOVE.L	(SP)+, D0                         ; Restore D0
	EXG	D0, A0                            ; Put the source address in A0
	EXG	D1, A1                            ; Put the pointer to destination in A1
	MOVE.W	(A0), D3                          ; Get opcode

	MOVE.L	#SWI_Debugger_Disassemble_table, A2
SWI_Debugger_Disassemble_loop
	MOVE.W	(A0), D3                          ; Get opcode
	MOVE.W	(A2)+, D2                         ; AND mask
	CMPI.W	#$0000, D2
	BEQ	SWI_Debugger_Disassemble_nfound
	AND.W	D2, D3                            ; Mask out unwanted bits
	CMP.W	(A2)+, D3                         ; Check if it`s the correct opcode
	BEQ	SWI_Debugger_Disassemble_found

SWI_Debugger_Disassemble_loop2
	CMP.B	#$00, (A2)+
	BNE	SWI_Debugger_Disassemble_loop2
	BRA	SWI_Debugger_Disassemble_loop

SWI_Debugger_Disassemble_found                    ; We`ve found the correct opcode
	MOVE.B	(A2)+, (A1)+
	BNE	SWI_Debugger_Disassemble_found
	MOVE.B	#$00, (A1)
	MOVE.L	(SP)+, D0
	RTS

SWI_Debugger_Disassemble_nfound
	MOVE.L	#'????', (A1)+
	MOVE.B	#$00, (A1)
	MOVE.L	(SP)+, D0
	RTS

SWI_Debugger_Disassemble_Error
	MOVE.L	(SP)+, D0                         ; Restore D0
	MOVE.L	#err_unaligned_address, D0        ; Pointer to error block
	ORI	#$02, CCR                         ; Set overflow fag
	RTS



; -----------------------------------------------------------------------------
; Built-in OS applications
;
; *ERROR
; -----------------------------------------------------------------------------
star_error
	BSR	not_supported
	RTS

; -----------------------------------------------------------------------------
; *FLASH
; -----------------------------------------------------------------------------
flash
	MOVE.L	A6, -(SP)                         ; Push A6 to the stack
	BSR	skip_spaces                       ; Loop until (A6) isn`t ASCII value 32 (space)
	CMP.B	#$0D, (A6)
	BEQ	flash_error                       ; No parameter, so generate error

flash2
	CMP.B	#$0D, (A6)
	BEQ	flash3
	CMP.B	#' ', (A6)                        ; Is it ASCII value 32 (space)?
	BNE	flash2                            ; No? Then check next byte
		                                  ; Too many parameters, so generate error
flash_error
	MOVE.L	#msg_flash_error, D0              ; Location of message in A0
	BSR	SWI_OS_Write0

flash_exit
	MOVE.L	(SP)+, A6                         ; Restore A6
	RTS

flash3
	MOVE.L	(SP)+, A6                         ; Restore A6
	MOVE.L	A6, D1                            ; Pointer to file in D1
	MOVE.L	#$00000005, D0                    ; OS_File reason 5: read catalogue information
	BSR	SWI_OS_File
	CMP.L	#$00020000, D4
	BNE	flash_wrongfile

	MOVE.L	A6, D1                            ; Pointer to file in D1
	MOVE.L	#$00000400, D2                    ; Load bios at $00000400
	MOVEQ	#$00, D3
	MOVE.L	#$000000FF, D0                    ; OS_File reason 5: read catalogue information
	BSR	SWI_OS_File

	MOVE.L	#msg_flash_warning, D0            ; Print flash warning
	BSR	SWI_OS_Write0

	MOVEQ	#$15, D0                          ; *FX 21,0 (Flush keyboard buffer)
	MOVEQ	#$00, D1
	BSR	SWI_OS_Byte

	BSR	SWI_OS_ReadC                      ; Read one character from the input stream
	BSR	SWI_OS_WriteC
	ANDI.B	#%11011111, D0                    ; Make character upper-case
	CMP.B	#'Y', D0                          ; Is it 'Y'?
	BNE	flash_exit

; Since the BIOS won`t be available when the flash ROM is switched to command mode, this code must be relocated to RAM
;	MOVE.L	#msg_flash_identify, D0           ; Print ID message
;	BSR	SWI_OS_Write0
;	BSR	flash_identify
	MOVE.L	#buffer, D1                       ; D1 is pointer to buffer
	MOVEQ	#$FF, D2                          ; Size of buffer
;	BSR	SWI_OS_ConvertHex2                ; Print device id
;	BSR	SWI_OS_Write0

	MOVE.L	#msg_flash_start, D0              ; Print flashing sector
	BSR	SWI_OS_Write0
	MOVE.B	D7, D0
	ADD.B	#'0', D0
	BSR	SWI_OS_WriteC
	BSR	SWI_OS_NewLine

	MOVE.L	#msg_flash_finish, D0             ; Location of message in A0
	BSR	SWI_OS_Write0
	BRA	flash_exit

flash_wrongfile
	MOVE.L	#msg_flash_wrongfile, D0          ; Location of message in A0
	BSR	SWI_OS_Write0
	RTS

flash_help
	MOVE.L	#msg_flash_help, D0               ; Location of message in A0
	BSR	SWI_OS_Write0
	RTS

; -----------------------------------------------------------------------------
; *GO
; -----------------------------------------------------------------------------
go
	BSR	get_hex_value                     ; Return hex value of string at (A6) in D1
	BCS	go_help                           ; Jump if the parameter wasn`t valid

	BSR	skip_spaces                       ; Loop until (A6) isn`t ASCII value 32 (space)
	BNE	go_help                           ; Jump if there were more parameters

	MOVE.L	D1, D0

	MOVE.L	#buffer, D1                       ; D1 is pointer to buffer
	MOVEQ	#$FF, D2                          ; Size of buffer
	BSR	SWI_OS_ConvertHex8
	BSR	SWI_OS_Write0

	BRA	SWI_OS_Byte_8E                    ; Check copyright message and ROM type and enter language ROM

go_help
	MOVE.L	#msg_go_help, D0                  ; Location of message in A0
	BSR	SWI_OS_Write0
	RTS

; -----------------------------------------------------------------------------
; *HELP
; -----------------------------------------------------------------------------
help
	MOVE.L	A6, -(SP)                         ; Push A0 to the stack
	CMPI.B	#$0D, (A6)                        ; Is the command terminated with a CR?
	BNE	help_extended
	MOVE.L	#msg_help, D0                     ; Location of message in D0
	BSR	SWI_OS_Write0
	MOVE.L	#msg_help_std, D0                 ; Location of message in D0
	BSR	SWI_OS_Write0
	BRA	help_to_host                      ; Pass the command to the host

help_extended
	ANDI.B	#$DF, (A6)
	CMPI.B	#$54, (A6)+                       ; Is the parameter 'TUBE'?
	BNE	help_swi
	ANDI.B	#$DF, (A6)
	CMPI.B	#$55, (A6)+                       ; Is the parameter 'TUBE'?
	BNE	help_exit
	ANDI.B	#$DF, (A6)
	CMPI.B	#$42, (A6)+                       ; Is the parameter 'TUBE'?
	BNE	help_exit
	ANDI.B	#$DF, (A6)
	CMPI.B	#$45, (A6)+                       ; Is the parameter 'TUBE'?
	BNE	help_exit
	CMPI.B	#$0D, (A6)                        ; Is the command terminated with a CR?
	BNE	help_exit

	MOVE.L	#msg_help, D0                     ; Location of short help message in D0
	BSR	SWI_OS_Write0
	MOVE.L	#msg_help_extd, D0                ; Location of extended help message in D0
	BSR	SWI_OS_Write0

help_exit
	MOVE.L	(SP)+, A6                         ; Pop A0 from the stack
	ANDI	#$FE, CCR                         ; Clear the carry-bit and exit
	RTS

help_swi
	ANDI.B	#$DF, -(A6)
	CMPI.B	#$53, (A6)+                       ; Is the parameter 'SWI'?
	BNE	help_to_host
	ANDI.B	#$DF, (A6)
	CMPI.B	#$57, (A6)+                       ; Is the parameter 'SWI'?
	BNE	help_to_host
	ANDI.B	#$DF, (A6)
	CMPI.B	#$49, (A6)+                       ; Is the parameter 'SWI'?
	BNE	help_to_host
	CMPI.B	#$0D, (A6)                        ; Is the command terminated with a CR?
	BNE	help_to_host

	MOVE.L	#msg_help, D0                     ; Location of short help message in D0
	BSR	SWI_OS_Write0

	LEA	SWITable, A6

help_swi_loop
	MOVE.B	#$20, D0
	BSR	SWI_OS_WriteC
	BSR	SWI_OS_WriteC
	BSR	SWI_OS_WriteC
	MOVE.L	(A6)+, D0                         ; SWI nr.
	MOVE.L	(A6)+, D0                         ; Jump address
	CMPI.L	#$FFFFFFFF, D0                    ; End of table?
	BEQ	help_exit

	MOVE.L	A6, D0                            ; Put pointer to SWI string in D0
	BSR	SWI_OS_Write0                     ; Print string
	MOVE.L	D0, A6
	CMPI.B	#$00, -(A6)                       ; Does nothing, only to decrease A0 by 1
	BSR	SWI_OS_NewLine

help_swi_loop2
	CMPI.B	#$FF, (A6)+
	BEQ	help_exit
	MOVE.L	A6, D0
	ANDI.B	#$03, D0
	BNE	help_swi_loop2
	BRA	help_swi_loop

help_to_host
	MOVE.L	(SP)+, A6                         ; Pop A0 from the stack
	ORI	#$01, CCR                         ; Set the carry bit and exit
	RTS

; -----------------------------------------------------------------------------
; *MON
; -----------------------------------------------------------------------------
mon
	CMPI.B	#$0D, (A6)                        ; Is the command terminated with a CR?
	BNE	monitor_help

	MOVE.L	#msg_monitor, D0                  ; Location of message in A0
	BSR	SWI_OS_Write0

monitor_loop
	BSR	SWI_OS_NewLine
	MOVE.B	#'>', D0                          ; > Prompt
	BSR	SWI_OS_WriteC

	MOVE.L	#buffer, D0                       ; Buffer address
	MOVE.B	#$FF, D1                          ; Maximum line length is 255
	MOVE.B	#$20, D2                          ; Minimum acceptable ASCII value
	MOVE.B	#$FF, D3                          ; Maximum acceptable ASCII value
	MOVE.L	#OS_ReadLine32, A0                ; SWI OS_ReadLine32: Read line from input
	TRAP	#SWI
	BCS	monitor_escape                    ; Jump if user pressed ESCAPE

	MOVE.L	#buffer, A6                       ; A6 points to the line read by the OS_ReadLine32 call
	MOVE.B	(A6)+, D0
	ANDI.B	#$DF, D0                          ; Make character upper-case

	CMP.B	#'D', D0                          ; Is it D?
	BEQ	monitor_dumpmem                   ; Yes, so jump to (D)ump memory
	CMP.B	#'F', D0                          ; Is it D?
	BEQ	monitor_fill                      ; Yes, so jump to (F)ill memory
	CMP.B	#'G', D0                          ; Is it D?
	BEQ	monitor_go                        ; Yes, so jump to (G)o
	CMP.B	#'H', D0                          ; Is it H?
	BEQ	monitor_inhelp                    ; Yes, so jump to (H)elp
	CMP.B	#'L', D0                          ; Is it H?
	BEQ	monitor_load                      ; Yes, so jump to (L)oad
	CMP.B	#'Q', D0                          ; Is it Q?
	BEQ	monitor_exit                      ; Yes, so jump to (Q)uit
	CMP.B	#'R', D0                          ; Is it R?
	BEQ	monitor_setreg                    ; Yes, so jump to Set (R)egister contents
	CMP.B	#'S', D0                          ; Is it H?
	BEQ	monitor_save                      ; Yes, so jump to (S)ave
	CMP.B	#'T', D0                          ; Is it H?
	BEQ	monitor_trace                     ; Yes, so jump to (T)race
	CMP.B	#'V', D0                          ; Is it H?
	BEQ	monitor_viewregs                  ; Yes, so jump to (V)iew register contents
	MOVE.L	#msg_monitor_unknown_cmd, D0      ; Location of message in A0
	BSR	SWI_OS_Write0
	BRA	monitor_loop

monitor_escape
	MOVEQ	#$7E, D0                          ; OSBYTE $7E: Acknowledge detection of an ESCAPE condition
	BSR	SWI_OS_Byte
	BRA	monitor_loop

monitor_dumpmem
	BSR	not_supported
	BRA	monitor_loop

monitor_fill
	BSR	not_supported
	BRA	monitor_loop

monitor_go
	BSR	not_supported
	BRA	monitor_loop

monitor_inhelp
	MOVE.L	#msg_monitor_help, D0             ; Location of message in A0
	BSR	SWI_OS_Write0
	BRA	monitor_loop

monitor_load
	BSR	not_supported
	BRA	monitor_loop

monitor_exit
	BSR	SWI_OS_NewLine                    ; Print newline
	RTS                                       ; Exit the monitor

monitor_setreg
	BSR	not_supported
	BRA	monitor_loop

monitor_save
	BSR	not_supported
	BRA	monitor_loop

monitor_trace
	BSR	not_supported
	BRA	monitor_loop

monitor_viewregs

	MOVE.B	#$20, D0
	MOVE.W	#$32, D1                          ; Number of spaces

monitor_viewregs2
	BSR	SWI_OS_WriteC
	DBF	D1, monitor_viewregs2

	MOVE.L	#msg_statusregister, D0           ; Location of message in A0
	BSR	SWI_OS_Write0
	BSR	SWI_OS_NewLine

	MOVE.L	#$44303A00, buffer                ; 'D0:' + $00 terminator
	MOVE.L	#buffer+3, D1                     ; D1 is pointer to buffer
	MOVE.L	#$000000FC, D2                    ; Size of buffer
	BSR	SWI_OS_ConvertHex8
	MOVE.B	#$20, buffer+$0B

	MOVE.L	#$44313A00, buffer+$0C            ; 'D1:' + $00 terminator
	MOVE.L	D1, D0
	MOVE.L	#buffer+$0F, D1                   ; D1 is pointer to buffer
	MOVE.L	#$000000F4, D2                    ; Size of buffer
	BSR	SWI_OS_ConvertHex8
	MOVE.B	#$20, buffer+$17

	MOVE.L	#$44323A00, buffer+$18            ; 'D2:' + $00 terminator
	MOVE.L	D2, D0
	MOVE.L	#buffer+$1B, D1                   ; D1 is pointer to buffer
	MOVE.L	#$000000E8, D2                    ; Size of buffer
	BSR	SWI_OS_ConvertHex8
	MOVE.B	#$20, buffer+$23

	MOVE.L	#$44333A00, buffer+$24            ; 'D3:' + $00 terminator
	MOVE.L	D3, D0
	MOVE.L	#buffer+$27, D1                   ; D1 is pointer to buffer
	MOVE.L	#$000000DC, D2                    ; Size of buffer
	BSR	SWI_OS_ConvertHex8
	MOVE.B	#$20, buffer+$2F

	MOVE.L	#$53523A00, buffer+$30            ; 'SR:' + $00 terminator
	MOVE.L	#buffer+$33, D1                   ; D1 is pointer to buffer
	MOVE.L	#$000000CD, D2                    ; Size of buffer
	MOVE.W	SR, D0                            ; Put the status register in D0
	BSR	SWI_OS_ConvertBinary2             ; Prepare string
	MOVE.B	#$20, buffer+$43
	MOVE.L	#$20200A0D, buffer+$44            ; Add some spaces to make it long-word aligned

	MOVE.L	#$44343A00, buffer+$48            ; 'D4:' + $00 terminator
	MOVE.L	D4, D0
	MOVE.L	#buffer+$4B, D1                   ; D1 is pointer to buffer
	MOVE.L	#$000000B8, D2                    ; Size of buffer
	BSR	SWI_OS_ConvertHex8
	MOVE.B	#$20, buffer+$53

	MOVE.L	#$44353A00, buffer+$54            ; 'D5:' + $00 terminator
	MOVE.L	D5, D0
	MOVE.L	#buffer+$57, D1                   ; D1 is pointer to buffer
	MOVE.L	#$000000AC, D2                    ; Size of buffer
	BSR	SWI_OS_ConvertHex8
	MOVE.B	#$20, buffer+$5F

	MOVE.L	#$44363A00, buffer+$60            ; 'D6:' + $00 terminator
	MOVE.L	D6, D0
	MOVE.L	#buffer+$63, D1                   ; D1 is pointer to buffer
	MOVE.L	#$000000A0, D2                    ; Size of buffer
	BSR	SWI_OS_ConvertHex8
	MOVE.B	#$20, buffer+$6B

	MOVE.L	#$44373A00, buffer+$6C            ; 'D7:' + $00 terminator
	MOVE.L	D7, D0
	MOVE.L	#buffer+$6F, D1                   ; D1 is pointer to buffer
	MOVE.L	#$00000094, D2                    ; Size of buffer
	BSR	SWI_OS_ConvertHex8
	MOVE.B	#$20, buffer+$77

	MOVE.L	#$55533A00, buffer+$78            ; 'US:' + $00 terminator
	MOVE.L	USP, A0
	MOVE.L	A0, D0
	MOVE.L	#buffer+$7B, D1                   ; D1 is pointer to buffer
	MOVE.L	#$00000094, D2                    ; Size of buffer
	BSR	SWI_OS_ConvertHex8
	MOVE.B	#$20, buffer+$83
	MOVE.L	#$20200A0D, buffer+$84            ; Add some spaces to make it long-word aligned

	MOVE.L	#$41303A00, buffer+$88            ; 'A0:' + $00 terminator
	MOVE.L	A0, D0
	MOVE.L	#buffer+$8B, D1                   ; D1 is pointer to buffer
	MOVE.L	#$00000094, D2                    ; Size of buffer
	BSR	SWI_OS_ConvertHex8
	MOVE.B	#$20, buffer+$93

	MOVE.L	#$41313A00, buffer+$94            ; 'A1:' + $00 terminator
	MOVE.L	A1, D0
	MOVE.L	#buffer+$97, D1                   ; D1 is pointer to buffer
	MOVE.L	#$00000094, D2                    ; Size of buffer
	BSR	SWI_OS_ConvertHex8
	MOVE.B	#$20, buffer+$9F

	MOVE.L	#$41323A00, buffer+$A0            ; 'A2:' + $00 terminator
	MOVE.L	A2, D0
	MOVE.L	#buffer+$A3, D1                   ; D1 is pointer to buffer
	MOVE.L	#$00000094, D2                    ; Size of buffer
	BSR	SWI_OS_ConvertHex8
	MOVE.B	#$20, buffer+$AB

	MOVE.L	#$41333A00, buffer+$AC            ; 'A3:' + $00 terminator
	MOVE.L	A3, D0
	MOVE.L	#buffer+$AF, D1                   ; D1 is pointer to buffer
	MOVE.L	#$00000094, D2                    ; Size of buffer
	BSR	SWI_OS_ConvertHex8
	MOVE.B	#$20, buffer+$B7

	MOVE.L	#$53533A00, buffer+$B8            ; 'SS:' + $00 terminator
	MOVE.L	SP, D0
	MOVE.L	#buffer+$BB, D1                   ; D1 is pointer to buffer
	MOVE.L	#$00000094, D2                    ; Size of buffer
	BSR	SWI_OS_ConvertHex8
	MOVE.B	#$20, buffer+$C3
	MOVE.L	#$20200A0D, buffer+$C4            ; Add some spaces to make it long-word aligned

	MOVE.L	#$41333A00, buffer+$C8            ; 'A4:' + $00 terminator
	MOVE.L	A4, D0
	MOVE.L	#buffer+$CB, D1                   ; D1 is pointer to buffer
	MOVE.L	#$00000094, D2                    ; Size of buffer
	BSR	SWI_OS_ConvertHex8
	MOVE.B	#$20, buffer+$D3

	MOVE.L	#$41353A00, buffer+$D4            ; 'A5:' + $00 terminator
	MOVE.L	A5, D0
	MOVE.L	#buffer+$D7, D1                   ; D1 is pointer to buffer
	MOVE.L	#$00000094, D2                    ; Size of buffer
	BSR	SWI_OS_ConvertHex8
	MOVE.B	#$20, buffer+$DF

	MOVE.L	#$41363A00, buffer+$E0            ; 'A6:' + $00 terminator
	MOVE.L	A6, D0
	MOVE.L	#buffer+$E3, D1                   ; D1 is pointer to buffer
	MOVE.L	#$00000094, D2                    ; Size of buffer
	BSR	SWI_OS_ConvertHex8
	MOVE.B	#$20, buffer+$EB

	MOVE.L	#$41373A00, buffer+$EC            ; 'A7:' + $00 terminator
	MOVE.L	A7, D0
	MOVE.L	#buffer+$EF, D1                   ; D1 is pointer to buffer
	MOVE.L	#$00000094, D2                    ; Size of buffer
	BSR	SWI_OS_ConvertHex8
	MOVE.B	#$20, buffer+$F7

	MOVE.L	#$50433A00, buffer+$F8            ; 'PC:' + $00 terminator

	MOVE.L	#buffer, D0                       ; D1 is pointer to buffer
	BSR	SWI_OS_Write0                     ; And write it to

	BSR	SWI_OS_NewLine
	BRA	monitor_loop


;	                                                    TRSM-INT---XNZVC
;	 D0:00000000 D1:00000000 D2:00000000 D3:00000000 SR:0000000000000000
;	 D4:00000000 D5:00000000 D6:00000000 D7:00000000 US:00000000
;	 A0:00000000 A1:00000000 A2:00000000 A3:00000000 SS:000007FC
;	 A4:00000000 A5:00000000 A6:00000000 A7:00000000 PC:00000000


monitor_help
	MOVE.L	#msg_mon_help, D0                 ; Location of message in A0
	BSR	SWI_OS_Write0
	RTS

; -----------------------------------------------------------------------------
; *QUIT
; -----------------------------------------------------------------------------
quit
	BSR	not_supported
	RTS

; -----------------------------------------------------------------------------
; *S
; -----------------------------------------------------------------------------
s
	BSR	get_hex_value                     ; Return hex value of string at (A6) in D1
	BCS	s_help                            ; Jump if the parameter wasn't valid
	BSR	skip_spaces                       ; Loop until (A6) isn't ASCII value 32 (space)
	BNE	s_help                            ; Jump if there were more parameters
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	RTS

s_help
	MOVE.L	#msg_s_help, D0                   ; Location of message in A0
	BSR	SWI_OS_Write0
	RTS

; -----------------------------------------------------------------------------
; *TZAP
; -----------------------------------------------------------------------------
tzap
	BSR	not_supported
	RTS

; -----------------------------------------------------------------------------
; *XFER
; -----------------------------------------------------------------------------
xfer
	BSR	get_hex_value                     ; Return hex value of string at (A6) in D1
	BCS	xfer_help                         ; Jump if the parameter wasn't valid
	MOVE.L	D1, D7                            ; Move host address to D7

	BSR	skip_spaces                       ; Loop until (A6) isn't ASCII value 32 (space)
	BSR	get_hex_value                     ; Return hex value of string at (A6) in D1
	BCS	xfer_help                         ; Jump if the parameter wasn't valid
	MOVE.L	D1, D6                            ; Move parasite address to D6

	BSR	skip_spaces                       ; Loop until (A6) isn't ASCII value 32 (space)
	BSR	get_hex_value                     ; Return hex value of string at (A6) in D1
	BCS	xfer_help                         ; Jump if the parameter wasn't valid
	MOVE.L	D1, D5                            ; Move length to D5

	BSR	skip_spaces                       ; Loop until (A6) isn't ASCII value 32 (space)
	MOVE.B	(A6)+, D0
	ANDI.B	#$DF, D0                          ; Make character upper-case
	CMP.B	#'R', D0                          ; Is it a R(ead)?
	BEQ	xfer2
	CMP.B	#'W', D0                          ; ...or a W(rite)?
	BNE	xfer_help
	MOVE.B	#$06, D4
	BRA	xfer2a
xfer2
	MOVE.B	#$07, D4                          ; Put R/W in D4
xfer2a
	BSR	skip_spaces                       ; Loop until (A6) isn`t ASCII value 32 (space)
	BEQ	xfer2b                            ; End of command string has been reached
	MOVE.B	(A6)+, D0
	ANDI.B	#$DF, D0                          ; Make character upper-case
	CMP.B	#'S', D0                          ; Is it a S(hadow memory)?
	BEQ	xfer3
	CMP.B	#'M', D0                          ; ...or a M(ain memory)?
	BNE	xfer_help
	MOVE.B	#$40, D3                          ; Byte for Main memory
	BRA	xfer3a

xfer2b
	MOVE.B	#$10, D0                          ; Byte for Shadow memory
	BRA	xfer3a

xfer3
	MOVE.B	#$20, D3
xfer3a
	BSR	skip_spaces                       ; Loop until (A6) isn`t ASCII value 32 (space)
	BNE	go_help                           ; Jump if there were more parameters

	MOVE.B	#$0E, buffer                      ; XY+00: Send block length
	MOVE.B	#$10, buffer+1                    ; XY+01: Receive block length
	MOVE.L	D7, buffer+2                      ; XY+02: Host address
	MOVE.L	D6, buffer+6                      ; XY+06: 68k address
	MOVE.W	D5, buffer+10                     ; XY+0A: Length
	MOVE.B	D4, buffer+12                     ; XY+0C: Function (&06 for write, &07 for read)
	MOVE.B	D3, buffer+13                     ; XY+0D: Memory access control

	MOVE.L	#buffer, D1                       ; D1 = pointer to control block
	MOVE.L	#$000000FA, D0                    ; OSWORD &FA call
	BSR	SWI_OS_Word

	RTS

xfer_help
	MOVE.L	#msg_xfer_help, D0                ; Location of message in A0
	BSR	SWI_OS_Write0
	RTS



; -----------------------------------------------------------------------------
; Vector handlers
;
; Default vector $01 ErrorV handler
; -----------------------------------------------------------------------------
_errorv
	BSR	SWI_OS_NewLine

	ADD.L	#$04, D0                          ; Ignore error number for now
	BSR	SWI_OS_Write0                     ; Print error message at (D0)

	CMP.L	#$00000000, error_pc
	BEQ	_errorv1
	MOVE.L	#msg_errorv_at, D0                ; Location of message in A0
	BSR	SWI_OS_Write0
	MOVE.L	error_pc, D0
	MOVE.L	#buffer, D1
	MOVE.B	#$FF, D2
	BSR	SWI_OS_ConvertHex8
	BSR	SWI_OS_Write0

_errorv1
	BSR	SWI_OS_NewLine
	MOVE.L	(ramsize), A7                     ; Reset stack pointer
	BRA	main_loop

; -----------------------------------------------------------------------------
; Default vector $02 IrqV handler
; -----------------------------------------------------------------------------
_irqv
	MOVE.L	D0, -(SP)                         ; Push D0 to the stack
	MOVE.L	#err_unknown_interrupt, D0        ; Location of message in D0
	MOVE.L	#0, error_pc
	BSR	SWI_OS_GenerateError              ; Print the message
	MOVE.L	(SP)+, D0                         ; Pull D0 from the stack
	RTS

; -----------------------------------------------------------------------------
; Default vector $03 WrchV handler
; -----------------------------------------------------------------------------
_wrchv
	BTST	#6, r1status                      ; Check the NF1 flag of TUBE ULA Status register 1 (1st data register not filled)
	BEQ	SWI_OS_WriteC                     ; Loop until TUBE ULA Data register 2 is empty
	MOVE.B	D0, r1data                        ; ?r1data = D0.B

	IFNE DEBUG
		MOVE.L	D0, -(SP)
		MOVE.L	D1, -(SP)
		MOVE.B	D0, D1
		MOVE.B	#6, D0
		TRAP	#15
		MOVE.L	(SP)+, D1
		MOVE.L	(SP)+, D0
	ENDC

	RTS

; -----------------------------------------------------------------------------
; Default vector $10 EventV handler
; -----------------------------------------------------------------------------
_eventv
	RTS

; -----------------------------------------------------------------------------
; Default vector $18 UKSWIV handler
; -----------------------------------------------------------------------------
_ukswiv
	MOVE.L	D0, -(SP)                         ; Push D0 to the stack
	MOVE.L	#err_no_such_swi, D0              ; Location of message in D0
	MOVE.L	#0, error_pc
	BSR	SWI_OS_GenerateError              ; Print the message
	MOVE.L	(SP)+, D0                         ; Pull D0 from the stack
	RTS

; -----------------------------------------------------------------------------
; Default vector $1E ChangeEnvironmentV handler
; -----------------------------------------------------------------------------
_ChangeEnvironmentV
	CMP	#6, D0
	BEQ	_ChangeEnvironmentV_6             ; ErrorV
	CMP	#12, D0
	BEQ	_ChangeEnvironmentV_10            ; EventV
	CMP	#12, D0
	BEQ	_ChangeEnvironmentV_12            ; UKSWIV
	MOVE.L	#err_bad_environment_number, D0
	ORI	#$02, CCR                         ; Set overflow fag
	RTS


_ChangeEnvironmentV_6
	CMPI.L	#$00000000, D1
	BNE	_ChangeEnvironmentV_6_write
	MOVE.L	errorv-bios_address, D1
	RTS

_ChangeEnvironmentV_6_write
	MOVE.L	D2, -(SP)                         ; Push D2 to the stack
	MOVE.L	errorv-bios_address, D2
	MOVE.L	D1, errorv-bios_address
	EXG	D2, D0
	MOVE.L	(SP)+, D2                         ; Pull D2 from the stack

	CMPI.L	#$00000000, D3
	BNE	_ChangeEnvironmentV_6b_write
	MOVE.L	errorbuffer, D3
	RTS

_ChangeEnvironmentV_6b_write
	MOVE.L	D2, -(SP)                         ; Push D2 to the stack
	MOVE.L	errorbuffer, D2
	MOVE.L	D3, errorbuffer
	EXG	D2, D3                            ; D3 is previous error buffer address
	MOVE.L	(SP)+, D2                         ; Pull D2 from the stack
	RTS


_ChangeEnvironmentV_10
	CMPI.L	#$00000000, D1
	BNE	_ChangeEnvironmentV_10_write
	MOVE.L	eventv-bios_address, D1
	RTS

_ChangeEnvironmentV_10_write
	MOVE.L	D2, -(SP)                         ; Push D2 to the stack
	MOVE.L	eventv-bios_address, D2
	MOVE.L	D1, eventv-bios_address
	EXG	D2, D1
	MOVE.L	(SP)+, D2                         ; Pull D2 from the stack
	RTS


_ChangeEnvironmentV_12
	CMPI.L	#$00000000, D1
	BNE	_ChangeEnvironmentV_12_write
	MOVE.L	ukswiv-bios_address, D1
	RTS

_ChangeEnvironmentV_12_write
	MOVE.L	D2, -(SP)                         ; Push D2 to the stack
	MOVE.L	ukswiv-bios_address, D2
	MOVE.L	D1, ukswiv-bios_address
	EXG	D2, D1
	MOVE.L	(SP)+, D2                         ; Pull D2 from the stack
	RTS

; -----------------------------------------------------------------------------
; Default escape handler (no software vector)
; -----------------------------------------------------------------------------
_escapev
	BSET.B	#6, escapeflag                    ; Set the Escape flag
	RTS



; -----------------------------------------------------------------------------
; Various utility functions
;
; Loop until (A0) isn`t ASCII value 32 (space)
; -----------------------------------------------------------------------------
skip_spaces
	CMP.B	#' ', (A6)+                       ; Is it ASCII value 32 (space)?
	BEQ	skip_spaces                       ; No? Then check next byte
	MOVE.B	-(A6), D0                         ; Put the first non-SPACE character in D0
	CMP.B	#$0D, D0                          ; Check if <cr>
	RTS

; -----------------------------------------------------------------------------
; Return hex value of string at (A6) in D1
; -----------------------------------------------------------------------------
get_hex_value
	CLR	D1                                ; Clear contents of D1

get_hex_value1
	MOVE.B	(A6)+, D0                         ; Get byte in D0
	MOVE.L	D0, -(SP)                         ; Push D0 to the stack

	CMPI.B	#'0', D0
	BLO	get_hex_value3                    ; Invalid character
	CMPI.B	#'9', D0
	BLS	get_hex_value2                    ; Character is between 0-9, so process it now

	ANDI.B	#%11011111, D0                    ; Make character upper-case
	CMPI.B	#'A', D0
	BLO	get_hex_value3                    ; Invalid character
	CMPI.B	#'F', D0
	BHI	get_hex_value3                    ; Invalid character
	SUBI.B	#$07, D0

get_hex_value2
	ANDI.B	#%00001111, D0                    ; Mask out the highest nibble
	ASL.L	#4, D1                            ; Shift D1 to make room for new nibble
	ADD.B	D0, D1                            ; Add the nibble to the result

	MOVE.L	(SP)+, D0                         ; Pull D0 from the stack (not used; just to clean things up)
	BRA	get_hex_value1

get_hex_value3
	MOVE.L	(SP)+, D0                         ; Pull D0 from the stack

	CMPI.B	#$0D, D0                          ; Is the next character CR?
	BEQ	get_hex_value4                    ; Yes? Then exit with C=0

	CMPI.B	#' ', D0                          ; Is the next character SPACE?
	BEQ	get_hex_value4                    ; Yes? Then exit with C=0

	SUBQ	#1, A6                            ; Optimize?
	ORI	#$01, CCR                         ; Set the carry bit and exit ; 00000160	Bad hex
	RTS

get_hex_value4
	SUBQ	#1, A6                            ; Optimize?
	ANDI	#$FE, CCR                         ; Clear the carry-bit and exit
	RTS


; -----------------------------------------------------------------------------
; Set interrupt level
; -----------------------------------------------------------------------------
set_interrupt
	ANDI.L	#$7, D0                           ; Mask out other bits
	ASL.W	#8, D0                            ; 

	MOVE.L	D1, -(SP)                         ; Push D1 to the stack
;	MOVE.C	CCR, D1                           ; Get current status register
	ANDI.W	#$F8FF, D1                        ; Mask out the ISR bits
	OR.W	D0, D1                            ; And set the interrupt level
;	MOVE.C	D1, CCR

	MOVE.L	(SP)+, D1                         ; Pull A6 from the stack
	RTS


; -----------------------------------------------------------------------------
; Strings and error messages
; -----------------------------------------------------------------------------
startup_message1
	DC.B	$0A
	DC.B	'Acorn 68', $00

startup_message2
	DC.B	' second processor ', $07, $00

startup_message_cpus
	DC.B	'008000010020030040060070300'

startup_message3
	DC.B	'K', $0A, $0D, $0A, $0A, $00

msg_swi_removecursors
	DC.B	23, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0

msg_swi_restorecursors
	DC.B	23, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0

msg_help
	DC.B	$0A
	DC.B	$0D
;	DC.B	'68030 TUBE ', bios_version, $0A, $0D, $00
	DC.B	'CiscOS v1.99e (20150222)', $0A, $0D, $00

msg_help_std
	DC.B	'   SWI', $0A, $0D
	DC.B	'   TUBE', $0A, $0D, $00

msg_help_extd
	DC.B	'   ERROR [number] <message>', $0A, $0D
	DC.B	'   FLASH <fsp>', $0A, $0D
	DC.B	'   GO <addr>', $0A, $0D
	DC.B	'   MON', $0A, $0D
	DC.B	'   S <addr>', $0A, $0D
	DC.B	'   SR <start addr> <end start> <"string">', $0A, $0D
	DC.B	'   TZAP [addr]', $0A, $0D
	DC.B	'   XFER <io addr.> <addr> <length> <R>|<W> [S|M]', $0A, $0D, $00

msg_flash_help
	DC.B	'Syntax : FLASH <fsp>', $0A, $0D, $00

msg_go_help
	DC.B	'Syntax : GO <addr>', $0A, $0D, $00

msg_mon_help
	DC.B	'Syntax : MON', $0A, $0D, $00

msg_s_help
	DC.B	'Syntax : S <addr>', $0A, $0D, $00

msg_sr_help
	DC.B	'Syntax : SR <start addr> <end start> <"string">', $0A, $0D, $00

msg_tzap_help
	DC.B	'Syntax : TZAP [addr]', $0A, $0D, $00

msg_xfer_help
	DC.B	'Syntax : XFER <io addr.> <addr> <length> <R|W> [S|M]', $0A, $0D, $00

msg_monitor
	DC.B	'CiscOS Monitor', $00

msg_monitor_help
	DC.B	'D  Dump memory', $0A, $0D
	DC.B	'F  Fill <start addr> <end addr> <fill Byte|Word|Long>', $0A, $0D
	DC.B	'G  Go', $0A, $0D
	DC.B	'H  Help', $0A, $0D
	DC.B	'L  Load', $0A, $0D
	DC.B	'Q  Quit', $0A, $0D
	DC.B	'R  Set register contents', $0A, $0D
	DC.B	'S  Save', $0A, $0D
	DC.B	'T  Trace', $0A, $0D
	DC.B	'V  View register contents', $0A, $0D, $00

msg_monitor_unknown_cmd
	DC.B	'Unknown command, use H for help.', $0A, $0D, $00

msg_statusregister
	DC.B	'T-S--INT---XNZVC', $00

msg_flash_wrongfile
	DC.B	'File is not a CiscOS file', $0A, $0D, $00

msg_flash_warning
	DC.B	'WARNING! THIS WILL FLASH THE BIOS', $0A, $0D
	DC.B	'USE AT YOUR OWN RISK!', $0A, $0D, $0A, $0D
	DC.B	'Do you want to continue? (Y/N) : ', $00

msg_flash_identify
	DC.B	'Flash ROM ID=$', $00

msg_flash_start
	DC.B	$0A, $0D
	DC.B	'Flashing sector ', $0A, $0D, $00

msg_flash_finish
	DC.B	'Flash completed successfully', $0A, $0D, $00

msg_flash_error
	DC.B	'Flash error at ', $00

msg_days_short
	DC.B	'MonTueWedThuFriSatSun', $00

msg_months_short
	DC.B	'JanFebMarAprMayJunJulAugSepOctNovDec', $00

msg_errorv_at
	DC.B	' at ', $00

msg_bus_error
	DC.B	'Bus error', $0A, $0D, $00

msg_address_error
	DC.B	'Address error exception', $0A, $0D, $00

msg_address_error_mf
	DC.B	' Access type&function:', $00

msg_address_error_aa
	DC.B	' Access address      :', $00

msg_address_error_ir
	DC.B	' Instruction register:', $00

msg_address_error_sr
	DC.B	' Status register     :TRSM-INT---XNZVC', $0A, $0D
	DC.B	'                      ', $00

msg_address_error_pc
	DC.B	' Program counter     :', $00

	ALIGN	4
err_illegal_instruction
	DC.L	$80000100
	DC.B	'Illegal instruction', $00

	ALIGN	4
err_unknown_interrupt
	DC.L	$80000104
	DC.B	'Unknown IRQ at &', $00

	ALIGN	4
err_int_divide_by_zero
	DC.L	$80000169
	DC.B	'Integer divide by zero', $00

	ALIGN	4
err_privilege_violation
	DC.L	$80000008
	DC.B	'Privilege violation', $00

	ALIGN	4
err_out_of_range
	DC.L	$00000001
	DC.B	'Out of range', $00

	ALIGN	4
err_escape
	DC.L	$00000011
	DC.B	'Escape', $00

	ALIGN	4
err_not_a_language
	DC.L	$000000FF
	DC.B	'This is not a language', $00

	ALIGN	4
err_cannot_run_code
	DC.L	$000000FF
	DC.B	'I cannot run this code', $00

	ALIGN	4
err_unknown_exception
	DC.L	$000000FF
	DC.B	'Unknown exception', $00

	ALIGN	4
err_not_supported
	DC.L	$000000FF
	DC.B	'Not supported', $00

	ALIGN	4
err_bad_environment_number
	DC.L	$000001B0
	DC.B	'Bad environment number', $00

	ALIGN	4
err_no_such_swi
	DC.L	$000001E6
	DC.B	'No such SWI', $00

	ALIGN	4
err_bad_vector_number
	DC.L	$000001A0
	DC.B	'Bad vector number', $00

	ALIGN	4
err_unknown_percent_field
	DC.L	$000002C2
	DC.B	'Unknown ''%'' field', $00

	ALIGN	4
err_bad_fscontrol_call
	DC.L	$00000401
	DC.B	'Bad FSControl call', $00

	ALIGN	4
err_unaligned_address
	DC.L	$00000807
	DC.L	'Unaligned address', $00

; -----------------------------------------------------------------------------
; Data
;
; Table which specifies OSWORD call parameter block length
; -----------------------------------------------------------------------------
osword_send_table
	                                          ; Length of parameterblock to be send to the host processor
	DC.B	$00                               ; OSWORD $01: Parameter block length = $00
	DC.B	$05                               ; OSWORD $02: Parameter block length = $05
	DC.B	$00                               ; OSWORD $03: Parameter block length = $00
	DC.B	$05                               ; OSWORD $04: Parameter block length = $05
	DC.B	$02                               ; OSWORD $05: Parameter block length = $02
	DC.B	$05                               ; OSWORD $06: Parameter block length = $05
	DC.B	$08                               ; OSWORD $07: Parameter block length = $08
	DC.B	$0E                               ; OSWORD $08: Parameter block length = $0E
	DC.B	$04                               ; OSWORD $09: Parameter block length = $04
	DC.B	$01                               ; OSWORD $0A: Parameter block length = $01
	DC.B	$01                               ; OSWORD $0B: Parameter block length = $01
	DC.B	$05                               ; OSWORD $0C: Parameter block length = $05
	DC.B	$00                               ; OSWORD $0D: Parameter block length = $00
	DC.B	$01                               ; OSWORD $0E: Parameter block length = $01
	DC.B	$20                               ; OSWORD $0F: Parameter block length = $20
	DC.B	$10                               ; OSWORD &10: Parameter block length = &10
	DC.B	$0D                               ; OSWORD &11: Parameter block length = &0D
	DC.B	$00                               ; OSWORD &12: Parameter block length = &00
	DC.B	$04                               ; OSWORD &13: Parameter block length = &04
	DC.B	$80                               ; OSWORD &14: Parameter block length = &80

osword_recv_table
	                                          ; Length of parameterblock to be received from the host processor
	DC.B	$05                               ; OSWORD &01: Parameter block length = &05
	DC.B	$00                               ; OSWORD &02: Parameter block length = &00
	DC.B	$05                               ; OSWORD &03: Parameter block length = &05
	DC.B	$00                               ; OSWORD &04: Parameter block length = &00
	DC.B	$05                               ; OSWORD &05: Parameter block length = &05
	DC.B	$00                               ; OSWORD &06: Parameter block length = &00
	DC.B	$00                               ; OSWORD &07: Parameter block length = &00
	DC.B	$00                               ; OSWORD &08: Parameter block length = &00
	DC.B	$05                               ; OSWORD &09: Parameter block length = &05
	DC.B	$09                               ; OSWORD &0A: Parameter block length = &09
	DC.B	$05                               ; OSWORD &0B: Parameter block length = &05
	DC.B	$00                               ; OSWORD &0C: Parameter block length = &00
	DC.B	$08                               ; OSWORD &0D: Parameter block length = &08
	DC.B	$18                               ; OSWORD &0E: Parameter block length = &18
	DC.B	$00                               ; OSWORD &0F: Parameter block length = &00
	DC.B	$01                               ; OSWORD &10: Parameter block length = &01
	DC.B	$0D                               ; OSWORD &11: Parameter block length = &0D
	DC.B	$80                               ; OSWORD &12: Parameter block length = &80
	DC.B	$04                               ; OSWORD &13: Parameter block length = &04
	DC.B	$80                               ; OSWORD &14: Parameter block length = &80

; -----------------------------------------------------------------------------
; SWI jump and lookup table
; -----------------------------------------------------------------------------
; Align first entry to long word; SWI string is padded with spaces to align to long word
	ALIGN	4

SWITable
	DC.L	$00000000                         ; SWI number
	DC.L	SWI_OS_WriteC                     ; Pointer to handler
	DC.B	'OS_WriteC', $00, $00, $00        ; SWI name followed by terminator and alignment bytes

	DC.L	$00000001
	DC.L	SWI_OS_WriteS
	DC.B	'OS_WriteS', $00, $00, $00

	DC.L	$00000002
	DC.L	SWI_OS_Write0
	DC.B	'OS_Write0', $00, $00, $00

	DC.L	$00000003
	DC.L	SWI_OS_NewLine
	DC.B	'OS_NewLine', $00, $00

	DC.L	$00000004
	DC.L	readcv
	DC.B	'OS_ReadC', $00, $00, $00, $00

	DC.L	$00000005
	DC.L	cliv-bios_address
	DC.B	'OS_CLI', $00, $00

	DC.L	$00000006
	DC.L	bytev
	DC.B	'OS_Byte', $00

	DC.L	$00000007
	DC.L	wordv
	DC.B	'OS_Word', $00

	DC.L	$00000008
	DC.L	filev
	DC.B	'OS_File', $00

	DC.L	$00000009
	DC.L	argsv
	DC.B	'OS_Args', $00

	DC.L	$0000000A
	DC.L	bgetv
	DC.B	'OS_BGet', $00

	DC.L	$0000000B
	DC.L	bputv
	DC.B	'OS_BPut', $00

	DC.L	$0000000C
	DC.L	gbpbv
	DC.B	'OS_GBPB', $00

	DC.L	$0000000D
	DC.L	findv
	DC.B	'OS_Find', $00

	DC.L	$0000000E
	DC.L	readlinev
	DC.B	'OS_ReadLine', $00

	DC.L	$0000000F
	DC.L	SWI_OS_Control
	DC.B	'OS_Control', $00, $00

	DC.L	$00000010
	DC.L	SWI_OS_GetEnv
	DC.B	'OS_GetEnv', $00, $00, $00

	DC.L	$00000012
	DC.L	SWI_OS_SetEnv
	DC.B	'OS_SetEnv', $00, $00, $00

	DC.L	$00000013
	DC.L	SWI_OS_IntOn
	DC.B	'OS_IntOn', $00, $00, $00, $00

	DC.L	$00000014
	DC.L	SWI_OS_IntOff
	DC.B	'OS_IntOff', $00, $00, $00

	DC.L	$00000016
	DC.L	SWI_OS_EnterOS
	DC.B	'OS_EnterOS', $00, $00

	DC.L	$00000019
	DC.L	SWI_OS_UnusedSWI
	DC.B	'OS_UnusedSWI', $00, $00, $00, $00

	DC.L	$00000029
	DC.L	fscontrolv
	DC.B	'OS_FSControl', $00, $00, $00, $00

	DC.L	$0000002B
	DC.L	SWI_OS_GenerateError
	DC.B	'OS_GenerateError', $00, $00, $00, $00

	DC.L	$0000002C
	DC.L	SWI_OS_ReadEscapeState
	DC.B	'OS_ReadEscapeState', $00, $00

	DC.L	$00000032
	DC.L	SWI_OS_ReadPoint
	DC.B	'OS_ReadPoint', $00, $00, $00, $00

	DC.L	$00000034
	DC.L	SWI_OS_CallAVector
	DC.B	'OS_CallAVector', $00, $00

	DC.L	$00000036
	DC.L	SWI_OS_RemoveCursors
	DC.B	'OS_RemoveCursors', $00, $00, $00, $00

	DC.L	$00000037
	DC.L	SWI_OS_RestoreCursors
	DC.B	'OS_RestoreCursors', $00, $00, $00

	DC.L	$00000038
	DC.L	SWI_OS_SWINumberToString
	DC.B	'OS_SWINumberToString', $00, $00, $00, $00

	DC.L	$00000039
	DC.L	SWI_OS_SWINumberFromString
	DC.B	'OS_SWINumberFromString', $00, $00

	DC.L	$0000003A
	DC.L	SWI_OS_ValidateAddress
	DC.B	'OS_ValidateAddress', $00, $00

	DC.L	$0000003F
	DC.L	SWI_OS_CheckModeValid
	DC.B	'OS_CheckModeValid', $00, $00, $00

	DC.L	$00000040
	DC.L	SWI_OS_ChangeEnvironment
	DC.B	'OS_ChangeEnvironment', $00, $00, $00, $00

	DC.L	$00000045
	DC.L	SWI_OS_Plot
	DC.B	'OS_Plot', $00

	DC.L	$00000046
	DC.L	SWI_OS_WriteN
	DC.B	'OS_WriteN', $00, $00, $00

	DC.L	$00000056
	DC.L	SWI_OS_SetECFOrigin
	DC.B	'OS_SetECFOrigin', $00

	DC.L	$0000005D
	DC.L	SWI_OS_PrintChar
	DC.B	'OS_PrintChar', $00, $00, $00, $00

	DC.L	$0000005B
	DC.L	SWI_OS_CRC
	DC.B	'OS_CRC', $00, $00

	DC.L	$0000007C
	DC.L	SWI_OS_LeaveOS
	DC.B	'OS_LeaveOS', $00, $00

	DC.L	$0000007D
	DC.L	SWI_OS_ReadLine32
	DC.B	'OS_ReadLine32', $00, $00, $00

	DC.L	$000000D0
	DC.L	SWI_OS_ConvertHex1
	DC.B	'OS_ConvertHex1', $00, $00

	DC.L	$000000D1
	DC.L	SWI_OS_ConvertHex2
	DC.B	'OS_ConvertHex2', $00, $00

	DC.L	$000000D2
	DC.L	SWI_OS_ConvertHex4
	DC.B	'OS_ConvertHex4', $00, $00

	DC.L	$000000D3
	DC.L	SWI_OS_ConvertHex6
	DC.B	'OS_ConvertHex6', $00, $00

	DC.L	$000000D4
	DC.L	SWI_OS_ConvertHex8
	DC.B	'OS_ConvertHex8', $00, $00

	DC.L	$000000D5
	DC.L	SWI_OS_ConvertCardinal1
	DC.B	'OS_ConvertCardinal1', $00

	DC.L	$000000D6
	DC.L	SWI_OS_ConvertCardinal2
	DC.B	'OS_ConvertCardinal2', $00

	DC.L	$000000D7
	DC.L	SWI_OS_ConvertCardinal3
	DC.B	'OS_ConvertCardinal3', $00

	DC.L	$000000D8
	DC.L	SWI_OS_ConvertCardinal4
	DC.B	'OS_ConvertCardinal4', $00

	DC.L	$000000DD
	DC.L	SWI_OS_ConvertBinary1
	DC.B	'OS_ConvertBinary1', $00, $00, $00

	DC.L	$000000DE
	DC.L	SWI_OS_ConvertBinary2
	DC.B	'OS_ConvertBinary2', $00, $00, $00

	DC.L	$000000DF
	DC.L	SWI_OS_ConvertBinary3
	DC.B	'OS_ConvertBinary3', $00, $00, $00

	DC.L	$000000E0
	DC.L	SWI_OS_ConvertBinary4
	DC.B	'OS_ConvertBinary4', $00, $00, $FF          ; End of table marker


; -----------------------------------------------------------------------------
; Built-in application jump table
; -----------------------------------------------------------------------------

OS_CLI_command_table
	ALIGN	4
	DC.L	star_error
	DC.B	'ERROR  ', $00

	DC.L	flash
	DC.B	'FLASH  ', $00

	DC.L	go
	DC.B	'GO ', $00

	DC.L	help
	DC.B	'HELP   ', $00

	DC.L	mon
	DC.B	'MON', $00

	DC.L	quit
	DC.B	'QUIT   ', $00

	DC.L	tzap
	DC.B	'TZAP   ', $00

	DC.L	s
	DC.B	'S  ', $00

	DC.L	xfer
	DC.B	'XFER   ', $00

	DC.L	$FFFFFFFF                         ; End of table marker
	DC.B	$00

; -----------------------------------------------------------------------------
; OS_FSControl_jumptable
; -----------------------------------------------------------------------------
	ALIGN	4

OS_FSControl_jumptable
	DC.L	not_supported                     ; &00: Set current directory
	DC.L	not_supported                     ; &01: Set library directory
	DC.L	not_supported                     ; &02: Inform of start of new app
	DC.L	not_supported                     ; &03: Run type (internal use only)
	DC.L	not_supported                     ; &04: Run a file
	DC.L	not_supported                     ; &05: Catalogue a directory
	DC.L	not_supported                     ; &06: Examine current directory
	DC.L	not_supported                     ; &07: Catalogue library
	DC.L	not_supported                     ; &08: Examine library
	DC.L	not_supported                     ; &09: Examine objects
	DC.L	SWI_OS_FSControl0A                ; &0A: Set file system options
	DC.L	not_supported                     ; &0B
	DC.L	not_supported                     ; &0C
	DC.L	not_supported                     ; &0D
	DC.L	not_supported                     ; &0E
	DC.L	not_supported                     ; &0F
	DC.L	not_supported                     ; &10
	DC.L	not_supported                     ; &11
	DC.L	not_supported                     ; &12
	DC.L	not_supported                     ; &13
	DC.L	not_supported                     ; &14
	DC.L	not_supported                     ; &15
	DC.L	SWI_OS_FSControl16                ; &16: Close all open files


; -----------------------------------------------------------------------------
; SWI_Debugger_Disasseble table
; -----------------------------------------------------------------------------
	ALIGN	4

SWI_Debugger_Disassemble_table
	DC.W	$FEC0, $E6C0
	DC.B    'RO               ', $00
	DC.W	$FEC0, $E4C0
	DC.B    'ROX              ', $00
	DC.W	$FEC0, $E2C0
	DC.B    'LS               ', $00
	DC.W	$FEC0, $E0C0
	DC.B    'AS               ', $00
	DC.W	$F018, $E018
	DC.B    'RO               ', $00
	DC.W	$F018, $E010
	DC.B    'ROX              ', $00
	DC.W	$F018, $E008
	DC.B    'LS               ', $00
	DC.W	$F018, $E000
	DC.B    'AS               ', $00
	DC.W	$F0C0, $D0C0
	DC.B    'ADD       <EA>,A@', $00
	DC.W	$F130, $D100
	DC.B    'ADDX             ', $00
	DC.W	$F000, $D000
	DC.B    'ADD              ', $00
	DC.W	$F1F8, $C188
	DC.B    'EXG              ', $00
	DC.W	$F1F8, $C148
	DC.B    'EXG              ', $00
	DC.W	$F1F8, $C140
	DC.B    'EXG              ', $00
	DC.W	$F1F0, $C100
	DC.B    'ABCD             ', $00
	DC.W	$F1C0, $C1C0
	DC.B    'MULS             ', $00
	DC.W	$F1C0, $C0C0
	DC.B    'MULU             ', $00
	DC.W	$F000, $C000
	DC.B    'AND              ', $00
	DC.W	$F0C0, $B0C0
	DC.B    'CMP     <EA>,A@  ', $00
	DC.W	$F138, $B108
	DC.B    'CMPM             ', $00
	DC.W	$F100, $B100
	DC.B    'EOR              ', $00
	DC.W	$F000, $B000
	DC.B    'CMP              ', $00
	DC.W	$F0C0, $90C0
	DC.B    'SUB       <EA>,A@', $00
	DC.W	$F130, $9100
	DC.B    'SUBX             ', $00
	DC.W	$F000, $9000
	DC.B    'SUB              ', $00
	DC.W	$F1F0, $8100
	DC.B    'SBCD             ', $00
	DC.W	$F1C0, $81C0
	DC.B    'DIVS             ', $00
	DC.W	$F1C0, $80C0
	DC.B    'DIVU             ', $00
	DC.W	$F000, $8000
	DC.B    'OR               ', $00
	DC.W	$F100, $7000
	DC.B    'MOVEQ            ', $00
	DC.W	$FF00, $6100
	DC.B    'BSR              ', $00
	DC.W	$FF00, $6000
	DC.B    'BRA              ', $00
	DC.W	$F000, $6000
	DC.B    'B                ', $00
	DC.W	$F0F8, $50C8
	DC.B    'DB               ', $00
	DC.W	$F0C0, $50C0
	DC.B    'S                ', $00
	DC.W	$F100, $5100
	DC.B    'SUBQ             ', $00
	DC.W	$F100, $5000
	DC.B    'ADDQ             ', $00
	DC.W	$F1C0, $41C0
	DC.B    'LEA              ', $00
	DC.W	$F1C0, $4180
	DC.B    'CHK              ', $00
	DC.W	$FFC0, $4EC0
	DC.B    'JMP              ', $00
	DC.W	$FFC0, $4E80
	DC.B    'JSR              ', $00
	DC.W	$FFFF, $4E77
	DC.B    'RTR              ', $00
	DC.W	$FFFF, $4E76
	DC.B    'TRAPV            ', $00
	DC.W	$FFFF, $4E75
	DC.B    'RTS              ', $00
	DC.W	$FFFF, $4E73
	DC.B    'RTE              ', $00
	DC.W	$FFFF, $4E72
	DC.B    'STOP             ', $00
	DC.W	$FFFF, $4E71
	DC.B    'NOP              ', $00
	DC.W	$FFFF, $4E70
	DC.B    'RESET            ', $00
	DC.W	$FFF8, $4E68
	DC.B    'MOVE FROM USP    ', $00
	DC.W	$FFF8, $4E60
	DC.B    'MOVE TO USP      ', $00
	DC.W	$FFF8, $4E58
	DC.B    'UNLINK           ', $00
	DC.W	$FFF8, $4E50
	DC.B    'LINK             ', $00
	DC.W	$FFF0, $4E40
	DC.B    'TRAP             ', $00
	DC.W	$FF80, $4C80
	DC.B    'MOVEM FROM REG.  ', $00
	DC.W	$FFC0, $4AC0
	DC.B    'TAS              ', $00
	DC.W	$FF00, $4A00
	DC.B    'TST              ', $00
	DC.W	$FFF8, $48C0
	DC.B    'EXT.L            ', $00
	DC.W	$FFF8, $4880
	DC.B    'EXT.W            ', $00
	DC.W	$FF80, $4880
	DC.B    'MOVEA TO REG.    ', $00
	DC.W	$FFF8, $4840
	DC.B    'SWAP             ', $00
	DC.W	$FFC0, $4840
	DC.B    'PEA              ', $00
	DC.W	$FFC0, $4800
	DC.B    'NBCD             ', $00
	DC.W	$FFC0, $46C0
	DC.B    'MOVE TO SR       ', $00
	DC.W	$FF00, $4600
	DC.B    'NOT              ', $00
	DC.W	$FFC0, $44C0
	DC.B    'MOVE TO CCR      ', $00
	DC.W	$FF00, $4400
	DC.B    'NEG              ', $00
	DC.W	$FF00, $4200
	DC.B    'CLR              ', $00
	DC.W	$FFC0, $40C0
	DC.B    'MOVE.W SR,       ', $00
	DC.W	$FF00, $4000
	DC.B    'NEGX             ', $00
	DC.W	$F000, $3000
	DC.B    'MOVE.W           ', $00
	DC.W	$F000, $2000
	DC.B    'MOVE.L           ', $00
	DC.W	$F000, $1000
	DC.B    'MOVE.B           ', $00
	DC.W	$FF00, $0C00
	DC.B    'CMP       #      ', $00
	DC.W	$FF00, $0A00
	DC.B    'EOR       #      ', $00
	DC.W	$FF00, $0600
	DC.B    'ADD       #      ', $00
	DC.W	$FF00, $0400
	DC.B    'SUB       #      ', $00
	DC.W	$FF00, $0200
	DC.B    'AND       #      ', $00
	DC.W	$FF00, $0000
	DC.B    'OR        #      ', $00
	DC.W	$F138, $0108
	DC.B    'MOVEP            ', $00
	DC.W	$FFC0, $08C0
	DC.B    'BSET             ', $00
	DC.W	$FFC0, $0880
	DC.B    'BCLR             ', $00
	DC.W	$FFC0, $0840
	DC.B    'BCHG             ', $00
	DC.W	$FFC0, $0800
	DC.B    'BTST             ', $00
	DC.W	$F1C0, $01C0
	DC.B    'BSET             ', $00
	DC.W	$F1C0, $0180
	DC.B    'BCLR             ', $00
	DC.W	$F1C0, $0140
	DC.B    'BCHG             ', $00
	DC.W	$F1C0, $0100
	DC.B    'BTST             ', $00
	DC.W	$0000

	ORG     bios_address+$00007FFF-$3B5D
EasterEggImage
	INCBIN	"68000C"

	END	START

*~Font name~Courier New~
*~Font size~10~
*~Tab type~1~
*~Tab size~4~
